/*
 * Copyright (c) 1998-2002 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package jgroup.test.performance.upgrade;

import java.rmi.RemoteException;

import jgroup.core.GroupManager;
import jgroup.core.JgroupException;
import jgroup.core.MemberId;
import jgroup.core.MembershipListener;
import jgroup.core.View;
import jgroup.test.performance.SpeedTest;
import jgroup.upgrade.UpgradeListener;
import jgroup.util.Abort;

import org.apache.log4j.Logger;

/**
 *  Jgroup server for measuring the performance of GMI invocations and
 *  server upgrade/recovery time.
 *
 *  The <code>UpgradeListener</code> is implemented to allow replicas of
 *  <code>UpgradeServer</code> to be upgraded on the fly.  The magic is
 *  dealt with using reflection on this class to construct a group
 *  manager that contains also the upgrade layer.  This class simply
 *  have to implement the <code>upgraded</code> method of the
 *  <code>UpgradeListener</code> interface.
 *
 *  @author Alberto Montresor
 *  @author Hein Meling
 *  @since Jgroup 1.2
 */
public class UpgradeServer
  implements SpeedTest, MembershipListener, UpgradeListener
{

/** Obtain logger for this class */
  private static Logger log = Logger.getLogger(UpgradeServer.class);

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields (non-shared state)
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Keep the time of the last view change */
  private long lastViewChange = 0;

  /** The member identifier of this replica. */
  private MemberId me;

  /** Buffer for storing log output */
  private StringBuilder buf = new StringBuilder();

  /** True indicates that this replica may fail at some random point in time */
  private static boolean randomFailure = false;

  /** Last view received */
  private View lastView;

  /** number of test invocations of each of the test methods: 1 - test(),
   * 2 - mtest(), 3 - test(byte[]), 4 -mtest(byte[])
   */
  private int test[] = new int[4];
  
  /** 
   * times of the first invocation of this method
   */
  private long testStart[] = new long[4];
  private long testStop[] = new long[4];
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructor
  ////////////////////////////////////////////////////////////////////////////////////////////

  public UpgradeServer()
    throws JgroupException
  {
    buf.append("UPG1------------------------------------------------------\n");
    buf.append("Log time : Number of members : Time since last view change\n");

    GroupManager gm = GroupManager.getGroupManager(this);
    log.info("UpgradeServer initialized");
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from SpeedTest
  ////////////////////////////////////////////////////////////////////////////////////////////

  public void mtest()
    throws RemoteException
  {
     if(test[1] == 0) testStart[1]=System.currentTimeMillis();
     test[1]++;
    testStop[1] = System.currentTimeMillis();
  }

  public byte[] mtest(byte[] x)
    throws RemoteException
  {
    if(test[3] == 0) testStart[3]=System.currentTimeMillis();
    test[3]++;
    testStop[3] = System.currentTimeMillis();
    return x;
  }

  public void vmtest(byte[] x)
    throws RemoteException
  {
    if(test[3] == 0) testStart[3]=System.currentTimeMillis();
    test[3]++;
    testStop[3] = System.currentTimeMillis();
  }

  public String mtest(String x)
    throws RemoteException
  {
    if (log.isDebugEnabled()) log.debug("mtest: received: " + x);
    if (log.isDebugEnabled()) log.debug("Last view: " + lastView);
    log.info("Invocation number since last reset: [0] - " + test[0] +
             ", [1] - " + test[1] +", [2] - " + test[2] +", [3] - " + test[3] );
    
    float f[] = new float[4];
    long cTime= System.currentTimeMillis();
    for(int i=0; i<4; i++){
      if(test[i] >1) f[i] = (float)(test[i]-1)/(testStop[i]-testStart[i]);
      test[i]=0;
    }
    log.info("Average invocation frequency: reset: [0] - " + f[0] +
             ", [1] - " + f[1] +", [2] - " + f[2] +", [3] - " + f[3] );
    
    return x;
  }

  public void test()
    throws RemoteException
  {
    if(test[0] == 0) testStart[0]=System.currentTimeMillis();
     test[0]++;
    testStop[0] = System.currentTimeMillis();
  }

  public byte[] test(byte[] x)
    throws RemoteException
  {
    if(test[2] == 0) testStart[2]=System.currentTimeMillis();
    test[2]++;
    testStop[2] = System.currentTimeMillis();
    return x;
  }

  public String test(String x)
    throws RemoteException
  {
    return x;
  }

  /**
   *  Shut down the server
   */
  public void shutdown()
    throws RemoteException
  {
  }
/*    new Thread() {
      public void run() {
        try {
          sleep(random.nextInt(MAX_SHUTDOWN_DELAY));
        } catch (InterruptedException e) {
        }
        Abort.exit(0);
      }
    }.start();
  }

*/
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from UpgradeListener
  ////////////////////////////////////////////////////////////////////////////////////////////

  public void upgraded()
  {
    System.out.println("I've been upgraded!  Now I should leave the group and exit.");
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from MembershipListener
  ////////////////////////////////////////////////////////////////////////////////////////////

  public void viewChange(View view)
  {
    long currentTime = System.currentTimeMillis();

    buf.append(currentTime);
    buf.append(" : ");
    buf.append(view.size());
    buf.append(" : ");
    buf.append((lastViewChange != 0) ? (currentTime - lastViewChange) : 0);
    buf.append("\n");
    buf.append(view);
    buf.append("\n");
    if (log.isDebugEnabled()) log.debug(buf.toString());

    /* Keep the time of this view change */
    lastViewChange = currentTime;

    if (randomFailure && view.memberHasPosition(0, me)) {
      try {
        shutdown();
      } catch (RemoteException e) {
      }
    }
    lastView = view;
  }


  public void hasLeft() { 
  	if (log.isDebugEnabled()) log.debug("has_left() called");
  }


  public void prepareChange() { 
    if (log.isDebugEnabled()) log.debug("prepareChange() called");
    }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Main method
  ////////////////////////////////////////////////////////////////////////////////////////////

  public static void main(String[] argv)
    throws JgroupException
  {
    for (int i = 0 ; i < argv.length ; i++) {
      if (argv[i].equals("--randomFailure")) {
        randomFailure = true;
      } else {
        System.err.println("Unknown option " + argv[i]);
        Abort.usage("upgrserver [--randomFailure]");
      }
    }

    UpgradeServer server = new UpgradeServer();
  }

} // END UpgradeServer
