/*
 * Copyright (c) 1998-2002 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package jgroup.test.performance.upgrade;

import java.util.Date;

import jgroup.core.ConfigManager;
import jgroup.core.registry.DependableRegistry;
import jgroup.core.registry.RegistryFactory;
import jgroup.test.performance.SpeedTest;
import jgroup.upgrade.UpgradeManagementClient;

import org.apache.log4j.Logger;



/**
 *  Test suite for measuring the performance of a server during the upgrade process.
 *  The class parses the experiment parameter from command line. The following
 *  parameters are interpreted:
 *
 *  <p>
 *  @param -method type of test calls sent to the server
 *  @param -calibrate number of calibration calls
 *  @param -rate the rate at which requests are sent to the test server
 *  @param -time the time of the cycle in seconds
 *  @param -cycles number of cycles of request bursts
 *  @param -upgradeStart the cycle number at which the upgrade starts
 *  @param -clients number of local clients generating the requests
 *  @param -server name at which the test server is registered at the repository
 *  @param -verbose verbose way of performing the experiment 
 *
 *  @author       Marcin Solarski
 *  @since        Jgroup 2.1
 */
public class UpgradeTestSuite
{


  /** Obtain logger for this class */
  private static Logger log = Logger.getLogger(UpgradeTestSuite.class);
  private static UpgradeTestSuite instance = null;


  public static void main(String[] argv) throws Exception
  {
    ConfigManager.init();
    if (instance == null)
      instance = new UpgradeTestSuite(argv);
  }
  
  public UpgradeTestSuite(String argv[]) throws Exception {
    if( argv.length <1 ) help();
    boolean verbose = PUtil.getSwitch(argv, "-v");
    int method = PUtil.getIntParameter(argv, "-method", "1");
    int calibReqCallNo = PUtil.getIntParameter(argv, "-calibrate", "10");
    int rps = PUtil.getIntParameter(argv, "-rate", "1");
    int expTime = PUtil.getIntParameter(argv, "-time", "10");
    int cycles = PUtil.getIntParameter(argv, "-cycles", "1");
    int upgrade_cycle = PUtil.getIntParameter(argv, "-upgradeStart", "-1");
    int client_no = PUtil.getIntParameter(argv, "-clients", "1");
    String serverName = PUtil.getStringParameter(argv, "-server", "UpgradeServer");
    if( verbose ) {
       String date = new Date(System.currentTimeMillis()).toString();
       log.info("Experiment started at: " + date);
       log.info("Testing suite with the following parameters: ");
       log.info("Server name:                \t"+ serverName);
       log.info("Number of calibration calls:\t" + calibReqCallNo);
       log.info("Test method:                \t" + method);
       log.info("Request rate per sec:       \t" + rps);
       log.info("Cycle time:                 \t" + expTime);
       log.info("Cycles number:              \t"+ cycles);
       log.info("Staring upgrade at cycle:   \t" + upgrade_cycle);
    }

    ClientThreadPool client[] = new ClientThreadPool[client_no];

    int times = expTime*rps;
    DependableRegistry reg = RegistryFactory.getRegistry();
    SpeedTest server = (SpeedTest) reg.lookup(serverName);
    int call_no = 0, time_tick = expTime/cycles;
    UpgradeStart upgradeT;
    upgradeT = new UpgradeStart();
    ClientThreadPool.init(client_no, cycles);
    for(int i=0; i<client_no; i++) {
        client[i]= new ClientThreadPool(method, calibReqCallNo, rps, times, server);
        client[i].start();
    }
    for(int j=0; j< cycles; j++) { 
      //try{ Thread.sleep(5000); } catch(InterruptedException e) {};
      ClientThreadPool.allStart();
      if(j == upgrade_cycle ) upgradeT.start();
    }
   
    for(int i=0; i< client_no; i++) 
      client[i].CycleDone(cycles);
    ClientThreadPool.allStop();

    //header printout
    StringBuilder header = new StringBuilder("Cycle\tT0");
    for(int i=0; i< client_no; i++) 
      header.append("\t" + "RT[" + i +"]");
    log.info(header);

    for(int j=1; j< cycles; j++) {
      long roundTime = j>1 ? ClientThreadPool.getStartCycleTime(j)-ClientThreadPool.getStartCycleTime(j-1) : 0;
      StringBuilder line = new StringBuilder(""+j + "\t" + roundTime);
      for(int i=0; i< client_no; i++) 
          line.append("\t" + ClientThreadPool.getRespTime(i, j));
      log.info(line);
    }
      
    //log.info("Stopping all threads");
    //ClientThreadPool.allStop();
    System.exit(0);
  } 

  static void help() {
     System.out.println("UpgradeTestSuite [-v][-method M][-calibrate C][-rate R][-time T][cycles Y][-clientsi N][-upgradeStart U][-server ServerName]");
  } 


  class UpgradeStart extends Thread { 
	  String argv[] = new String[] {"u"}; 
     public void run(){
	 try { UpgradeManagementClient.main(argv);
	 } catch(Exception e) {log.error("Exception in main", e);}

     }


  }

  
}
