/*
 * Copyright (c) 1998-2002 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package jgroup.arm;

import static jgroup.util.log.ReplicaEvent.Type.Failed;
import static jgroup.util.log.ReplicaEvent.Type.Shutdown;

import java.io.BufferedReader;
import java.io.InputStreamReader;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.lang.reflect.Modifier;

import jgroup.core.ConfigManager;
import jgroup.core.GroupManager;
import jgroup.core.arm.ExecException;
import jgroup.core.arm.ExecService;
import jgroup.relacs.config.AppConfig;
import jgroup.util.log.Eventlogger;
import jgroup.util.log.ReplicaEvent;

import org.apache.log4j.LogManager;
import org.apache.log4j.Logger;

/**
 *  The application starter class is wrapper for starting a server replica
 *  in a separate child JVM process, and sets up a pipe between the parent
 *  and child process.  This allows the parent process (the ExecDaemon) to
 *  control the child process (the server replica), for example the ExecDaemon
 *  may wish to kill the replica.  Note that, unless the server replica is
 *  using the <code>Replication</code> service, or if the server has specific
 *  shutdown needs, the server should consider adding a shutdown hook so as
 *  to ensure that the replica releases any held resources when being killed.
 *
 *  This class can also be used to start a server through its constructor
 *  without having to recompile it.  That is, this wrapper will take care of
 *  obtaining the group manager for the server.  This means that the server
 *  can be made into a replicated server without changing it.  This assumes
 *  that the server does not need to use any of the services or listeners
 *  provided through the Jgroup group manager.  For simple replication
 *  this should be no problem.
 *
 *  @author Hein Meling
 *  @since Jgroup 2.1
 */
public final class ApplicationStarter
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Logger
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Obtain logger for this class */
  private static final Logger log = Logger.getLogger(ApplicationStarter.class);


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /* The application object */
  private Object obj;

  /* 
   * The GroupManager of the application associated with 
   * this ApplicationStarter object
   */
  private GroupManager gm;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructor
  ////////////////////////////////////////////////////////////////////////////////////////////

  public ApplicationStarter(final String[] args)
    throws ExecException, ClassNotFoundException
  {
    Class c = Class.forName(args[0]);
    String mainError = args[0]
      + " does not implement a 'public static void main(String[] args)' method.";

    final AppConfig app = AppConfig.getApplication(c);
    /*
     * Set up a shutdown hook for this ApplicationStarter object, to
     * allow detection of replica crash in offline analysis.
     */
    Runtime.getRuntime().addShutdownHook(new Thread("AS-ShutdownHook") {
      public void run() {
        log.info("AS-ShutdownHook: shutdown for " + app);
        if (Eventlogger.ENABLED) {
          Eventlogger.logEventFlush(new ReplicaEvent(Failed, app.getGroupId()));
        }
    }});

    try {

      /*
       * Check for 'main(String[] args)' method in the specified class.
       */
      Method main = c.getDeclaredMethod("main", new Class[] { String[].class });
      int m = main.getModifiers();
      if (!Modifier.isPublic(m) || !Modifier.isStatic(m)) {
        throw new ExecException(mainError);
      }
      
      /* Copy the argument part of the array */
      String[] argv = new String[args.length-1];
      if (argv.length > 0)
        System.arraycopy(args, 1, argv, 0, args.length);

      /*
       * Invoke the main method to start the replica in this JVM.
       */
      main.invoke(null, new Object[] { argv });

    } catch (InvocationTargetException e) {
      log.warn(args[0] + " has thrown an exception", e.getCause());
      throw new ExecException(args[0]
        + " has thrown an exception during invocation of the main method.", e.getCause());
    } catch (IllegalAccessException e) {
      throw new ExecException(args[0] + " failed to start", e);
    } catch (NoSuchMethodException e) {
      throw new ExecException(mainError, e);
    }

    /*
     * This thread connects to its parent process to receive commands.
     * That is, the execdaemon creates a replica through the ApplicationStarter
     * and uses this command controller to shut down the replica once
     * it should be removed.
     */
    Thread shutdownThrd = new Thread("AS-CommandParser") {
      public void run() {
        BufferedReader in = new BufferedReader(new InputStreamReader(System.in));
        String cmd;
        try {
          while ((cmd = in.readLine()) != null) {
            if (log.isDebugEnabled()) {
              log.debug("COMMAND RECEIVED: " + cmd);
            }
            if (cmd.equals(ExecService.SHUTDOWN_REPLICA)) {
              if (Eventlogger.ENABLED) {
                Eventlogger.logEventFlush(new ReplicaEvent(Shutdown, app.getGroupId()));
              }
              LogManager.shutdown();
              System.out.println(new ReplicaEvent(Shutdown, app.getGroupId()));
              Runtime.getRuntime().halt(0);
            }
          }
          in.close();
        } catch (java.io.IOException e) {
          log.error("Replica failed reading command from ExecDaemon", e);
        }
      }
    };
    shutdownThrd.setPriority(Thread.NORM_PRIORITY+1);
    shutdownThrd.setDaemon(true);
    shutdownThrd.start();

    if (log.isDebugEnabled())
      log.debug("Created replica: " + args[0]);
  }


  public ApplicationStarter(String className)
    throws Exception
  {
    Class appClass = Class.forName(className);
    AppConfig app = AppConfig.getApplication(appClass);
    obj = appClass.newInstance();
    gm = GroupManager.getGroupManager(obj);
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // ApplicationStarter methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  public void shutdown()
  {
    System.out.println(obj.getClass().getName() + " shutting down...");
    System.exit(0);
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Main method
  ////////////////////////////////////////////////////////////////////////////////////////////

  public static void main(String[] args)
    throws Exception
  {
    ConfigManager.init();
    new ApplicationStarter(args);
  }

} // END ApplicationStarter
