/*
 * Copyright (c) 1998-2002 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */
 
package jgroup.arm;

import java.io.IOException;
import java.io.ObjectInput;
import java.io.ObjectOutput;
import java.net.InetAddress;
import java.rmi.RemoteException;

import jgroup.core.arm.ARMEvent;
import jgroup.core.arm.DistributionScheme;
import jgroup.relacs.config.DistributedSystemConfig;
import jgroup.relacs.config.Host;
import jgroup.util.Network;

import org.apache.log4j.Logger;

/**
 *  Event to notify the replication manager of the availability of an
 *  execution daemon running on the initiating host.
 *
 *  @author Hein Meling
 *  @since Jgroup 1.2
 */
final class EDPresentEvent
  implements ARMEvent
{

  private static final long serialVersionUID = 5047243468822963123L;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Logger
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Obtain logger for this class */
  private static final Logger log = Logger.getLogger(EDPresentEvent.class);


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Private Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  An integer based <code>InetAddress</code> for the machine running
   *  the execution daemon that generated this event.
   */
  private int intAdr;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructors
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Default constructor for externalization.
   */
  public EDPresentEvent() { }


  /**
   *  Construct a new <code>EDPresentEvent</code> instance.
   */
  EDPresentEvent(InetAddress inet)
  {
    intAdr = Network.translate(inet);
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // ARMEvent interface methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  The event handling is performed by the event object itself, at the
   *  (replication manager) receiver side. <p>
   *
   *  The handler method is invoked by the replication manager, allowing
   *  an execution daemon to notify its presence to the replication
   *  manager.  The host location for the execution daemon is provided
   *  through the constructor. <p>
   * 
   *  We update the internal tables of the host object associated with
   *  the internet address attached with this object.  We assume that
   *  the notifying host is specified in the configuration, otherwise
   *  nothing happens.
   *
   *  @exception RemoteException
   *    Raised if the remote operation failed.
   *  @exception java.net.UnknownHostException
   *    Raised if the invoking host could not be found through DNS.
   */
  public void handle(DistributionScheme distScheme)
    throws Exception
  {
    InetAddress inet = Network.translate(intAdr);
    if (log.isDebugEnabled())
      log.debug("Received new host: " + inet);
    Host host = DistributedSystemConfig.getHost(inet);
    if (host != null) {
      // update the host internal tables
      host.suspect();
    }
  }


  /* (non-Javadoc)
   * @see jgroup.core.arm.ARMEvent#getGroupId()
   */
  public int getGroupId()
  {
    return -1;
  }


  /* (non-Javadoc)
   * @see jgroup.core.arm.ARMEvent#getObject()
   */
  public Object getObject()
  {
    return null;
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from the Externalizable interface (as required by the ARMEvent interface)
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Restores the content of this object from the marshalled data
   *  contained in the specified input stream.
   *
   *  @param in
   *    The stream to be read from.
   *  @exception IOException
   *    Thrown if there was problems unmarshalling the input stream.
   *  @exception ClassNotFoundException
   *    Thrown if the local classloader cannot find a class marshalled
   *    in the stream.
   */
  public void readExternal(ObjectInput in)
    throws IOException, ClassNotFoundException
  {
    intAdr = in.readInt();
  }


  /**
   *  Marshals the content of this object to the specified output
   *  stream.
   * 
   *  @param out
   *    The stream to write this object into.
   *  @exception IOException
   *    Thrown if there was problems unmarshalling the input stream.
   */
  public void writeExternal(ObjectOutput out)
    throws IOException
  {
    out.writeInt(intAdr);
  }

} // END EDPresentEvent
