/*
 * Copyright (c) 1998-2004 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package jgroup.arm;

import static jgroup.util.log.ViewEvent.Type.ARM;

import java.io.Serializable;
import java.util.Timer;
import java.util.TimerTask;
import java.util.concurrent.locks.Lock;
import java.util.concurrent.locks.ReentrantLock;

import jgroup.core.MemberId;
import jgroup.core.MemberTable;
import jgroup.core.View;
import jgroup.core.arm.RecoveryStrategy;
import jgroup.relacs.config.AppConfig;
import jgroup.relacs.config.Host;
import jgroup.relacs.config.HostSet;
import jgroup.util.log.Eventlogger;
import jgroup.util.log.ViewEvent;

import org.apache.log4j.Logger;

/**
 * @author Hein Meling
 */
public class GroupData
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Logger
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Obtain logger for this class */
  private static final Logger log = Logger.getLogger(GroupData.class);


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Transient fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** The period to await a view install before initiating recovery actions */
  private transient int awaitViewPeriod;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** The application associated with this group data object */
  private final AppConfig app;

  /** The currently known (including crashed) members of the application group */
  private final MemberTable membertable = new MemberTable();

  /** The current set of assigned hosts */
  private HostSet assignedHosts = new HostSet();

  /** The set of removed hosts (permanently) */
  private final HostSet removedHosts = new HostSet();

  /** The set of new hosts since the last view change */
  private final HostSet newHosts = new HostSet();

  /** The view monitor timer for this application */
  private Timer viewMonitorTimer = null;

  /** Field indicating if this RM replica is the leader or not */
  private boolean iamLeader;

  /** The most recent view installed by this RM replica */
  private View view;

  /** Lock to ensure exclusive access to the view monitor timer */
  private final Lock lock = new ReentrantLock();


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructor
  ////////////////////////////////////////////////////////////////////////////////////////////

  public GroupData(AppConfig app, int awaitViewPeriod)
  {
    this.app = app;
    this.awaitViewPeriod = awaitViewPeriod;
  }
  

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  void addHosts(HostSet hosts)
  {
    if (assignedHosts.isEmpty()) {
      this.assignedHosts = hosts;
      newHosts.addHosts(hosts);
    } else {
      for (Host host : hosts) {
        addHost(host);
      }
    }
    if (log.isDebugEnabled()) {
      log.debug("assignedHosts: " + assignedHosts);
      log.debug("     newHosts: " + newHosts);
    }
  }

  void addHost(Host newHost)
  {
    if (log.isDebugEnabled())
      log.debug("adding: " + newHost);
    if (!removedHosts.containsHost(newHost)) {
      assignedHosts.addHost(newHost);
      newHosts.addHost(newHost);
      ReplicaCount.updateHost(newHost);
    }
    if (log.isDebugEnabled()) {
      log.debug("assignedHosts: " + assignedHosts);
      log.debug("     newHosts: " + newHosts);
    }
  }

  void removeHost(Host host)
  {
    if (log.isDebugEnabled())
      log.debug("removing: " + host);
    removedHosts.addHost(host);
    boolean removed = assignedHosts.removeHost(host);
    if (removed) {
      /*
       * The above removeHost() did actually remove the host from
       * the host set, and thus we must invoke the removeReplica().
       */
      ReplicaCount.removeReplica(host);
    }
    if (log.isDebugEnabled()) {
      log.debug("assignedHosts: " + assignedHosts);
      log.debug(" removedHosts: " + removedHosts);
    }
  }

  HostSet getAssignedHosts()
  {
    return assignedHosts;
  }

  void viewChangeEvent(View view)
  {
    /* Update the view for this application group. */
    app.viewChange(view);
    logStatus();
    /*
     * All RM replicas should log this ViewChangeEvent, since in
     * some crash scenarios it would not be logged.  Offline analysis
     * of the logs should detect the duplicate entries.
     */
    if (Eventlogger.ENABLED || log.isDebugEnabled()) {
      ViewEvent viewEvent = new ViewEvent(ARM, view);
      if (Eventlogger.ENABLED)
        Eventlogger.logEventFlush(viewEvent);
      if (log.isDebugEnabled())
        log.debug(viewEvent.shortToString());
    }
    /* Reschedule the view monitor for this application */
    scheduleViewMonitor();
    if (iamLeader) {
      System.out.println(view);
    }
  }

  void viewChange(View view, boolean iamLeader)
  {
    this.view = view;
    this.iamLeader = iamLeader;
    if (log.isDebugEnabled())
      log.debug("Updating group: " + app.getGroupId());
    membertable.viewChange(view);
    MemberId[] members = membertable.members();
    for (int i=0; i < members.length; i++) {
      switch (membertable.getStatus(members[i])) {
      case MemberTable.NEWMEMBER:
        /* This is a new member */
        if (log.isDebugEnabled())
          log.debug("New member " + members[i]);
        membertable.put(members[i], new HostSet());
        break;
      case MemberTable.RECOVERING:
        /* This member has crashed and then recovered */
        if (log.isDebugEnabled())
          log.debug("New incarnation " + members[i]);
        membertable.put(members[i], new HostSet());
        break;
      case MemberTable.SURVIVED:
        /* This member is survived from the previous view */
        if (newHosts.isEmpty())
          break;
        if (log.isDebugEnabled())
          log.debug("Inserting new hosts: " + newHosts);
        HostSet hosts = (HostSet) membertable.get(members[i]);
        if (log.isDebugEnabled())
          log.debug("Before: " + members[i] + " = " + hosts);
        hosts.addHosts(newHosts);
        if (log.isDebugEnabled())
          log.debug("After : " + members[i] + " = " + hosts);
        break;
      }
    }
    newHosts.removeAllHosts();
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Inner class and methods for state merge
  ////////////////////////////////////////////////////////////////////////////////////////////

  public static class GroupMergeRecord
    implements Serializable
  {

    private static final long serialVersionUID = 932291238562617927L;

    /*
     * Note that we need to transfer the <code>AppConfig</code> object as
     * they are stored in the AppConfig static tables, and the serialized
     * copy of these objects may contain a view (and other information)
     * not known to the receiving replication manager.
     */
    private AppConfig app;

    /*
     * The set of hosts removed locally (to be removed)
     */
    private HostSet removedHosts;

    /*
     * The set of hosts added locally (to be added)
     */
    private HostSet hosts;

    GroupMergeRecord(AppConfig app, HostSet removedHosts, HostSet hosts)
    {
      this.app = app;
      this.removedHosts = (HostSet) removedHosts.clone();
      // no need to clone 'hosts' as it is already a clone from getState()
      this.hosts = hosts;
      if (log.isDebugEnabled()) {
        log.debug("getState: gid: " + getGroupId() +
            ", toadd: " + hosts +
            ", toremove: " + removedHosts);
      }
    }

    public int getGroupId()
    {
      return app.getGroupId();
    }

  } // END GroupMergeRecord

  GroupMergeRecord getMergeRecord(MemberId[] dests, MemberId me)
  {
    synchronized (this) {
      return new GroupMergeRecord(app, removedHosts, getState(dests, me));
    }
  }

  void putMergeRecord(GroupMergeRecord merge)
  {
    if (log.isDebugEnabled()) {
      logStatus();
      log.debug("putState: gid: " + merge.getGroupId());
      log.debug("toremove: " + merge.removedHosts);
      log.debug("   toadd: " + merge.hosts);
    }
    putState(merge.removedHosts, merge.hosts);
    logStatus();
  }

  private HostSet getState(MemberId[] dests, MemberId me)
  {
    HostSet hosts = ((HostSet) membertable.get(me));
    hosts = (HostSet) hosts.clone();
    if (log.isDebugEnabled())
      log.debug("Starting slist: " + hosts);
    for (int i = 0; i < dests.length; i++) {
      if (dests[i].equals(me))
        hosts.removeHosts((HostSet) membertable.get(dests[i]));
    }
    if (log.isDebugEnabled())
      log.debug("Local removed: " + removedHosts);
    hosts.removeHosts(removedHosts);
    if (log.isDebugEnabled())
      log.debug("Final: " + hosts);
    return hosts;
  }

  synchronized private void putState(HostSet toremove, HostSet toadd)
  {
    toremove.removeHosts(removedHosts);
    removedHosts.addHosts(toremove);
    for (Host host : toremove) {
      if (log.isDebugEnabled())
        log.debug("Removing " + host + " from assigned host set");
      boolean removed = assignedHosts.removeHost(host);
      if (removed)
        ReplicaCount.removeReplica(host);
    }

    for (Host host : toadd) {
      if (!removedHosts.containsHost(host) && !assignedHosts.containsHost(host)) {
        if (log.isDebugEnabled())
          log.debug("Adding " + host + " to assigned host set");
        assignedHosts.addHost(host);
        // Add the new host to the set of new hosts
        newHosts.addHost(host);
        ReplicaCount.updateHost(host);
      }
    }
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Log method
  ////////////////////////////////////////////////////////////////////////////////////////////

  public void logStatus()
  {
    if (log.isDebugEnabled()) {
      log.debug("GroupData: " + app);
      log.debug("  assignedHosts=" + assignedHosts);
      log.debug("  removedHosts =" + removedHosts);
      log.debug("  newHosts     =" + newHosts);
      log.debug("  ViewMonitor is " + (viewMonitorTimer != null ? "" : "not ") + "scheduled");
    }
  }

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Private ViewMonitor class and support methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * Cancel and nullify the view monitor, if a view monitor
   * has been scheduled.
   */
  public void cancelViewMonitor()
  {
    lock.lock();
    try {
      if (viewMonitorTimer != null) {
        viewMonitorTimer.cancel();
        viewMonitorTimer = null;
      }
    } finally {
      lock.unlock();
    }
  }

  /**
   * Reschedule this view monitor if recovery is still needed.
   * Otherwise, the view monitor timer is canceled.
   */
  public void scheduleViewMonitor()
  {
    lock.lock();
    try {
      if (viewMonitorTimer != null) {
        // cancel exiting timer for this application
        viewMonitorTimer.cancel();
      }
      if (app.needsRecovery()) {
        // reschedule and waiting for the next view
        viewMonitorTimer = new Timer("ViewMonitor-" + app.getGroupId(), true);
        if (log.isDebugEnabled()) {
          log.debug("View monitor expires in " + awaitViewPeriod + "ms for " + app);
        }
        viewMonitorTimer.schedule(new ViewMonitor(), awaitViewPeriod);
      } else {
        // we reached the expected view size; nullify the timer field
        viewMonitorTimer = null;
        if (log.isDebugEnabled()) {
          log.debug("View monitor removed for " + app);
          logStatus();
        }
      }
    } finally {
      lock.unlock();
    }
  }

  /**
   * This timer task is run only when a viewChange does not follow a
   * replica (re)assignment.  This is required to capture failures
   * occuring immediately after creation, and no common view has been
   * installed including the new member.  A view change from live
   * members is otherwise required to trigger recovery in such failure
   * scenarios.
   */
  private class ViewMonitor
    extends TimerTask
  {

    /* (non-Javadoc)
     * @see java.util.TimerTask#run()
     */
    public void run()
    {
      // Ensures that this is the last execution of the current timer.
      // Note: it is safe to do this in the run() of the same timer execution,
      // since it guarantees that it will be the last run.
      cancelViewMonitor();
      if (log.isDebugEnabled()) {
        log.debug("View monitor expired for " + app);
      }
      if (!iamLeader) {
        if (log.isDebugEnabled())
          log.debug("I am not the leader RM replica; rescheduling view monitor");
        scheduleViewMonitor();
        return;
      } else if (view.size() == 1) {
        MemberId myId = view.getLeader();
        if (membertable.getStatus(myId) == MemberTable.NEWMEMBER) {
          if (log.isDebugEnabled())
            log.info("I am a single new member; rescheduling view monitor");
          scheduleViewMonitor();
          return;
        }
      }
      // get the recovery strategy to use for the application in question
      RecoveryStrategy rs = app.getRecoveryStrategy();
      boolean recovered = rs.handleFailure(assignedHosts);
      if (recovered) {
        if (log.isDebugEnabled())
          log.debug("Successful recovery for " + app);
        scheduleViewMonitor();
      } else {
        if (log.isDebugEnabled())
          log.warn("Unsuccessful recovery for " + app);
      }
    }

  } // END ViewMonitor

} // END GroupData
