/*
 * Copyright (c) 1998-2002 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package jgroup.arm;

import static jgroup.util.log.ReplicaEvent.Type.ForcedRemove;
import static jgroup.util.log.ReplicaEvent.Type.Leaving;
import static jgroup.util.log.ViewEvent.Type.Server;

import java.rmi.RemoteException;
import java.util.Timer;
import java.util.TimerTask;
import java.util.concurrent.locks.Lock;
import java.util.concurrent.locks.ReentrantLock;

import jgroup.core.JgroupException;
import jgroup.core.MemberId;
import jgroup.core.MembershipListener;
import jgroup.core.MembershipService;
import jgroup.core.View;
import jgroup.core.arm.RecoveryService;
import jgroup.core.arm.ReplicationManager;
import jgroup.core.arm.ShutdownListener;
import jgroup.core.registry.DependableRegistry;
import jgroup.core.registry.RegistryFactory;
import jgroup.relacs.config.AppConfig;
import jgroup.relacs.config.DistributedSystemConfig;
import jgroup.relacs.config.Domain;
import jgroup.relacs.config.DomainSet;
import jgroup.relacs.config.Host;
import jgroup.util.Util;
import jgroup.util.log.Eventlogger;
import jgroup.util.log.ReplicaEvent;
import jgroup.util.log.ViewEvent;

import org.apache.log4j.Logger;

/**
 *  Recovery layer for providing the replication manager with view
 *  change events from groups assoicated with the replication manager.
 *
 *  Note that the <code>RecoveryListener</code> does not provide any
 *  upcalls (it is only a marker interface), thus to listen for view
 *  change events, the replica should implement the
 *  <code>MembershipListener</code>.  A server using the recovery layer
 *  need not implement the <code>RecoveryListener</code>, instead it
 *  simply needs to specify "Recovery" in the layer stack.
 *
 *  @author Hein Meling
 *  @since Jgroup 1.2
 */
public class RecoveryLayer
  implements RecoveryService, MembershipListener
{ 

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Logger
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Obtain logger for this class */
  private static final Logger log = Logger.getLogger(RecoveryLayer.class);


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constants
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** The default delay before removing a replica */
  private static final int DEFAULT_REMOVE_DELAY   = 2000;

  /** The default max time before canceling a removal */
  private static final int DEFAULT_REMOVE_LATENCY = 5000;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Underlying group membership service. */
  private MembershipService membershipService;

  /**
   *  The shutdown listener; note that the server (listener) is not required
   *  to implement the ShutdownListener interface, but if it does this field
   *  will be set to the implementor allowing us to invoke the shutdown method.
   */
  private ShutdownListener shutdownListener = null;

  /** Reference to the replication manager. */
  private ReplicationManager replicaManager = null;

  /** Application info object for the replication manager. */
  private AppConfig rmApp;

  /** Application info object for the server object associated with this group manager. */
  private AppConfig thisApp;

  /** My member identifier. */
  private MemberId thisMember = null;

  /** This applications group identifier */
  private int gid;

  /** The local host */
  private Host localHost;

  /** Timer used to delay the remove task when new views keep arriving */
  private Timer timer = null;

  /** The delay before removing a replica after it has been elected */
  private int removeDelay;

  /**
   * The largest time allowed between canceling view changes; that is, the
   * amount of time to wait before starting the remove replica task.  If a
   * view change arrive before the timer expires, the timer will be canceled
   * and a new one instansiated.
   */
  private int removeLatency;

  /** 
   * Timer used to periodically renew the lease for this application.
   */
  private Timer livenessTimer = null;

  /** Lock object for the exclusive access to notify the RM */
  private final Lock lock = new ReentrantLock();


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructor
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *
   */
  private RecoveryLayer(MembershipService pgms)
    throws JgroupException, RemoteException
  {
    membershipService = pgms;
    thisMember = membershipService.getMyIdentifier();

    /* Obtain the replication manager application object */
    rmApp = AppConfig.getApplication(ReplicaManagerImpl.class);
    localHost = DistributedSystemConfig.getLocalHost();
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Static factory
  ////////////////////////////////////////////////////////////////////////////////////////////

  public static RecoveryLayer getLayer(MembershipService pgms)
    throws JgroupException, RemoteException
  {
    return new RecoveryLayer(pgms);
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Layer interface methods (inherited from RecoveryService interface)
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Add a listener for membership events.
   * 
   *  Since this layer sends view change events to the replication
   *  manager, we must provide special handling to use this layer
   *  from the replication manager.  The replication manager must
   *  provide self-recovery through its local view changes.  This
   *  layer is used by the replication manager to remove excessive
   *  replicas from the RM group.
   */
  public void addListener(Object listener)
  {
    if (listener == null)
      throw new NullPointerException("No replica specified for the RecoveryLayer");
    if (listener instanceof DependableRegistry)
      throw new IllegalArgumentException("RecoveryLayer cannot be used by the dependable registry");
    if (listener instanceof ShutdownListener)
      shutdownListener = (ShutdownListener) listener;
    String className = listener.getClass().getName();
    thisApp = AppConfig.getApplication(className);
    gid = thisApp.getGroupId();
    removeLatency = thisApp.getIntParam("Recovery.RemoveLatency", DEFAULT_REMOVE_LATENCY);
    removeDelay = thisApp.getIntParam("Recovery.RemoveDelay", DEFAULT_REMOVE_DELAY);

    /* Check if this is the server is the replication manager or not. */
    if (!className.equals(rmApp.getClassName())) {
      try {
        /* Obtain a remote reference for the dependable registry */
        DependableRegistry dregistry = RegistryFactory.getRegistry();
        /* Obtain a remote reference for the replication manager. */
        replicaManager = (ReplicationManager) dregistry.lookup(rmApp.getRegistryName());
        if (log.isDebugEnabled()) {
          log.debug("Replication Manager is available");
        }
      } catch (Exception e) {
        log.error("Replication Manager is unavailable", e);
        throw new IllegalStateException("Replication manager is unavailable", e);
      }
    }
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // MembershipListener interface methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Upcall that is invoked by the MembershipLayer when a view change
   *  occur in this replica's object group.
   *
   *  @param view
   *    The new view object group.
   */
  public void viewChange(final View view)
  {
    if (Eventlogger.ENABLED)
      Eventlogger.logEventFlush(new ViewEvent(Server, view));
    if (log.isDebugEnabled())
      log.debug("--- RecoveryLayer.viewChange: ---" + view);

    /*
     * If the application using this layer is the replication manager,
     * we simply avoid to execute this code, since the replication manager
     * should not notify itself of neither view changes or ping events.
     */
    if (replicaManager != null) {

      /* Recompute ping rate for the application */
      recomputePingRate(view.size());

      /*
       * Only the leader will notify the replication manager (correlator)
       * of view changes related to this replica's object group.  
       */
      if (view.memberHasPosition(0, thisMember)) {
//        doNotify(view);
        doNotifyThread(view);
      }
    }

    /*
     * Always cancel the remove timer if we receive a new view; it should only
     * remain active if the view size exceeds the initial redundancy level.
     */
    if (timer != null) {
      if (log.isDebugEnabled())
        log.debug("timer.cancel");
      timer.cancel();
    }
    /*
     * Determine if the current number of replicas (view size) exceed
     * the initial (max) redundancy for this application.  If it does,
     * we will remove one of these replicas at a time, until the
     * redundancy level is back to normal.  But first we wait to see
     * if there are more views.
     */
    if (thisApp.getInitialRedundancy() < view.size()) {
      timer = new Timer("RemoveReplicaTask-" + gid, true);
      if (log.isDebugEnabled())
        log.debug("timer.schedule");
      timer.schedule(new RemoveReplicaTask(view), removeLatency);
    }

  } // end of viewChange()


  /**
   * I'm the member to notify the replication manager of this view change.
   * We do it in a separate thread to avoid length delays (due to communication
   * latency) of the view install phase. 
   * 
   * @param view
   */
  private void doNotifyThread(final View view)
  {
    Thread viewNotifyThread = new Thread("ViewNotifyThread-" + gid) {
      public void run() {
        if (log.isDebugEnabled())
          log.debug("Notifying the Replication Manager: Lock pending");
        lock.lock();
        try {
          doNotify(view);
        } finally {
          lock.unlock();
        }
      }
    };
    viewNotifyThread.setDaemon(true);
    // This thread should have priority in order to receive the result from the RM
    viewNotifyThread.setPriority(Thread.NORM_PRIORITY+2);
    viewNotifyThread.start();
  }


  /**
   * I'm the member to notify the replication manager of this view change.
   * We do it inline of the viewChange() method to ensure that views recorded
   * at the replication manager are not reordered.  That is we want to ensure
   * that the replication manager has received this new view, before we allow
   * the membership service to continue installing another view.
   * 
   * @param view
   */
  private void doNotify(final View view)
  {
    try {
      if (log.isDebugEnabled())
        log.debug("Notifying the Replication Manager");
      replicaManager.notifyEvent(new ViewChangeEvent(view));
      if (log.isDebugEnabled())
        log.debug("Notified the Replication Manager");

    } catch (Exception e) {
      /*
       * The replication manager seems to be unavailable, or it was
       * not prepared to accept the above view change notification.
       */
      log.error("Failed to notify the Replication Manager of the new view", e);
    }
  }

  /* (non-Javadoc)
   * @see jgroup.core.MembershipListener#hasLeft()
   */
  public void hasLeft()
  {
    if (log.isDebugEnabled())
      log.debug("I've left the group.");
    removeReplica();
  }

  /* (non-Javadoc)
   * @see jgroup.core.MembershipListener#prepareChange()
   */
  public void prepareChange()
  {
    if (log.isDebugEnabled())
      log.debug("Invalid view; preparing for new view");
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Private Methods and Inner Class
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * Remove the replica instance of the associated application on the localhost.
   */
  private void removeReplica()
  {
    if (log.isDebugEnabled())
      log.debug("Removing replica instance from localhost");
    boolean removed = localHost.removeReplica(thisApp);
    if (log.isDebugEnabled())
      log.debug(removed ? "Replica removed" : "Replica already removed");
  }

  /**
   * Update the ping rate according to the current group size, and
   * restart the liveness timer.
   */
  private void recomputePingRate(int groupSize)
  {
    ReplicaPingEvent pingEvent = ReplicaPingEvent.getNewPingEvent(gid, groupSize);
    if (pingEvent != null) {
      if (livenessTimer != null) {
        livenessTimer.cancel();
      }
      livenessTimer = new Timer("LivenessTimer-" + gid, true);
      int pingRate = pingEvent.getPingRate();
      livenessTimer.schedule(new LivenessTask(pingEvent), pingRate, pingRate);
      if (log.isDebugEnabled())
        log.debug("Recomputed ping rate: " + pingRate + "ms.");
    }
  }


  /**
   * A timer task used periodically renew the lease when only a single
   * replica remains in the group.
   */
  private class LivenessTask
    extends TimerTask
  {
    private ReplicaPingEvent pingEvent;

    public LivenessTask(ReplicaPingEvent pingEvent)
    {
      this.pingEvent = pingEvent;
    }

    /* (non-Javadoc)
     * @see java.util.TimerTask#run()
     */
    public void run()
    {
      if (replicaManager != null) {
        try {
          if (log.isDebugEnabled())
            log.debug("Sending pingEvent: " + pingEvent);
          replicaManager.notifyEvent(pingEvent);
        } catch (Exception e) {
          /*
           * The replication manager seems to be unavailable.
           */
          log.error("Failed to renew the lease with the Replication Manager", e);
        }
      }
    }
  } // END of LivenessTask


  /**
   * A timer task used to remove the local replica if selected.
   */
  private class RemoveReplicaTask
    extends TimerTask
  {

    /* Period to wait before a force remove, halting the JVM */
    private static final long FORCE_REMOVE_DELAY = 10000;

    /* The most recently installed view on which this task will operate */
    private View view;

    /**
     * Constructs a new timer task object for the provided view.
     */
    public RemoveReplicaTask(View view)
    {
      super();
      this.view = view;
    }

    /**
     * The run method is invoked at the scheduled time, unless it is canceled.
     */
    public void run()
    {
      if (log.isDebugEnabled())
        log.debug("RemoveReplicaTask.run");
      if (removeMeNext(view)) {
        if (log.isDebugEnabled())
          log.debug("RemovingMe");
        Util.sleep(removeDelay);
        if (log.isDebugEnabled())
          log.debug("Replica leaving from localhost: " + localHost);

        /*
         * If for some reason the "clean" leave from the group fails to
         * complete, we do a force remove.
         */
        Timer forceRemove = new Timer("ForceRemove-" + gid, true);
        TimerTask removeTask = new TimerTask() {
          public void run() {
            if (Eventlogger.ENABLED) {
              Eventlogger.logEventFlush(new ReplicaEvent(ForcedRemove, gid));
            }
            log.warn("ForcedRemove");
            removeReplica();
          }
        };
        forceRemove.schedule(removeTask, FORCE_REMOVE_DELAY);

        try {
          /*
           * Exit the server group; note that we only have one group for each
           * group manager, thus we don't need to qualify this method with a
           * group identifier.
           */
          membershipService.leave();
          if (Eventlogger.ENABLED) {
            Eventlogger.logEventFlush(new ReplicaEvent(Leaving, gid));
          }
          /* 
           * If the server implements the ShutdownListener interface, invoke
           * the shutdown method in order to allow for application specific
           * clean up code.
           */
          if (shutdownListener != null)
            shutdownListener.shutdown();
        } catch (JgroupException e) {
          log.error("Failed to shutdown local replica", e);
        }
      }
    }


    //////////////////////////////////////////////////////////////////////////////////////////
    // Private Methods
    //////////////////////////////////////////////////////////////////////////////////////////

    /**
     * Determine if the local replica should be removed next, yet maintaining
     * the policy of replica distribution among domains.  We do this by
     * selecting the domain who has got the largest number of replicas in
     * this particular view, and from that selecting an arbitrary host.
     * If that host is the localhost we can release this replica.
     * 
     * @param view The current view to check if the local member is to be removed.
     * @return True is returned if the local member should be removed; otherwise
     *    false is returned.
     */
    private boolean removeMeNext(View view)
    {
      Domain largestReplicaDomain = DomainSet.getLargestReplicaDomain(view.getMembers());
      /* The host from which to remove a replica */
      Host host = largestReplicaDomain.getHostSet().removeFirst();
      if (log.isDebugEnabled())
        log.debug("Selected host: " + host);
      return localHost.equals(host);
    }

  } // END of inner class RemoveReplicaTask

} // END RecoveryLayer
