/*
 * Copyright (c) 1998-2004 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package jgroup.arm;

import java.io.IOException;
import java.io.ObjectInput;
import java.io.ObjectOutput;
import java.util.HashMap;
import java.util.Map;
import java.util.Timer;
import java.util.TimerTask;

import jgroup.core.arm.ARMEvent;
import jgroup.core.arm.DistributionScheme;
import jgroup.relacs.config.AppConfig;
import jgroup.relacs.types.ViewImpl;

import org.apache.log4j.Logger;

/**
 * @author Hein Meling
 * @since Jgroup 2.1
 */
public class ReplicaPingEvent
  implements ARMEvent
{

  private static final long serialVersionUID = -6041617835370010146L;

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Logger
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Obtain logger for this class */
  private static final Logger log = Logger.getLogger(ReplicaPingEvent.class);


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constants
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** The smallest allowed ping rate */
  private static final int MIN_PING_RATE     = 1000;

  /** The default ping period for lease renewal */
  private static final int DEFAULT_PING_RATE = 2000;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** The group identifier for issuer of this ping event */
  private int gid;

  /** The period to wait before alarming ARM */
  private int suspectPeriod;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Static and transient fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Map of renew timers (groupid -> renew timer for group) */
  private static transient final Map<Integer,Timer> renewTimers =
    new HashMap<Integer,Timer>(5);

  /** Access point to the replication manager for external events */
  private transient DistributionScheme distScheme;

  /**
   * The actual ping rate associated with this PingEvent object.
   * This is the period between each renewal invocations performed
   * by the <code>RecoveryLayer</code>.
   *
   * Note that since this field is transient, it will not be accessible
   * to JVMs creating this object using the no-argument constructor.
   * (deserialization).
   */
  private transient int pingRate;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructors
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * Default constructor for externalization
   */
  public ReplicaPingEvent() { }

  /**
   * Private constructor to prevent external construction.
   *
   * @param gid
   * @param pingRate
   */
  private ReplicaPingEvent(int gid, int pingRate)
  {
    this.gid = gid;
    this.pingRate = pingRate;
    if (pingRate < MIN_PING_RATE)
      throw new IllegalArgumentException("Ping rate must be greater than " + MIN_PING_RATE);
    suspectPeriod = 2*pingRate;
  }

  /**
   * @param gid the group identifier of the group renewing its lease
   * @param groupSize the current size of the group
   */
  public static ReplicaPingEvent getNewPingEvent(int gid, int groupSize)
  {
    AppConfig app = AppConfig.getApplication(gid);
    boolean groupFailureSupport = app.getBooleanParam("Recovery.GroupFailureSupport");
    if (groupFailureSupport) {
      int basePingRate = app.getIntParam("Recovery.RenewalRate", DEFAULT_PING_RATE);
      int pingRate = (int) (Math.pow(basePingRate/1000, groupSize)*1000);
      return new ReplicaPingEvent(gid, pingRate);
    } else {
      return null;
    }
  }

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Access methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * Get the ping rate associated with this object. <p>
   * 
   * Note that this can only be called on the object created by
   * the argument constructor, since the <code>pingRate</code> field
   * is not serialized.
   */
  public int getPingRate()
  {
    return pingRate;
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // ARMEvent interface methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /* (non-Javadoc)
   * @see jgroup.core.arm.ARMEvent#handle()
   */
  public void handle(DistributionScheme distScheme)
  {
    this.distScheme = distScheme;
    synchronized (renewTimers) {
      Timer renewTimer = (Timer) renewTimers.get(gid);
      if (renewTimer != null) {
        renewTimer.cancel();
      }
      renewTimer = new Timer("RenewTimer-" + gid, true);
      renewTimers.put(gid, renewTimer);
      if (log.isDebugEnabled()) {
        log.debug("Lease expires @ " + (System.currentTimeMillis() + suspectPeriod)
            + ", group: " + gid);
      }
      renewTimer.schedule(new RenewTask(), suspectPeriod);
    }
  }


  /* (non-Javadoc)
   * @see jgroup.core.arm.ARMEvent#getGroupId()
   */
  public int getGroupId()
  {
    return gid;
  }


  /* (non-Javadoc)
   * @see jgroup.core.arm.ARMEvent#getObject()
   */
  public Object getObject()
  {
    // Currently we ignore these events.
    return null;
  }


  /**
   * This timer task is run only when a renewal does not occur.
   */
  private class RenewTask
    extends TimerTask
  {

    /* (non-Javadoc)
     * @see java.util.TimerTask#run()
     */
    public void run()
    {
      if (log.isDebugEnabled())
        log.debug("Lease expired @ " + System.currentTimeMillis());
      synchronized (renewTimers) {
        renewTimers.remove(gid);
      }
      distScheme.viewChangeEvent(new ViewImpl(gid));
    }

  } // END RenewTask


  public String toString()
  {
    return "ReplicaPingEvent: gid=" + gid + ", suspect_period=" + suspectPeriod;
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from the Externalizable interface (as required by the ARMEvent interface)
  ////////////////////////////////////////////////////////////////////////////////////////////

  /* (non-Javadoc)
   * @see java.io.Externalizable#readExternal(java.io.ObjectInput)
   */
  public void readExternal(ObjectInput in)
    throws IOException, ClassNotFoundException
  {
    gid = in.readInt();
    suspectPeriod = in.readInt();
  }

  /* (non-Javadoc)
   * @see java.io.Externalizable#writeExternal(java.io.ObjectOutput)
   */
  public void writeExternal(ObjectOutput out)
    throws IOException
  {
    out.writeInt(gid);
    out.writeInt(suspectPeriod);
  }

} // END ReplicaPingEvent
