/*
 * Copyright (c) 1998-2002 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */
 
package jgroup.arm;

import java.io.IOException;
import java.io.ObjectInput;
import java.io.ObjectOutput;
import java.rmi.RemoteException;

import jgroup.core.View;
import jgroup.core.arm.ARMEvent;
import jgroup.core.arm.DistributionScheme;
import jgroup.core.arm.UnknownGroupException;
import jgroup.relacs.types.ViewImpl;

/**
 *  Event to notify the replication manager of a view change that has
 *  occured in object group assigned to the system by the replication
 *  manager.
 *
 *  @author Hein Meling
 *  @since Jgroup 1.2
 */
final class ViewChangeEvent
  implements ARMEvent
{

  private static final long serialVersionUID = 7943725380815810998L;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** The view associated with this event */
  private View view;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructors
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Default constructor for externalization.
   */
  public ViewChangeEvent() { }


  /**
   *  Construct a new <code>ViewChangeEvent</code> instance.
   */
  ViewChangeEvent(View view)
  {
    this.view = view;
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // ARMEvent interface methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  The event handling is performed by the event object itself, at the
   *  (replication manager) receiver side. <p>
   *
   *  The handler method is invoked by the replication manager; it
   *  enables the recovery layer of the group communication stack to
   *  notify the replication manager of view change events that occured
   *  within its group.
   *
   *  @exception RemoteException
   *    Raised if the remote operation failed.
   *  @exception UnknownGroupException
   *    Raised of the group identifier of the view associated with this
   *    event was not known to the replication manager.
   *  @exception NullPointerException
   *    Thrown if this method is invoked at sender side (recovery
   *    layer); that is before view has been marshalled and then
   *    unmarshalled again.  The unmarshal process will initialize the
   *    object so that this method can be invoked.
   */
  public void handle(DistributionScheme distScheme)
    throws Exception
  {
    distScheme.viewChangeEvent(view);
  }


  /* (non-Javadoc)
   * @see jgroup.core.arm.ARMEvent#getGroupId()
   */
  public int getGroupId()
  {
    return view.getGid();
  }


  /* (non-Javadoc)
   * @see jgroup.core.arm.ARMEvent#getObject()
   */
  public Object getObject()
  {
    return view;
  }


  public String toString()
  {
    return "ViewChangeEvent: " + view;
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from the Externalizable interface (as required by the ARMEvent interface)
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Restores the content of this object from the marshalled data
   *  contained in the specified input stream.
   *
   *  @param in
   *    The stream to be read from.
   *  @exception IOException
   *    Thrown if there was problems unmarshalling the input stream.
   *  @exception ClassNotFoundException
   *    Thrown if the local classloader cannot find a class marshalled
   *    in the stream.
   */
  public void readExternal(ObjectInput in)
    throws IOException, ClassNotFoundException
  {
    view = new ViewImpl();
    view.readExternal(in);
  }


  /**
   *  Marshals the content of this object to the specified output
   *  stream.
   * 
   *  @param out
   *    The stream to write this object into.
   *  @exception IOException
   *    Thrown if there was problems unmarshalling the input stream.
   */
  public void writeExternal(ObjectOutput out)
    throws IOException
  {
    view.writeExternal(out);
  }

} // END ViewChangeEvent
