/*
 * Copyright (c) 1998-2002 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */
 
package jgroup.arm.recovery;

import java.util.HashMap;
import java.util.Map;
import java.util.Set;

import jgroup.relacs.config.Host;
import jgroup.relacs.config.HostSet;
import jgroup.relacs.config.Host.ReplicaState;

import org.apache.log4j.Logger;



/**
 *  If the current number of replicas falls below the minimal redundancy
 *  specified for this replica, and the number of available hosts in
 *  this partition is greater than the minimal redundancy then create
 *  new replicas to compensate for the lost redundancy.
 *
 *  @author Hein Meling
 *  @since Jgroup 1.2
 */
public class KeepMinimalInPartition
  extends AbstractRecoveryStrategy
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Logger
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Obtain logger for this class */
  private static final Logger log = Logger.getLogger(KeepMinimalInPartition.class);


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constants
  ////////////////////////////////////////////////////////////////////////////////////////////

  private static final int RETRY_LIMIT = 3;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Map of number of retries for a given host */
  private final Map<Host, Integer> hostRetries = new HashMap<Host, Integer>();

  /** The host on which a replacement replica has been installed (pending join) */
  private Host recoveredHost;

  /** The set of hosts in the current view */
  private HostSet viewHosts;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // RecoveryStrategy interface methods (those not implemented in AbstractRecoveryStrategy)
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * FIXME UPDATE DOCUMENTATION:
   *  The <code>KeepMinimalInPartition</code> implementation simply
   *  assume that the servers do not maintain any state or has some
   *  other means of recovering their state, and tries to relocate
   *  the replicas to a new set of hosts.
   *
   *  @param hosts
   *    The set of hosts on which the application was running,
   *    but has now failed.
   *
   *  @return 
   *    True if all replicas replacement repliacs has been created successfully,
   *    false otherwise.
   */
  public boolean handleFailure(HostSet hosts)
  {
    viewHosts = app.getViewHosts();
    if (recoveredHost != null && !viewHosts.containsHost(recoveredHost)) {
      /*
       * The last recovered host is not yet in the current view; check the
       * status of the replica on that host.
       */
      if (!checkReplicaStatus(recoveredHost)) {
        if (log.isDebugEnabled())
          log.debug("We are pending join from replica on " + recoveredHost);
        // Reschedule the ViewMonitor
        return true;
      }
      recoveredHost = null;
    }
    /*
     * We need to clone the assigned host set since the underlying
     * host set may be modified by the relocation below, which may
     * cause a concurrent modification exception.
     */
    HostSet assignedHosts = (HostSet) hosts.clone();
    if (log.isDebugEnabled())
      log.debug("assignedHosts: " + assignedHosts);
    if (log.isDebugEnabled())
      log.debug("handleFailure(): " + app);
    boolean groupFailed = prepareRecovery();
    boolean recovered = true;
    if (groupFailed && assignedHosts.getAvailHosts().isEmpty()) {
      for (Host host : assignedHosts) {
        recovered &= relocateReplica(host);
      }
    } else {
      recovered = doRecover(assignedHosts);
    }
    return recovered;
  }


  private boolean doRecover(HostSet assignedHosts)
  {
    /*
     * The assignedHosts set may have been updated in a previous run,
     * and if so should contain the new host on which the replica is being
     * installed (in case of recovery).
     */
    HostSet hostsToCheck = new HostSet();
    HostSet recoverHosts = new HostSet();
    hostsToCheck.addHosts(assignedHosts);
    hostsToCheck.removeHosts(viewHosts);
    if (log.isDebugEnabled()) {
      log.debug("    viewHosts: " + viewHosts);
      log.debug("assignedHosts: " + assignedHosts);
      log.debug(" hostsToCheck: " + hostsToCheck);
      log.debug(" recoverHosts: " + recoverHosts);
    }

    for (Host host : hostsToCheck) {
      if (!host.isAvailable()) {
        recoverHosts.addHost(host);
      } else {
        if (checkReplicaStatus(host)) {
          // The replica on the host has not installed a view; recovery needed
          recoverHosts.addHost(host);
        }
      }
    }

    if (log.isDebugEnabled())
      log.debug(" recoverHosts: " + recoverHosts);
    boolean recovered = true;
    if (!recoverHosts.isEmpty()) {
      /*
       * If more than one replica is missing, relocate multiple replicas
       * in this iteration of the recovery algorithm.  Otherwise, we only
       * relocate one replica for each iteration of the recovery algorithm.
       * The recovery algorithm is triggered by the ViewMonitor timeouts.
       * The missing variable is initialized by the prepareRecovery() method,
       * and updated by the relocateReplica() method.
       */
      while (missing > 0 && !recoverHosts.isEmpty()) {
        // We only attempt to recover from one host at a time
        Host failedHost = recoverHosts.removeFirst();
        // This call will actually modify the 'assignedHosts' set
        recovered &= relocateReplica(failedHost);
      }
    }
    return recovered;
  }


  /**
   * Check the status of the replica associated with the outer <code>GroupData</code>
   * object on the given host.
   * 
   * If this method returns true it means that a recovery action is needed for
   * the given host.  If false is returned it means that recovery may be needed,
   * but additional waits could regain the required redundancy levels.
   * 
   * @param host The host on which to check the replica status.
   * @return True is returned if the replica was created and later failed
   *   or is not responding by installing a view after several checks.
   *   False is returned if replica has been created on the given host,
   *   but has not yet installed a view reflecting this.
   */
  private boolean checkReplicaStatus(Host host)
  {
    Set replicas = host.queryReplicas();
    /* The host has not yet joined the group; check if the replica has been created. */
    if (replicas.contains(app.getClassData())) {
      if (host.isJoining(app)) {
        log.debug(app + " is pending to join the group on " + host);
      } else {
        ReplicaState state = host.getState(app);
        log.debug(app + " state: " + state + " on " + host);
      }
      /* The replica has been created, but not installed a view. */
      if (retriesExhausted(host)) {
        log.warn("Removing replica on " + host + "; it has not installed a view in due time");
//        host.removeReplica(app);
        /* This replica has been removed from this host; try another host */
//        return true;
        //FIXME find better solutions
        return false; // ignore this possibility
      } else {
        /* The replica has been created, but not installed a view. */
        log.info(app + " has been created on " + host + " (but not yet installed a view)");
        /* Just reschedule the view monitor */
        return false;
      }
    } else {
      /* The replica does not seem to have been created (or has crashed). */
      if (retriesExhausted(host)) {
        log.warn(app + ": Replica creation failed on " + host);
        /* Try another host */
        return true;
      } else {
        /* The replica was not created (or has crashed), but the host is still available. */
        boolean created = restartReplica(host);
        if (created) {
          log.info(app + " has been restarted on " + host);
          /* Just reschedule the view monitor */
          return false;
        } else {
          log.warn(app + " failed to restart on " + host);
          /* Try another host */
          return true;
        }
      }
    }
  }


  /**
   * Returns true if the given host has exhausted its retry count; False is
   * returned otherwise and the retry count is incremented.
   * 
   * @param host the host whose retry count to query
   */
  private boolean retriesExhausted(Host host)
  {
    Integer retries = hostRetries.get(host);
    if (retries == null)
      retries = 0;
    if (retries < RETRY_LIMIT) {
      hostRetries.put(host, ++retries);
      if (log.isDebugEnabled())
        log.debug("Checking " + app + " on " + host + "; retry: " + retries);
      return false;
    } else {
      resetRetries(host);
      log.warn(app + ": Exhausted the retries to check on " + host);
      return true;
    }
  }

  private void resetRetries(Host host)
  {
    // Reset the retry count for future use of this host
    hostRetries.put(host, 0);
  }


  /* (non-Javadoc)
   * @see jgroup.core.arm.RecoveryStrategy#restartReplica(jgroup.relacs.config.Host)
   */
  public boolean restartReplica(Host host)
  {
    if (log.isDebugEnabled())
      log.debug("restartReplica(): " + app);
    /*
     * Create a single new replica on the same host, if there still
     * are missing replicas to satisfy the requirements of this 
     * recovery strategy.
     */
    boolean recovered = false;
    if (missing > 0) {
      recovered = host.createReplica(app);
      if (recovered) {
        missing--;
        if (log.isDebugEnabled())
          log.debug("Restarted replica " + app.getClassName() + " on " + host);
      }
      return recovered;
    } else {
      /*
       * No missing replicas remains, recovery is consider to be successful
       */
      return true;
    }
  }


  /* (non-Javadoc)
   * @see jgroup.core.arm.RecoveryStrategy#relocateReplica(jgroup.relacs.config.Host)
   */
  public boolean relocateReplica(Host host)
  {
    if (log.isDebugEnabled())
      log.debug("relocateReplica: " + app.getClassName() + " away from " + host);

    /*
     * Create a replacement replica for this application on the
     * newly assigned host, if there are still missing replicas.
     */
    boolean recovered = false;
    if (missing > 0) {
      try {
        recoveredHost = distScheme.reassignReplica(app, host);
        recovered = recoveredHost.createReplica(app);
        if (recovered) {
          missing--;
          if (log.isDebugEnabled())
            log.debug("Reassigned replica to new host: " + recoveredHost);
        }
      } catch (Exception e) {
        // Ignore; recovered is false
      }
      return recovered;
    } else {
      /*
       * No missing replicas remains, recovery is consider to be successful
       */
      return true;
    }
  }

} // END KeepMinimalInPartition
