/*
 * Copyright (c) 1998-2002 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package jgroup.core;

import java.io.File;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.List;

import jgroup.core.registry.RegistryFactory;
import jgroup.core.registry.RegistryLocator;
import jgroup.relacs.config.BootstrapRegistryConfig;
import jgroup.relacs.config.ConfigParser;
import jgroup.relacs.config.DistributedSystemConfig;
import jgroup.util.Network;
import jgroup.util.Util;

import org.apache.log4j.LogManager;
import org.apache.log4j.joran.JoranConfigurator;

/**
 *  The configuration manager class is used initialize the required
 *  Jgroup system configuration data, including logging configuration.
 *  It will read the logging configuration and the Jgroup system
 *  configuration, by parsing the XML files supplied to the JVM using
 *  the following system properties: <p>
 *
 *  <code>jgroup.system.config</code>: XML-based configuration file for
 *  specifying the required system configurations for Jgroup/ARM.
 *  <p>
 *
 *  <code>jgroup.system.services</code>: XML-based configuration file for
 *  specifying the services required for Jgroup/ARM.
 *  <p>
 *
 *  <code>jgroup.system.applications</code>: XML-based configuration file for
 *  specifying the applications used in a Jgroup/ARM system.
 *  <p>
 *
 *  <code>jgroup.log.config</code>: XML-based configuration file for
 *  log4j, allows the user to limit the amount of debug information
 *  being sent to the console; among numerous other configuration
 *  options available through the log4j toolkit.  For details concerning
 *  the format of the configuration file, please refer to the log4j
 *  documentation.  The default for this property: 
 *  <code>config/log4j.xml</code>
 *  <p>
 *
 *  In addition, the following logging related system properties may be
 *  specified by the user: <p>
 * 
 *  <code>jgroup.registry.locator</code>: The registry locator class to be used
 *  by Jgroup.  The default is <code>jgroup.relacs.registry.RelacsRegistryLocator</code>
 *  <p>
 * 
 *  <code>jgroup.log.app</code>: The application name that will be appended
 *  to the log filename.  Note that, when using the ARM framework (the
 *  <code>ExecDaemon</code>) to create replicas, the log.app name for the
 *  replica will be given the name of the class (exclusive the package name).
 *  <p>
 *
 *  <code>jgroup.log.dir</code>: The directory into which log files will be
 *  written.  If not specified, the default is used (<code>./log</code>)
 *  <p>
 *
 *  <code>jgroup.log.msgcontent</code>: True/False.  Used to turn on/off
 *  the logging of message content at both the mss and daemon level.
 *  <p>
 * 
 *  The following system properties are initialized at runtime, allowing the
 *  log4j toolkit to use these properties in its XML based log file for
 *  constructing unique files.  Passing these properties to the Java runtime
 *  will have no effect as they will be overwritten. <p>
 *
 *  <code>jgroup.log.machine</code>: The local host of this replica.
 *  <p>
 * 
 *  <code>jgroup.log.time</code>: The time when this replica was initialized. 
 *  <p>
 * 
 *  <code>jgroup.log.date</code>: The date when this replica was initialized.
 *  <p>
 *
 *  @author Hein Meling
 *  @since Jgroup 1.2
 */
public final class ConfigManager
{

  public static final String DEFAULT_REGISTRY_LOCATOR    = "jgroup.relacs.registry.RelacsRegistryLocator";
  public static final String DEFAULT_REPLICATION_MANAGER = "jgroup.arm.ReplicaManagerImpl";

  private static final String DEFAULT_SYSTEM_CONFIG       = "file:target/classes/siteconfig/lx-config.xml";
  private static final String DEFAULT_SYSTEM_SERVICES     = "file:target/classes/config/services.xml";
  private static final String DEFAULT_SYSTEM_APPLICATIONS = "file:target/classes/config/applications.xml";
  private static final String DEFAULT_LOGGING_CONFIG      = "file:target/classes/logging/log4j.xml";
  private static final String DEFAULT_LOG_DIR             = File.separator + "tmp" + File.separator + "jgroup-logs";


  /** 
   *  Indicates if message content should be logged.  To change this
   *  use the system property: <code>jgroup.log.msgcontent</code>.
   */
  public static boolean logMsgContent = false;

  /**
   *  Variable used to avoid duplicate reading/parsing of the
   *  configuration files.  That is to make the init methods idempotent.
   */
  private static boolean initialized = false;

  /**
   * Variable used to avoid duplicate reading/initialization of the
   * logging configuration.  That is to make the configureLogging method
   * idempotent.
   */
  private static boolean logInitialized = false;

  /** The port used by the bootstrap registry */
  private static int bootstrapPort;

  /** The distributed system configuration object */
  private static DistributedSystemConfig dsc;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Public static methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Invoked to initialize various system configuration variables
   *  required by Jgroup/ARM.  This includes reading and parsing the
   *  logging configuration and the distributed system configuration.
   *
   *  The method is idempotent (it can be invoked many times without
   *  changing its behavior), meaning that no configuration files will
   *  be read and parsed several times.
   *
   *  @exception ConfigurationException
   *    Raised if there was an error parsing the configuration files,
   *    or the system properties required to specify the configuration
   *    file names were unset.
   */
  public static void init()
    throws ConfigurationException
  {
    configureLogging();
    String urlConfig = System.getProperty("jgroup.system.config", DEFAULT_SYSTEM_CONFIG);
    init(urlConfig);
  }


  /**
   * Helper method to parse a specified configuration file instead of the
   * system configured default.  This is useful for the Java based experiment
   * framework.
   *
   * @param urlConfig
   * @throws ConfigurationException
   */
  public static void init(String urlConfig)
    throws ConfigurationException
  {
    if (initialized)
      return;
    initialized = true;

    /*
     * Parse the given URL (should be either a distributed system config
     * or an experiment config; the latter will invoke the parser again
     * to obtain the distributed system config.)
     */
    ConfigParser.parse(urlConfig);
    /*
     * Parse the service and application configuration. Multiple services and applications 
     * configuration files can be specified and separated by comma.
     */
    urlConfig = System.getProperty("jgroup.system.services", DEFAULT_SYSTEM_SERVICES);
    String urlConfigs[] = urlConfig.split(",");
    for (int i=0; i < urlConfigs.length; i++) {
      ConfigParser.parse(urlConfigs[i]);
    }
    urlConfig = System.getProperty("jgroup.system.applications", DEFAULT_SYSTEM_APPLICATIONS);
    urlConfigs = urlConfig.split(",");
    for (int i=0; i < urlConfigs.length; i++) {
      ConfigParser.parse(urlConfigs[i]);      
    }
    setRegistryLocator();
    BootstrapRegistryConfig brc = 
      (BootstrapRegistryConfig) ConfigParser.getConfig(BootstrapRegistryConfig.class);
    bootstrapPort = brc.getPort();
    dsc = (DistributedSystemConfig) ConfigParser.getConfig(DistributedSystemConfig.class);
  }


  /**
   * Method to set up log directory, and initialize the Log4j configuration.
   * This can be used externally when custom configuration is required, such
   * as the experiment framework.
   *
   * @throws ConfigurationException
   *   Thrown if the specified log directory could not be created, or
   *   if the log4j configuration URL was malformed.
   */
  public static void configureLogging()
    throws ConfigurationException
  {
    if (logInitialized)
      return;
    logInitialized = true;

    /*
     * Get the directory in which log files should be stored;
     * if the directory does not exist, it is created.
     */
    String logDir = System.getProperty("jgroup.log.dir", DEFAULT_LOG_DIR);
    File dir = new File(logDir);
    if (!dir.exists()) {
      /* Directory does not exist; create it. */
      if (!dir.mkdir())
        throw new ConfigurationException("Could not create the specified log directory: " + logDir);
    }
    /*
     * Check if the application name is defined; if not 'noapp' is
     * set to this property so that log4j use that when constructing
     * the logging filenames.
     */
    String logApp = System.getProperty("jgroup.log.app");
    if (logApp == null || logApp.length() == 0) {
      System.setProperty("jgroup.log.app", "noapp");
    }
    /*
     * Obtain the property used to set if the message content should be logged.
     */
    logMsgContent = Boolean.getBoolean("jgroup.log.msgcontent");
    /* 
     * Set the values for the local machine, date and time when the Jgroup 
     * related application was initialized.  These values can be used in the
     * log4j.xml file to avoid overwriting the log files of concurrently
     * running applications on the same host.  The can be used along with the
     * app property defined above.
     */
    System.setProperty("jgroup.log.machine", Network.getLocalMachineName());
    System.setProperty("jgroup.log.time", Util.getTimeString());
    System.setProperty("jgroup.log.date", Util.getDateString());
    /*
     * Initialize and parse the log4j configuration file.
     */
    String logConfig = System.getProperty("jgroup.log.config", DEFAULT_LOGGING_CONFIG);
    // JoranConfigurator requires log4j version 1.3
    JoranConfigurator jc = new JoranConfigurator();
    try {
      jc.doConfigure(new URL(logConfig), LogManager.getLoggerRepository());
    } catch (MalformedURLException e) {
      throw new ConfigurationException("Bad URL provided for log4j configuration: " + logConfig, e);
    }
    List errorList = jc.getErrorList();
    for (int i = 0; i < errorList.size(); i++) {
      System.out.println(errorList.get(i));
    }
  }

  /**
   * Set the registry to use based on a system property, 
   * jgroup.registry. Default is DependableRegistry.
   * 
   * @exception ConfigurationException
   *   Raised if unable to create the registry.
   */
  private static void setRegistryLocator()
    throws ConfigurationException
  {
    String locatorName = System.getProperty("jgroup.registry", DEFAULT_REGISTRY_LOCATOR);
    try {
      Class clazz = Class.forName(locatorName);
      RegistryLocator locator = (RegistryLocator) clazz.newInstance();
      RegistryFactory.setLocator(locator);
    } catch (ClassNotFoundException e) {
      throw new ConfigurationException("Registry locator class not found: " + locatorName);
    } catch (InstantiationException e) {
      throw new ConfigurationException("Could not instanstiate: " + locatorName);
    } catch (IllegalAccessException e) {
      throw new ConfigurationException("Illegal access to : " + locatorName);
    }
  }


  /**
   *  Retrieve an instance of the given configuration class, that is
   *  containing the configuration information.  For example, the
   *  configuration data object associated with the XML tag
   *  <code>Transport</code>, can be retreived using the class literal
   *  <code>TransportConfig.class</code>.
   *
   *  @param cl
   *    The class literal of the configuration object to retreive.
   *  @return 
   *    An instance of the configuration object retreived.
   *
   *  @exception NullPointerException
   *    If there is no such configuration object in the internal
   *    configuration map.  This may occur for two reasons; (i) there
   *    is no such configuration tag specified in the configuration file,
   *    or (ii) the specified class literal is a non-existant class.
   *  @exception IllegalStateException
   *    If the init method was not invoked prior to invoking this method.
   */
  public static Object getConfig(Class cl)
  {
    return ConfigParser.getConfig(cl);
  }


  /**
   *  Returns the <code>DistributedSystemConfig</code> object.
   */
  public static DistributedSystemConfig getDistributedSystem()
  {
    return dsc;
  }


  /**
   *  Returns the port for the bootstrap registry.
   */
  public static int getBootstrapPort()
  {
    return bootstrapPort;
  }

} // END ConfigManager
