/*
 * Copyright (c) 1998-2002 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */
 
package jgroup.core;

import java.lang.reflect.Proxy;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import jgroup.core.Layer.FinalizeLayer;
import jgroup.relacs.config.AppConfig;
import jgroup.relacs.config.DistributedSystemConfig;
import jgroup.relacs.config.LayerConfig;
import jgroup.relacs.config.ServiceConfig;
import jgroup.relacs.rmi.IntGroupHandler;
import jgroup.util.Network;

import org.apache.log4j.Logger;

/**
 *  The <code>GroupManager</code> class is used to configure the Jgroup
 *  system.  In order to make use of the Jgroup facilities, a group
 *  manager must be obtained through the invocation of static factory
 *  method <code>getGroupManager</code>.  The main parameter to this
 *  method is the server object that implements a number of listener
 *  interfaces.  That is, the server object is a listener for various
 *  services.  A group manager is constructed by collecting and
 *  configuring the services needed by the server object.  In order for
 *  a server object to gain access to one of the provided services, the
 *  method <code>getService</code> can be invoked on the group manager,
 *  by specifying the service interface class or the shorthand service
 *  name (as given in the <code>services.xml</code> file) of the desired
 *  service.
 *
 *  @author Hein Meling
 *  @author Alberto Montresor
 *  @since Jgroup 0.8 (revised Jgroup 1.2)
 */
public final class GroupManager
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Logger
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Obtain logger for this class */
  private static final Logger log = Logger.getLogger(GroupManager.class);


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Mapping from classes to their corresponding service objects */
  private Map map;

  /** The group identifier for this group manager instance */
  private int groupId;

  /** The set of local group managers (in this JVM) */
  private static final Map<Integer, GroupManager> groupManagers =
    new HashMap<Integer, GroupManager>(5);


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructors
  ////////////////////////////////////////////////////////////////////////////////////////////

  private GroupManager(Map map, int groupId)
  {
    this.map = map;
    this.groupId = groupId;
    groupManagers.put(groupId, this);
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Service methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Returns an handler for a Jgroup service (e.g., the membership
   *  service) specified by the interface class <code>cl</code>.
   *  Services provided by this instance of the group manager, must have
   *  been initialized when invoking the static factory method
   *  <code>getGroupManager</code>.  That is, the service must have been
   *  specified either in the <code>LayerStack</code> entry of the
   *  corresponding application in the <code>application.xml</code>
   *  file, or through the <code>serviceSet</code> array provided to the
   *  <code>getGroupManager</code> method.
   *
   *  @param cl
   *    Class specifying the interface of the required service.
   *  @exception ClassCastException
   *    If the generated group manager is unable to provide the requested
   *    service class.
   */
  public Object getService(Class cl)
  {
    try {
      return getService((String) map.get(cl));
    } catch (ClassCastException e) {
      throw new ClassCastException("Class " + cl.getName() + " not found");
    }
  }


  /**
   *  Returns an handler for a Jgroup service (e.g., the membership
   *  service) specified by the shorthand service name, as specified in
   *  the <code>services.xml</code> file.  Services provided by this
   *  instance of the group manager, must have been initialized when
   *  invoking the static factory method <code>getGroupManager</code>.
   *  That is, the service must have been specified either in the
   *  <code>LayerStack</code> entry of the corresponding application in
   *  the <code>application.xml</code> file, or through the
   *  <code>serviceSet</code> array provided to the
   *  <code>getGroupManager</code> method.
   *
   *  @param serviceName
   *    Shorthand name for the required service.
   *  @exception ClassCastException
   *    If the generated group manager is unable to provide the requested
   *    service class.
   */
  public Object getService(String serviceName)
  {
    Object service = map.get(serviceName);
    if (service == null)
      throw new ClassCastException("Service " + serviceName + " not found");
    return service;
  }

  /**
   *  Returns the group identifier for this group manager instance.
   */
  public int getGroupId()
  {
    return groupId;
  }

  /**
   *  Returns the local member identifier associated with this group manager.
   */
  public MemberId getLocalMember()
  {
    MembershipService pgms = (MembershipService) getService("PGMS");
    return pgms.getMyIdentifier();
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Static methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  public static GroupManager getGroupManager(int gid)
  {
    return groupManagers.get(gid);
  }


  public static Object getService(int gid, Class service)
  {
    GroupManager gm = groupManagers.get(gid);
    if (gm == null)
      return null;
    else
      return gm.getService(service);
  }


  public static boolean contains(int gid)
  {
    return groupManagers.containsKey(gid);
  }


  public static boolean isEmpty()
  {
    return groupManagers.isEmpty();
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Static factory methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Returns a group manager object through which handlers, for each of
   *  the services associated with the listener interfaces implemented
   *  by the specified object, can be accessed.
   *
   *  @param listener
   *    Object implementing the listeners interfaces associated with the
   *    the requested services.
   *  @exception JgroupException
   *    Raised if some configuration error occurs.
   */
  public static GroupManager getGroupManager(Object listener)
    throws JgroupException
  {
    if (listener == null)
      throw new JgroupException("No listener specified for the group manager");
    Class serverClass = listener.getClass();

    ConfigManager.init();

    /* Get application info object for this server (group manager listener) */
    AppConfig app = AppConfig.getApplication(serverClass);
    if (app == null)
      throw new ConfigurationException(serverClass.getName()
        + " is not available in the application configuration.");

    /* Get the set of services (layers) used by this application. */
    String[] serviceSet = app.getServiceSet();
    int groupId = app.getGroupId();
    return getGroupManager(groupId, serviceSet, listener);
  }


  /**
   *  Returns a group manager object through which handlers, for each of
   *  the services associated with the listener interfaces implemented
   *  by the specified object, can be accessed.
   *
   *  @param serviceSet
   *    An array of <code>String</code>s, each representing a service or
   *    layer that should be incorporated with the returned GroupManager.
   *  @param listener
   *    Object implementing the listeners interfaces associated with the
   *    the requested services.
   *  @exception JgroupException
   *    Raised if some configuration error occurs.
   */
  @SuppressWarnings("unchecked")
  public static GroupManager getGroupManager(int groupId, String[] serviceSet, Object listener)
    throws JgroupException
  {
    if (listener == null)
      throw new JgroupException("No listener specified for the group manager");
    Class serverClass = listener.getClass();
    if (log.isDebugEnabled())
      log.debug(serverClass.getName() + ": wants a group manager");

    ConfigManager.init();
    DistributedSystemConfig dsc = ConfigManager.getDistributedSystem();
    if (!dsc.containsLocalHost())
      throw new JgroupException("Distributed system configuration does not contain local host: "
        + Network.getLocalHostName() + "\nThis is required for servers, but not for clients.");

    /* This map should be immutable from outside the GroupManager class. */
    Map layerMap = new HashMap(serviceSet.length);

    /* Construct the group manager object and add its reference to the layerMap */
    GroupManager gm = new GroupManager(layerMap, groupId);
    layerMap.put("GM", gm);

    /*
     * Initialize the layers.
     */
    LayerConfig.initializeLayers();

    /*
     * Create a layer object for each service specified in the
     * <code>LayerStack<code> of the given application; that is the
     * layers to be used by the application.  The layers are placed in a
     * map, associated with the service that they provide as a key.
     */
    List<Class> internalIfaces = new ArrayList<Class>();
    for (int layerNo = 0; layerNo < serviceSet.length; layerNo++) {
      /*
       * Create an instance of the next layer obtained from the service
       * set.  For layers that share implementation class the
       * getLayerObject() method must ensure to return the same layer
       * instance for all layers.  Furthermore, note that we need to pass
       * the layerMap in order to be able to check dependencies between
       * layers, and also the internalIfaces list, so as to be able to
       * construct a IGMI proxy for layers that use IGMI.
       */
      LayerConfig.constructLayerObject(serviceSet[layerNo], layerMap, internalIfaces);
    }

    /*
     * Iterate over the interfaces implemented by the server, and add it
     * as a listener for all interfacs that are listener interfaces for
     * services defined in the XML file.
     */
    Class serverInterfaces[] = serverClass.getInterfaces();
    for (int i = 0; i < serverInterfaces.length; i++) {

      /*
       * Find the internal interfaces implemented by this server.
       */
      if (InternalGMIListener.class.isAssignableFrom(serverInterfaces[i])) {
        internalIfaces.add(serverInterfaces[i]);
        /*
         * Add also the server interface for this server to the layer
         * map, so that the interfaces assignable from a layer
         * interface can be reached.  This will allow using
         * getService(jgroup.test.hello.InternalHello.class).  The
         * actual layer implementing the IGMI service will be
         * replaced by a proxy for the layer (see below).
         */
        layerMap.put(serverInterfaces[i], "IGMI");
      }
    }

    /*
     * Iterate over the services to obtain the service name and its
     * corresponding listener interface to be matched against the
     * interfaces implemented by the server.
     */
    for (int layerNo = 0; layerNo < serviceSet.length; layerNo++) {
      boolean foundIface = false;
      ServiceConfig serviceConfig = ServiceConfig.getService(serviceSet[layerNo]);
      Class listenerIface = serviceConfig.getListener();
      if (listenerIface == null)
        foundIface = true;

      for (int i = 0; !foundIface && i < serverInterfaces.length; i++) {
        /*
         * Check if this particular service's listener interface is
         * implemented by the server.  If so we invoke addListener on
         * that service's layer to allow this particular server to be a
         * listener for its events.
         */
        if (listenerIface.isAssignableFrom(serverInterfaces[i])) {
          Layer layer = (Layer) layerMap.get(serviceSet[layerNo]);
          if (layer == null)
            throw new ConfigurationException("The layer implementing " + serviceSet[layerNo]
              + " is not be initialized. \nThis layer is required by the " + serverInterfaces[i]
              + " implemented by the server class: " + serverClass.getName());
          layer.addListener(listener);
          if (log.isDebugEnabled()) {
            log.debug(serverClass.getName() + ": listen to events from " + serviceSet[layerNo]);
          }
          foundIface = true;
        }
      }

      if (!foundIface && serviceConfig.isMarker()) {
        /*
         * If the listener interface of the current service is a marker
         * interface that is not implemented by the server, we still
         * need to add the server as a listener for this layer.
         */
        Layer layer = (Layer) layerMap.get(serviceSet[layerNo]);
        layer.addListener(listener);
        if (log.isDebugEnabled()) {
          log.debug(serverClass.getName() + ": registered with " + serviceSet[layerNo]);
        }
      }
    }

    /*
     * For applications that require internal GMI, we replace the
     * IntGMILayer instances in the layer map with the proxy instance
     * that is required for use by the server or layer.
     */
    if (internalIfaces.size() > 0) {

      /*
       * Get the interfaces to be implemented by the proxy, that is those
       * found to implement the InternalGMIService interface and that
       * interface itself.
       */
      internalIfaces.add(InternalGMIService.class);
      Class[] interfaces = new Class[internalIfaces.size()];
      internalIfaces.toArray(interfaces);

      /* Get the invocation handler IntGroupHandler (also known as IntGMILayer) */
      IntGroupHandler handler = (IntGroupHandler) layerMap.get("IGMI");
      if (handler == null)
        throw new ConfigurationException(serverClass.getName()
          + " requires the IGMI layer, but it is not specified for this application.");

      /* Get the class loader for the server */
      ClassLoader classLoader = serverClass.getClassLoader();

      /* Now we can obtain the proxy object */
      Object proxy = Proxy.newProxyInstance(classLoader, interfaces, handler);

      /*
       * Replace the IntGMILayer instance with the proxy instance in the
       * layer map.
       */
      layerMap.put("IGMI", proxy);
    }

    /*
     * For each of the layers in the map, check if it implements the
     * FinalizeLayer interface.  For those that do, invoke the complete()
     * method.
     * 
     * Herein, we provide special handling for the MembershipService, since
     * the MembershipLayer may perform the join() operation automatically,
     * and this should be done last.  This is since otherwise, we often
     * run into bootstrap problems for a number of references that have not
     * yet been allocated when used in the viewChange() method that is invoked
     * during the join() operation.
     */
    for (Iterator iter = layerMap.values().iterator(); iter.hasNext();) {
      Object layer = iter.next();
      if (layer instanceof FinalizeLayer && !(layer instanceof MembershipService)) {
        ((FinalizeLayer) layer).complete(listener);
      }
    }
    /* Finalize the MembershipService (PGMS) layer construction (autojoin) */
    FinalizeLayer pgmsLayer = (FinalizeLayer) layerMap.get("PGMS");
    pgmsLayer.complete(listener);

    return gm;
  }

} // END GroupManager
