/*
 * Copyright (c) 1998-2002 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package jgroup.core;

import java.io.Externalizable;

import jgroup.relacs.types.LocalId;
import net.jini.jeri.Endpoint;

/**
 *  A class implementing this interface uniquely identify a member
 *  object in a group. <p>
 *
 *  A member identifier is composed by three parts:
 *  <UL>
 *  <LI> an IP address, uniquely identifying the machine hosting the
 *  member; </LI>
 *  <LI> the incarnation time of the Jgroup subsystem hosting the
 *  member, i.e. the time at which the Jgroup subsystem was created;
 *  </LI>
 *  <LI> a member counter, which is used to distinguish multiple members
 *  running in the same Java virtual machine. </LI>
 *  </UL>
 *
 *  When created, a Jgroup subsytem takes control over some UDP
 *  communication port.  Given the UDP specification, it is impossible
 *  that two Jgroup subsystem share the same port in the same machine.
 *  In this way, it is possible to discover if a remote Jgroup subsystem
 *  has crashed and then recovered; in this case, their incarnation
 *  times differ (the more recent is greater than the older).
 *
 *  When a <code>MemberId</code> is created, its IP address and its
 *  incarnation time are set equal to each member id created in that
 *  Java virtual machine.  So, the identifiers of members hosted in the
 *  same Java virtual machine differs only for the member counter. <p>
 *
 *  Member ids may be compared in the following ways:
 *  <UL>
 *  <LI> two <code>MembersId</code>s are equal (method <tt>equals</tt>)
 *  if and only if they have the same IP address, the same incarnation
 *  time and the same member counter; </LI>
 *  <LI> two <code>MembersId</code>s are <i>neighbours</i> (method
 *  <tt>isNeighbour</tt>) if and only if they have the same IP address
 *  and the same incarnation time; i.e., if they are hosted in the same
 *  virtual machine;</LI>
 *  <LI> a <code>MembersId</code> is <i>newer</i> than another
 *  <code>MembersId</code> (method <tt>isNewer</tt>) if and only if
 *  their IP addresses are the same and the incarnation time of the
 *  former id is greater than the incarnation time of the latter id; in
 *  other words, if the Java virtual machine hosting the latter member
 *  id has crashed (freeing the control over the UDP port) and then
 *  recovered. </LI>
 *  </UL>
 *
 *  @author Alberto Montresor
 *  @author Hein Meling
 *  @author Marcin Solarski
 *  @since Jgroup 0.9
 */
public interface MemberId
  extends Externalizable, Comparable<MemberId>
{

  /**
   *  Returns the EndPoint (IP, port) of the java virtual machine
   *  hosting this member.
   */
  public EndPoint getEndPoint();

  /**
   * Returns the TcpEndpoint of this member.
   */
  public Endpoint getTcpEndpoint();

  /**
   *  Returns a LocalId object used to distinguish this particular
   *  member from other members of the same group hosted in
   *  the same Java virtual machine.
   */
  public LocalId getLocalId();

  /**
   *  Returns the canonical hostname associated with the
   *  JVM on which the member is running.
   */
  public String getCanonicalHostName();

  /**
   *  Returns the server port associated with this member host.
   *  The server port is computed deterministically, based on the
   *  the host port number and the member number.
   */
  public int getServerPort();

  /**
   * Returns true if and only if the IP addresses of this member id and
   * the specified id are the same, and the incarnation time of the former
   * is greater than the incarnation time of the latter; in other words,
   * if the Java virtual machine hosting the specified member id has crashed
   * and then recovered.
   */
  public boolean isNewer(MemberId id);

  /**
   * Returns true if and only if this member id and the specified member id
   * have the same IP address and the same incarnation time; i.e., if they
   * are hosted in the same virtual machine.
   */
  public boolean isNeighbour(MemberId id);
  
} // END MemberId
