/*
 * Copyright (c) 1998-2002 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package jgroup.core;

import java.io.Externalizable;

/**
 *  The Partitionable Group Membership Service contained in Jgroup
 *  tracks changes in the group's composition and installs them as views
 *  at member objects.  Installated views are an abstraction of the
 *  current environment with respect to failures and recoveries.  Member
 *  objects can obtain information about the current view through the
 *  <code>View</code> interface.  Each view is uniquely identified by a
 *  view identifier, that can be obtained by invoking method
 *  <code>getVid</code>.  The composition of the view may be obtained by
 *  invoking method <code>getMembers</code>.
 *
 *  @author Alberto Montresor
 *  @author Hein Meling
 *  @since Jgroup 0.1
 */
public interface View
  extends Externalizable, Cloneable
{

  /**
   *  Returns the identifier of the group for which this view has been
   *  generated.
   */
  public int getGid();

  /**
   *  Returns the identifier of this view.
   */
  public long getVid();

  /**
   *  Returns the number of views (or partitions) merging into this view.
   */
  public int mergingViews();

  /**
   *  Returns an array of <code>MemberId</code>s identifying the members
   *  included in this view.
   */
  public MemberId[] getMembers();

  /**
   *  Return the <code>MemberId</code> of the group leader (member in
   *  position 0).
   */
  public MemberId getLeader();

  /**
   *  Returns true if the given member is the group leader.
   */
  public boolean isLeader(MemberId member);

  /**
   *  Returns an array of <code>MemberId</code>s that are common to this
   *  view and the provided member array.
   */
  public MemberId[] commonMembers(MemberId[] members);

  /**
   *  Returns an array of <code>MemberId</code>s that are common to this
   *  view and the provided view.
   */
  public MemberId[] commonMembers(View view);

  /**
   *  Returns an array of <code>MemberId</code>s that are new to this
   *  view; that is they were not in the previous view.
   */
  public MemberId[] newMembers(View previousView);

  /**
   *  Returns an array of <code>MemberId</code>s that are not longer
   *  part of this view; that is they were members of the previous view
   *  but are no longer members of the this view.
   */
  public MemberId[] oldMembers(View previousView);

  /**
   *  Returns true if the given member is contained in the current view.
   *  Otherwise, false is returned.
   */
  public boolean contains(MemberId memberId);

  /**
   *  Returns true if this view contains all members of the provided view.
   *  False is returned otherwise. <p>
   *
   *  This is useful for checking if a new view is a contraction of the
   *  previous view, or if there are new members.
   */
  public boolean containsAll(View view);

  /**
   * Returns true if this view has exactly the same members are the provided
   * view.  False is returned otherwise. <p>
   * 
   * Note that this is different from using <code>Object.equals()</code>, since
   * this method will not check that the view identifiers are the same.
   */
  public boolean hasSameMembers(View prevView);

  /**
   *  Returns the number of members in this view.
   */
  public int size();

  /**
   * Returns the position of the given member in this view.  If the member
   * is not in this view, a negative value is returned.
   */
  public int getMemberPosition(MemberId id);

  /**
   *  Returns true if the specified member is the member at the given
   *  position; otherwise false is returned.
   *
   *  @param pos
   *    The position in the view to check for equality with the given
   *    member.  A negative position value is considered to indicate
   *    a position from the end of the view array, while a positive
   *    value indicate a position from the beginning of the view
   *    array.  A position value outside the range of the view, will
   *    always return false.
   *  @param member
   *    The member whose position to check.
   *  @return 
   *    True if the given member is at the specified position in this
   *    view; false otherwise.
   */
  public boolean memberHasPosition(int pos, MemberId member);
  
  public Object clone();

} // END View
