/*
 * Copyright (c) 1998-2002 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */
 
package jgroup.core.arm;

import jgroup.relacs.config.AppConfig;
import jgroup.relacs.config.Host;
import jgroup.relacs.config.HostSet;


/**
 *  The <code>RecoveryStrategy</code> interface provides methods that
 *  must be implemented by a particular recovery service.  These methods
 *  will be invoked in response to a corresponding failure event, that
 *  the method intend to rectify. <p>
 *
 *  All implementations of this interface that are to be used through
 *  the <code>ReplicationManager</code> must implement a no-argument
 *  constructor.  Initialization of a particular recovery strategy
 *  should occur in the {@link #initialize(DistributionScheme, AppConfig) initialize}
 *  method.
 *
 *  @author Hein Meling
 *  @since Jgroup 1.2
 *  @see jgroup.arm.recovery.AbstractRecoveryStrategy
 *  @see jgroup.arm.recovery.KeepMinimalInPartition
 *  @see jgroup.arm.recovery.KeepInitialInPartition
 */
public interface RecoveryStrategy
{

  /**
   *  Initialize the recovery strategy for the given application.
   */
  public void initialize(DistributionScheme distScheme, AppConfig app);


  /**
   *  Returns true if the associated application needs a recovery action,
   *  otherwise false is returned.
   */
  public boolean needsRecovery();


  /**
   *  Invoked to initialize the recovery strategy implementation
   *  for the provided application.  The method returns true if
   *  the group has failed.
   */
  public boolean prepareRecovery();


  /**
   *  Handle multiple simulatenous failures; this may either be due to
   *  a network partitioning scenario or if several members of the group
   *  failed at the same time.
   *  
   *  FIXME UPDATE DOCUMENTATION:
   *  Invoked in the event that an entire group fails.  This particular
   *  type of failure can be characterized as a design failure in most
   *  cases.  This can be used by special recovery strategies that are
   *  capable of recovery with different versions of a replica.  For
   *  instance if a replica group running version X of the application
   *  software, may be replaced by version Y of the same software, in
   *  case we get a group failure.
   *
   *  @param hosts
   *    The set of hosts on which the application were running.
   *
   *  @return 
   *    True if recovery was successful; false otherwise.
   */
  public boolean handleFailure(HostSet hosts);


  /**
   *  Restart the given replica on the given host.  This method differes
   *  from the one below in that it will try to instatiate the replica
   *  on the same host on which the failed replica was running.  That is
   *  it will be invoked only when the execution service is still believed
   *  to be available on that host.
   *
   *  @param host
   *    The host on which the replica was running.
   *
   *  @return 
   *    True if recovery was successful; false otherwise.
   */
  public boolean restartReplica(Host host);


  /**
   *  Relocate the application replica that were running on the given
   *  host.  The new location should be obtain through the
   *  <code>DistributionScheme</code> provided by the replication
   *  manager.
   *
   *  @param host
   *    The host on which the replica were running.
   *
   *  @return 
   *    True if recovery was successful; false otherwise.
   */
  public boolean relocateReplica(Host host);

} // END RecoveryStrategy
