/*
 * Copyright (c) 1998-2002 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package jgroup.core.registry;

import java.net.UnknownHostException;
import java.rmi.AccessException;
import java.rmi.AlreadyBoundException;
import java.rmi.NotBoundException;
import java.rmi.Remote;
import java.rmi.RemoteException;
import java.rmi.registry.LocateRegistry;
import java.rmi.registry.Registry;
import java.rmi.server.RMIClientSocketFactory;
import java.rmi.server.RMIServerSocketFactory;
import java.util.Collections;
import java.util.HashMap;
import java.util.Map;

import jgroup.core.ConfigManager;
import jgroup.relacs.simulator.SocketStatusImpl;
import jgroup.util.Abort;

import org.apache.log4j.Logger;

/**
 *  Provides static methods for binding bootstrap references to a local
 *  RMI registry.  It will create local registry if no such registry
 *  exists and bind the given name-remote object pair.
 *
 *  Currently, used by ExecDaemon, RegistryImpl (DependableRegistry),
 *  the Jgroup Daemon and the partition simulator to bootstraping their
 *  references.
 *
 *  @author Hein Meling
 *  @since Jgroup 1.2
 */
public class BootstrapRegistry
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Logger
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Obtain logger for this class */
  private static final Logger log = Logger.getLogger(BootstrapRegistry.class);

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Static fields and constants
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Bootstrap registry test string */
  private static final String BOOTREGISTRY_TEST = "BOOTSTRAP_REGISTRY_TEST";

  /** Mapping: host -> registry stub */
  private static final Map<String,Registry> registryMap =
    Collections.synchronizedMap(new HashMap<String,Registry>());

  /** The port used by the bootstrap registry */
  private static int port;

  /** The registry stub for the local bootstrap registry*/
  private static Registry reg;

  /**
   * An object used to obtain information about link connectivity when
   * the partition simulator has been activated.
   */
  private static SocketStatusImpl status = null;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Static block
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Initialize the bootstrap registry. */
  static {
    port = ConfigManager.getBootstrapPort();
    if (port < 1024)
      Abort.exit("Illegal bootstrap port: " + port);

    /* Check if the partition simulator should be activated. */
    if (Boolean.getBoolean("jgroup.simulator")) {
      status = SocketStatusImpl.instance();
    }
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Public static methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Update the bootstrap registry reference.  This should be used in the
   *  case when the JVM hosting the bootstrap registry has crashed, and
   *  the bootstrap registry should be recreated within a new JVM.
   */
  public static void refreshRegistryStub()
    throws RemoteException
  {
    reg = getRegistry();
  }

  /**
   * Update the bootstrap registry references to the remote registries.
   */
  public static void refreshRemoteRegistries()
    throws RemoteException
  {
    for (String hostname : registryMap.keySet()) {
      try {
        getRemoteRegistry(hostname, null, false);
      } catch (UnknownHostException e) {
        log.error("Unknown host: " + hostname, e);
      }
    }
  }

  /**
   *  Returns an array of the names bound in the bootstrap registry.
   *  The array will contain a snapshot of the names bound in this
   *  registry at the time of the given invocation of this method.
   *
   *  @return an array of the names bound in this registry 
   *
   *  @exception AccessException 
   *    if this operation is not permitted (if originating from a
   *    non-local host, for example)
   *  @exception RemoteException 
   *    if registry could not be contacted
   */
  public static String[] list() 
    throws AccessException, RemoteException
  {
    if (reg == null)
      reg = getRegistry();
    return reg.list();
  }


  /**
   *  Obtains a remote object reference for the specified object name,
   *  from the local bootstrap registry.
   *
   *  @param name  
   *    Name of remote object to be retreived. 
   * 
   *  @exception AccessException 
   *    if this operation is not permitted (if originating from a
   *    non-local host, for example)
   *  @exception RemoteException 
   *    if registry could not be contacted
   *  @exception NotBoundException 
   *    if name is not bound in the registry
   */
  public static Remote lookup(String name)
    throws AccessException, RemoteException, NotBoundException
  {
    if (reg == null)
      reg = getRegistry();
    return reg.lookup(name);
  }


  /**
   *  Obtains a remote object reference for the specified object name,
   *  from the bootstrap registry located on the given hostname.
   *
   *  @param hostname  
   *    name of remote host from which the object reference is obtained
   *  @param name  
   *    name of remote object to be retreived
   * 
   *  @exception AccessException 
   *    if this operation is not permitted (if originating from a
   *    non-local host, for example)
   *  @exception RemoteException 
   *    if registry could not be contacted
   *  @exception NotBoundException 
   *    if name is not bound in the registry
   */
  public static Remote lookup(String hostname, String name)
    throws AccessException, UnknownHostException, RemoteException, NotBoundException
  {
    Registry remoteReg = getRemoteRegistry(hostname, null, true);
    if (log.isDebugEnabled())
      log.debug("lookup(" + hostname + ", " + name + ")");
    return remoteReg.lookup(name);
  }


  /**
   *  Binds the specified name to a new remote object.  This bind
   *  operation works on the existing local BootstrapRegistry
   *  instance.
   *
   *  @param name  
   *    Name to be associated with the remote object 
   *  @param obj  
   *    Remote object to associate with the name
   * 
   *  @exception AccessException 
   *    if this operation is not permitted (if originating from a
   *    non-local host, for example) 
   *  @exception RemoteException 
   *    if registry could not be contacted
   *  @exception AlreadyBoundException 
   *    if name is already bound
   */
  public static void bind(String name, Remote obj) 
    throws AccessException, RemoteException, AlreadyBoundException
  {
    if (log.isDebugEnabled())
      log.debug("Binding object with name: " + name + ", remoteRef: " + obj);
    if (reg == null)
      reg = getRegistry();
    reg.bind(name, obj);
  }


  /**
   *  Rebinds the specified name to a new remote object.  Any
   *  existing binding for the name is replaced.  This rebind
   *  operation works on the existing local BootstrapRegistry
   *  instance.
   *
   *  @param name  
   *    name to be associated with the remote object 
   *  @param obj  
   *    new remote object to associate with the name
   * 
   *  @exception AccessException 
   *    if this operation is not permitted (if originating from a
   *    non-local host, for example) 
   *  @exception RemoteException 
   *    if registry could not be contacted
   */
  public static void rebind(String name, Remote obj) 
    throws AccessException, RemoteException
  {
    if (log.isDebugEnabled())
      log.debug("Rebinding object with name: " + name + ", remoteRef: " + obj);
    if (reg == null)
      reg = getRegistry();
    reg.rebind(name, obj);
  }

  /**
   *  Removes the binding for the specified name in the bootstrap
   *  registry.
   *
   *  @param name
   *    The name of the binding to remove
   *
   *  @exception AccessException 
   *    if this operation is not permitted (if originating from a
   *    non-local host, for example)
   *  @exception RemoteException 
   *    if registry could not be contacted
   *  @exception NotBoundException 
   *    if name is not bound in the registry
   */
  public static void unbind(String name)
    throws AccessException, RemoteException, NotBoundException
  {
    if (reg == null)
      reg = getRegistry();
    reg.unbind(name);
  }


  /**
   *  Get a reference to a local RMI registry associated with the
   *  bootstrap port specified in the configuration files.  The socket
   *  factories used by the registry will be the default once.
   *
   *  @return 
   *    a <code>Registry</code> object
   * 
   *  @exception AccessException 
   *    if this operation is not permitted (if originating from a
   *    non-local host, for example) 
   *  @exception RemoteException 
   *    if registry could not be contacted
   */
  public static Registry getRegistry()
    throws RemoteException
  {
    return getLocalRegistry(null, null);
  }


  /**
   *  Get a reference to a local RMI registry associated with the
   *  bootstrap port specified in the configuration files.  The socket
   *  factories may be specified.
   *
   *  @param csf
   *    client-side <code>Socket</code> factory used to make
   *    connections to the registry
   *  @param ssf
   *    server-side <code>ServerSocket</code> factory used to accept
   *    connections to the registry
   *  @return 
   *    a <code>Registry</code> object
   * 
   *  @exception AccessException 
   *    if this operation is not permitted (if originating from a
   *    non-local host, for example) 
   *  @exception RemoteException 
   *    if registry could not be contacted
   */
  public static Registry getLocalRegistry(RMIClientSocketFactory csf, RMIServerSocketFactory ssf)
    throws AccessException, RemoteException
  {
    try {
      reg = LocateRegistry.getRegistry(null, port, csf);
      /* 
       * The <code>LocateRegistry.getRegistry()</code> method above simply
       * returns a stub for the bootstrap registry, even in the case when
       * no registry exists.  To circumvent this problem and to ensure that
       * this method always returns a valid registry, we invoke a dummy
       * rebind() method.  If this throws a <code>RemoteException</code> we
       * know that the stub returned above was not a valid registry, and we
       * can safely create a registry in this JVM instead.
       *
       * Further, note that the rebind method cannot take <code>null</code>
       * as its second argument, therefore we simply pass the stub of the
       * registry returned from the above <code>getRegistry()</code> method,
       * since that is a remote object instance.
       */
      reg.rebind(BOOTREGISTRY_TEST, reg);
      if (log.isDebugEnabled()) {
        log.debug("Found bootstrap registry on localhost at port " + port);
      }
    } catch (RemoteException e) {
      if (log.isDebugEnabled())
        log.debug("No bootstrap registry found on localhost");
      /*
       * If createRegistry() below throws a RemoteException we leave
       * it to the invoking class to deal with the problem; it should
       * probably abort.
       */
      try {
        reg = LocateRegistry.createRegistry(port, csf, ssf);
      } catch (RemoteException ex) {
        log.error("Failed to create bootstrap registry on localhost at port " + port);
        throw ex;
      }
      if (log.isDebugEnabled())
        log.debug("Bootstrap registry successfully created on localhost at port " + port);
    }
    /* Return an existing or newly created registry stub. */
    return reg;
  }


  /**
   *  Get a reference to a local RMI registry associated with the specified
   *  bootstrap port and socket factories.
   *
   *  Note that for non-local RMI registries, the returned reference (stub)
   *  may not work since we cannot create a registry on the remote host.
   *  This case may be used to determine if the host is running a registry.
   *
   *  @param hostname
   *    the hostname to query for bootstrap registry
   *  @param csf
   *    client-side <code>Socket</code> factory used to make
   *    connections to the registry
   *  @param checkCache
   *    use true to check the cache for the registry reference; false will
   *    not query the cache, but will update it.
   *  @return 
   *    a <code>Registry</code> object
   * 
   *  @exception AccessException 
   *    if this operation is not permitted (if originating from a
   *    non-local host, for example) 
   *  @exception RemoteException 
   *    if registry could not be contacted
   */
  public static Registry getRemoteRegistry(String hostname, RMIClientSocketFactory csf, boolean checkCache)
    throws UnknownHostException, RemoteException
  {
    if (hostname == null)
      throw new NullPointerException("hostname not specified");

    Registry reg = null;
    if (checkCache)
      reg = registryMap.get(hostname);

    if (status == null || status.isReliable(hostname)) {
      if (status != null)
        log.debug("Partition simulator: link is   up: " + hostname);
      if (reg != null)
        return reg;
    } else {
      if (log.isDebugEnabled())
        log.debug("Partition simulator: link is down: " + hostname);
      throw new RemoteException("Partition simulator: broken link to " + hostname);
    }

    try {
      reg = LocateRegistry.getRegistry(hostname, port, csf);
      /* 
       * The <code>LocateRegistry.getRegistry()</code> method above simply
       * returns a stub for the bootstrap registry, even in the case when
       * no registry exists.
       */
      registryMap.put(hostname, reg);
      if (log.isDebugEnabled())
        log.debug("Found bootstrap registry on " + hostname + " at port " + port);
    } catch (RemoteException e) {
      if (log.isDebugEnabled())
        log.debug("No bootstrap registry found on: " + hostname);
      throw e;
    }

    /* Return the remote registry stub. */
    return reg;
  }


  /**
   *  Create a local RMI registry on the default bootstrap port and
   *  socket factories.
   *
   *  @return 
   *    a <code>Registry</code> object
   * 
   *  @exception AccessException 
   *    if this operation is not permitted (if originating from a
   *    non-local host, for example) 
   *  @exception RemoteException 
   *    if registry could not be contacted
   */
  public static Registry createRegistry()
    throws AccessException, RemoteException
  {
    return createRegistry(null, null);
  }


  /**
   *  Create a local RMI registry on the specified bootstrap port and
   *  socket factories.
   *
   *  @param csf
   *    client-side <code>Socket</code> factory used to make
   *    connections to the registry
   *  @param ssf
   *    server-side <code>ServerSocket</code> factory used to accept
   *    connections to the registry
   *  @return 
   *    a <code>Registry</code> object
   * 
   *  @exception AccessException 
   *    if this operation is not permitted (if originating from a
   *    non-local host, for example) 
   *  @exception RemoteException 
   *    if registry could not be contacted
   */
  public static Registry createRegistry(RMIClientSocketFactory csf, RMIServerSocketFactory ssf)
    throws AccessException, RemoteException
  {
    if (reg != null)
      return reg;
    try {
      reg = LocateRegistry.createRegistry(port, csf, ssf);
      return reg;
    } catch (RemoteException ex) {
      log.error("Failed to create local BootstrapRegistry", ex);
      throw ex;
    }
  }

} // END BootstrapRegistry
