/*
 * Copyright (c) 1998-2002 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package jgroup.core.registry;

import java.io.Externalizable;
import java.lang.reflect.InvocationHandler;
import java.rmi.AccessException;
import java.rmi.NotBoundException;
import java.rmi.RemoteException;
import java.util.Iterator;

import jgroup.core.IID;

/**
 * For obtaining references to remote objects, RMI includes a registry service,
 * that provides methods for storing and retrieving remote object references.
 * Each registry object acts independently and maintains a different set
 * of bindings (name, remote object), thus constituting a single point of
 * failure. Furthermore, a registry object running on a certain host can be
 * used to register only remote objects running on the same host; thus, in
 * order to obtain a stub for a remote object, a client need to be aware of
 * its location.
 * <p>
 * The registry implementation included with Jgroup has several advantages
 * over the standard Java RMI registry. First, it enables a set of replicated
 * remote objects to register themselves under the same name. This set
 * forms a remote object group that can be retrieved as a single entity.
 * All the Jgroup registry objects running in a distributed system
 * cooperate in order to maintain a replicated database of bindings and
 * offer a reliable and high-available repository facility. Clients no
 * longer need to be aware of the location of a service: they can simply
 * access the group of registry replicas and obtain a stub for every
 * service registered in the distributed system. A stub for the registry
 * group is obtained through the <code>RegistryFactory.getRegistry</code>
 * methods.
 * <p>
 * A distributed system can be designed by including a certain number of registry
 * replicas running on different hosts and possibly on distinct portions of the
 * communication network. Client access these replicas through standard
 * RMI interactions as if they were a single registry, and are guaranteed
 * that their invocation will successfully terminate, provided that at
 * least one operational replica is running in their partition.
 * <p>
 * During a partitioning, the Jgroup registry presents a partitioned
 * behavior reflecting the current failure scenario. The registration
 * of a remote object inside a partition will not affect the registry
 * replicas not contained in that partition, while a retrieval operation
 * will not success if the service we are looking for has been registered
 * in another partition. Nevertheless, the replicas contained in a
 * partition consistently maintain the same set of bindings and act as
 * a single entity; moreover, the disappearance of the partitioning
 * causes the execution of a reconciliation protocol in order to
 * re-establish a consistent set of bindings among the replicas that
 * belonged to different partitions. It is important to note that this
 * behavior is perfectly reasonable for a partitionable distributed
 * system, since clients asking for remote services are interested only
 * in servers running in their current partitions.
 * <p>
 * The <code> DependableRegistry</code> interface describes the methods
 * of a dependable registry that can be remotely invoked.
 * To create a registry that runs in an application, use one of the
 * <code>RegistryFactory.createRegistry</code> methods. To obtain
 * a reference to a remote object registry, use one of the
 * <code>RegistryFactory.getRegistry</code> methods.
 *
 *  @author  Alberto Montresor
 *  @author  Hein Meling
 *  @since   Jgroup 0.3
 */
public interface DependableRegistry
  extends LookupRegistry
{
	
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constants
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Service name used to register each dependable registry replica in
   *  the local registry.
   */
  public static final String DEPENDABLE_REGISTRY_NAME = "DependableRegistry";


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Nested interface
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * This interface must be implemented by all types that needs
   * to be stored in a dependable registry.
   */
  public static interface RegistryEntry
    extends Externalizable
  {
    
    /**
     * Returns the entry type.  An <code>DependableRegistry</code>
     * implementation must check the returned value to see if it can
     * handle it.
     * 
     * @return a <code>java.lang.String</code> representing the entry type.
     */
    public String getType();

    /**
     * Returns an invocation handler of the given RegistryEntry type,
     * for the given serviceName.  The handler will include all entries
     * enumerated from the provided iterator.
     */
    public InvocationHandler createInvocationHandler(String serviceName, Iterator entries)
      throws RemoteException;

  } // END of RegistryEntry


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * Add a replicated member object to the remote object group associated with
   * the specified name in the registry.
   *
   * @param name
   *   The name associated with the remote object group
   * @param entry
   *   The RegistryEntry type associated with the member object implementation
   * @param serverType
   *   The class object of the member object implementation
   * @return
   *   A <code>IID</code> identifier that can be used to remove a
   *   replicated remote object from the binding set.
   *
   * @exception RemoteException Raised when remote operation failed.
   * @exception AccessException Raised when this operation is not permitted.
   */
  public IID bind(String name, RegistryEntry entry, Class serverType)
    throws RemoteException, AccessException;

  /**
   * Add a replicated member object to the remote object group associated with
   * the specified name in the registry.
   *
   * @param name
   *   The name associated with the remote object group
   * @param entry
   *   The RegistryEntry type associated with the member object implementation
   * @param serverType
   *   The class object of the member object implementation
   * @param leaseTime
   *   The lease time for the service
   * @return
   *   A <code>IID</code> identifier that can be used to remove a
   *   replicated remote object from the binding set.
   *
   * @exception RemoteException Raised when remote operation failed.
   * @exception AccessException Raised when this operation is not permitted.
   */
  public IID bind(String name, RegistryEntry entry, Class serverType, long leaseTime)
    throws RemoteException, AccessException;

  /**
   * Remove a replicated remote object identified by the <code>IID</code>
   * object obtained when the replica has been registered.
   *
   * @param bid       the binding identifier obtained when the replica has
   *                  been registered
   *
   * @exception RemoteException If remote operation failed.
   * @exception NotBoundException if there is no object with this name in the
   *              registry.
   * @exception AccessException If this operation is not permitted.
   */
  public void unbind(IID bid)
    throws RemoteException, NotBoundException, AccessException;

  /**
   * Update the timestamp (lease time) for the server associated with the given
   * binding identifier (<code>IID</code>).  This method is used by the server-side
   * <code>LeaseLayer</code> and is invoked through the <code>RegistryLayer</code>.
   * 
   * @param iid
   *   The <code>iid</code> of the server whose lease to refresh.
   * 
   * @exception RemoteException If remote operation failed.
   * @exception NotBoundException if there is no object in the registry matching
   *              the given <code>IID</code>.
   * @exception AccessException If this operation is not permitted.
   */
  public void refresh(IID iid)
    throws RemoteException, AccessException, NotBoundException;
    
} // END DependableRegistry
