/*
 * Copyright (c) 1998-2002 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */
 
package jgroup.relacs.config;

import java.io.Serializable;
import java.util.Map;

/**
 *  ClassData specifies the class and arguments of a replica.
 *
 *  @author Hein Meling
 *  @since Jgroup 1.2
 */
public class ClassData 
  implements Serializable
{

  private static final long serialVersionUID = -504244408760314856L;

  /////////////////////////////////////////////////////////////////////////////
  // Fields
  /////////////////////////////////////////////////////////////////////////////

  /** Fully qualified class name */
  private String className;

  /** Arguments passed to the main method of the above class */
  private String[] classArgs;

  /**
   * The map of system properties (key and value) 
   * to append to the JVM command line
   */
  private Map<String,String> properties = null;

  /** Class name without package information */
  private transient String shortClassName;


  /////////////////////////////////////////////////////////////////////////////
  // Constructors
  /////////////////////////////////////////////////////////////////////////////

  public ClassData(String className, String[] classArgs)
//    throws ClassNotFoundException
  {
//    Class cl = Class.forName(className);
    this.className = className;
    this.classArgs = classArgs;
  }


  /////////////////////////////////////////////////////////////////////////////
  // Access methods
  /////////////////////////////////////////////////////////////////////////////

  public Class getClassObject()
  {
    Class cl = null;
    try {
      cl = Class.forName(className);
    } catch (ClassNotFoundException cnf) {
      /*
       * Should never happen since we throw an exception in the
       * constructor if the class was not found, and thus we cannot get
       * this far if the class was not available.
       */
    }
    return cl;
  }

  public String getName()
  {
    return className;
  }

  public String getShortName()
  {
    if (shortClassName == null) {
      shortClassName = className.substring(className.lastIndexOf('.') + 1);
    }
    return shortClassName;
  }

  public String[] getArgs()
  {
    return classArgs;
  }

  public String getArgString()
  {
    StringBuilder buf = new StringBuilder();
    for (int i = 0; i < classArgs.length; i++) {
      buf.append((i != 0) ? " " : "");
      buf.append(classArgs[i]);
    }
    return buf.toString();
  }

  public void setProperties(Map<String,String> properties)
  {
    this.properties = properties;
  }

  public String getCommand(String jvmCmd, String appStarter)
  {
    StringBuilder buf = new StringBuilder(jvmCmd);
    buf.append(" -Djgroup.log.app=");
    buf.append(getShortName());
    if (properties != null) {
      for (Map.Entry<String,String> entry : properties.entrySet()) {
        buf.append(" -D");
        buf.append(entry.getKey());
        buf.append("=");
        buf.append(entry.getValue());
      }
    }
    buf.append(" ");
    buf.append(appStarter);
    buf.append(" ");
    buf.append(className);
    buf.append(" ");
    buf.append(getArgString());
    return buf.toString();
  }

  /////////////////////////////////////////////////////////////////////////////
  // Object methods; overriding from the Object class
  /////////////////////////////////////////////////////////////////////////////

  /**
   *  Returns a string representation of the object.
   */
  public String toString()
  {
    StringBuilder buf = new StringBuilder();
    buf.append("[");
    buf.append(className);
    if (classArgs.length > 0) {
      buf.append(", ");
      buf.append(getArgString());
    }
    buf.append("]");
    return buf.toString();
  }

  /**
   *  Returns a hashcode for the <code>ClassData</code> class.
   *  Two <code>ClassData</code> objects will return the same hashcode
   *  if their class name and arguments are equal.
   */
  public int hashCode()
  {
    int hash = className.hashCode();
    for (int i = 0; i < classArgs.length; i++)
      hash ^= classArgs[i].hashCode();
    return hash;
  }

  /**
   * Compares two objects for content equality.
   *
   * @param obj the object to compare with
   * @return  true if these objects are equal; false otherwise.
   */
  public boolean equals(Object obj)
  {
    if (!(obj instanceof ClassData))
      return false;
    ClassData cd = (ClassData) obj;
    if (!(className.equals(cd.className)) || classArgs.length != cd.classArgs.length) {
      return false;
    }
    boolean eq = true;
    for (int i = 0; eq && i < classArgs.length; i++) {
      eq &= classArgs.equals(cd.classArgs);
    }
    return eq;
  }

} // END of ClassData
