/*
 * Copyright (c) 1998-2002 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package jgroup.relacs.config;

import java.net.InetAddress;
import java.net.UnknownHostException;

import jgroup.core.ConfigurationException;
import jgroup.core.MemberId;

import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;


/**
 *  The <code>ClientsConfig</code> object contains the set of clients
 *  on which client applications may be started.
 *
 *  @author Hein Meling
 *  @since Jgroup 2.1
 */
public final class ClientsConfig
  implements ConfigurationObject
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Tags
  ////////////////////////////////////////////////////////////////////////////////////////////

  private static final String DOMAIN_TAG_NAME            = "Domain";
  private static final String HOST_TAG_NAME              = "Host";
  private static final String NAME_ATTRIB                = "name";
  private static final String ADDRESS_ATTRIB             = "address";
  private static final String PORT_ATTRIB                = "port";


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Static fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Set of all hosts in this distributed system; independent of domains. */
  private static final HostSet allHosts = new HostSet();

  /** The endpoint considered to be the local host. */
  private static Host localHost = null;

  /** The endpoint considered to be the local domain. */
  private static Domain localDomain = null;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Set of domains in this distributed system */
  private DomainSet domains = new DomainSet();


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructors
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Constructs an empty distributed system.
   */
  public ClientsConfig()
  {
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from ConfigurationObject
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Parses the specified element.
   */
  public void parse(Element elm)
    throws ConfigurationException
  {
    /* Nodes within DS should be Domain tags. */
    NodeList domainList = elm.getChildNodes();
    
    /* Parse the list of domains */
    for (int i=1; i <= domainList.getLength(); i++) {
      Node node = domainList.item(i);
      if (node != null && !node.getNodeName().startsWith("#")) {
        elm = (Element) node;
        if (ConfigParser.checkTag(elm, DOMAIN_TAG_NAME)) 
          parseDomain(elm);
      }
    }
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Private methods for parsing
  ////////////////////////////////////////////////////////////////////////////////////////////

  private void parseDomain(Element elm)
    throws ConfigurationException
  {
    /* 
     *  Read attributes; "-" in the domain name corresponds to empty 
     */
    String domainname = ConfigParser.getAttrib(elm, NAME_ATTRIB, true);
    if ("-".equals(domainname))
      domainname = "";
    String address = ConfigParser.getAttrib(elm, ADDRESS_ATTRIB, true);
    int port = ConfigParser.getIntAttrib(elm, PORT_ATTRIB, false);

    /* Create a new domain object with the parsed information */
    Domain domain = null;
    try {
      domain = new Domain(domainname, address, port);
    } catch (UnknownHostException e) {
      throw new ConfigurationException("Unknown multicast address: " + address);
    } catch (IllegalArgumentException e) {
      throw new ConfigurationException(e.getMessage());
    }

    /* Add the domain to the system */
    if (!domains.addDomain(domain))
      throw new ConfigurationException("Multiple definitions of domain: " + domain);

    /* Nodes within the domain should be host tags. */
    NodeList hostList = elm.getChildNodes();
    /* Parse the list of hosts */
    for (int i=1; i <= hostList.getLength(); i++) {
      Node node = hostList.item(i);
      if (node != null && !node.getNodeName().startsWith("#")) {
        elm = (Element)node;
        if (ConfigParser.checkTag(elm, HOST_TAG_NAME)) 
          parseHost(domain, elm);
      }
    }
  }


  private void parseHost(Domain domain, Element elm)
    throws ConfigurationException
  {
    String hostname = ConfigParser.getAttrib(elm, NAME_ATTRIB, true);
    int port = ConfigParser.getIntAttrib(elm, PORT_ATTRIB, true);

    /* Create a new host object with the parsed information */
    Host host = null;
    try {
      host = new Host(hostname, domain, port);
    } catch(UnknownHostException e) {
      throw new ConfigurationException("Unknown host: " + hostname);
    }

    /* Add the host to this domain */
    if (!domain.addHost(host))
      throw new ConfigurationException("Multple definitions of host: " + host);

    /* Add the host to the all hosts set */
    allHosts.addHost(host);

    /* Check if this host is a local host */
    if (host.isLocal()) {
      localHost = host;
      localDomain = domain;
    }
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Access methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Returns the parsed set of domains.
   */
  public DomainSet getDomainSet()
  {
    return domains;
  }


  /**
   *  Returns a set of all hosts parsed; independent of domain.
   */
  public HostSet getAllHosts()
  {
    return allHosts;
  }


  /**
   *  Returns the size of all hosts parsed; independent of domain.
   */
  public int size()
  {
    return allHosts.size();
  }


  /**
   *  Returns the number of domains parsed.
   */
  public int numOfDomains()
  {
    return domains.size();
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Access methods - only for server side
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Returns the host object associated with the given member.
   */
  public static Host getHost(MemberId member)
  {
    return allHosts.getHost(member);
  }


  /**
   *  Returns the host object associated with the given internet address.
   */
  public static Host getHost(InetAddress inetAdr)
  {
    return allHosts.getHost(inetAdr);
  }


  /**
   *  Returns the endpoint that is considered local to this host.
   */
  public Host getLocalHost()
  {
    return localHost;
  }


  /**
   *  Returns the domain endpoint that is considered local to this
   *  domain.
   */
  public Domain getLocalDomain()
  {
    return localDomain;
  }


  /**
   *  Returns true if the local host is specified in the distributed
   *  system configuration; false is returned otherwise.
   */
  public boolean containsLocalHost()
  {
    return localHost != null;
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Override methods in Object
  ////////////////////////////////////////////////////////////////////////////////////////////

  public int hashCode()
  {
    return domains.hashCode() ^ allHosts.hashCode();
  }


  /**
   *  Returns a string representation of this object
   */
  public String toString()
  {
    return toString(false);
  }

  /**
   *  Returns a string representation of this object;
   *  
   *  @param full If true, prints also the complete host set for each domain.
   */
  public String toString(boolean full)
  {
    StringBuilder buf = new StringBuilder();
    buf.append("[ClientsConfig:\n");
    buf.append(domains.toString(full));
    buf.append("]");
    return buf.toString();
  }

} // END ClientsConfig
