/*
 * Copyright (c) 1998-2002 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package jgroup.relacs.config;

import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.lang.reflect.Modifier;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;

import jgroup.core.ConfigurationException;
import jgroup.core.GroupManager;
import jgroup.core.InternalGMIListener;
import jgroup.core.JgroupException;
import jgroup.core.Layer;
import jgroup.core.MalformedLayerException;

import org.apache.log4j.Logger;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;




/**
 *  Object for holding layer information and creating instances of the
 *  layer, and add listeners between the layers.
 *
 *  @author Hein Meling
 *  @since Jgroup 1.2
 */
public final class LayerConfig
  implements ConfigurationObject
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Logger
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Obtain logger for this class */
  private static final Logger log = Logger.getLogger(LayerConfig.class);


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Tags
  ////////////////////////////////////////////////////////////////////////////////////////////

  private static final String DEPENDS_TAG_NAME  = "Depends";
  private static final String REPLACES_TAG_NAME = "Replaces";
  private static final String SERVICE_ATTRIB    = "service";
  private static final String CLASS_ATTRIB      = "class";


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Static collections
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * Map of layers (association from service name to LayerConfig object).
   * Note that this map must be static to allow cross access between
   * instances of the LayerConfig class.  This is because the parser
   * creates a new LayerConfig object for each Layer entry in the XML
   * file.
   */
  private static final Map<String, LayerConfig> layers =
    new HashMap<String, LayerConfig>();


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /* The service implemented by this layer. */
  private String service;

  /* The layer class implementing the above service. */
  private String layerClassName;
  private Class layerClass;

  /* The getLayer method of this layer. */
  private Method getLayerMethod;

  /* The set of interfaces implemented by this layer */
  private Set<Class> layerIfaces;

  /* Set of services on which this layer depends */
  private Set<String> depends;

  /* Set of services that this layer replaces. */
  private Set<String> replaces;

  /*
   * An instance representing this layer.  Can be obtained through the
   * getLayerObject() method.
   */
  private Layer layer = null;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructors
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Constructs an empty layer config object.
   */
  public LayerConfig() { }


  /**
   *  Constructs a layer config object for the specified class.
   */
  public LayerConfig(String service, String layerClassName)
  {
    this.service = service;
    this.layerClassName = layerClassName;
  }


  private void initLayer()
    throws ConfigurationException
  {
    try {
      layerClass = Class.forName(layerClassName);
    } catch (ClassNotFoundException e) {
      throw new ConfigurationException("Layer class not found: " + layerClassName);
    }
    // Ensure that the getLayer() method is nullified.
    getLayerMethod = null;

    Method[] methods = layerClass.getMethods();
    for (int i = 0; i < methods.length; i++) {
      if (methods[i].getName().equals("getLayer")) {
        if (getLayerMethod != null)
          throw new ConfigurationException(layerClassName 
            + " provides multiple getLayer() methods; only one is accepted.");
        getLayerMethod = methods[i];
        int m = getLayerMethod.getModifiers();
        if (!Modifier.isPublic(m) || !Modifier.isStatic(m))
          getLayerMethod = null;
      }
    }
    if (getLayerMethod == null) 
      throw new ConfigurationException("The layer (" + layerClassName
        + ") implementing the " + service
        + " service does not declare the required 'public static getLayer()' method.");

    Class[] ifaces = layerClass.getInterfaces();
    layerIfaces = new HashSet<Class>(ifaces.length);
    for (int i = 0; i < ifaces.length; i++)
      layerIfaces.add(ifaces[i]);
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from ConfigurationObject
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Parses the specified element.
   */
  public void parse(Element elm)
    throws ConfigurationException
  {
    String servicename = ConfigParser.getAttrib(elm, SERVICE_ATTRIB, true);
    String clazz       = ConfigParser.getAttrib(elm, CLASS_ATTRIB,   true);

    /* Create a new service object with the parsed information */
    LayerConfig layer = new LayerConfig(servicename, clazz);
    layers.put(servicename, layer);

    /* Nodes within the 'Layer' tag. */
    NodeList layerList = elm.getChildNodes();

    /* Parse the list of tags inside the Layer tag. */
    for (int i = 1; i <= layerList.getLength(); i++) {
      Node node = layerList.item(i);
      if (node != null && !node.getNodeName().startsWith("#")) {
        elm = (Element) node;
        if (ConfigParser.checkTag(elm, DEPENDS_TAG_NAME)) 
          layer.addDependency(ConfigParser.getAttrib(elm, SERVICE_ATTRIB, true));
        else if (ConfigParser.checkTag(elm, REPLACES_TAG_NAME))
          layer.addReplaces(ConfigParser.getAttrib(elm, SERVICE_ATTRIB, true));
      }
    }
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Static Access methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * Check that layer dependencies are satisfied, and prepare for
   * construction the layers given in the service set. 
   */
  public static void initializeLayers()
    throws ConfigurationException
  {
    /*
     * Check dependencies and reset the layers
     */
    for (Iterator iter = layers.entrySet().iterator(); iter.hasNext();) {
      Map.Entry layerEntry = (Map.Entry) iter.next();
      String serviceName = (String) layerEntry.getKey(); 
      /* Get service interface for the specified service */
//      Class serviceIface = ServiceConfig.getServiceIface(serviceName);
      /* Get the layer that implements the specified service */
      LayerConfig layerConfig = (LayerConfig) layerEntry.getValue();
      /*
       * Check if the layer implements the required service interface,
       * or an interface derived from the given service interface
       */
//      if (!layerConfig.contains(serviceIface))
//        throw new ConfigurationException("The layer providing the " + serviceName 
//          + " service does not implement the required service interface: " + serviceIface);
      /*
       * Reset each of layers in the current group manager stack.
       * Note that since multiple group managers may exist within
       * the same JVM, we have to perform this reset operation before
       * we can create the layers.
       */
      layerConfig.resetLayerObject();
    }
  }


  /**
   * Construct the layer for the given service name.
   *
   * @param serviceName
   * @param layerMap
   * @param internalIfaces
   * @throws JgroupException
   */
  @SuppressWarnings("unchecked")
  public static void constructLayerObject(String serviceName, Map layerMap, List<Class> internalIfaces)
    throws JgroupException
  {
    LayerConfig layerConfig = layers.get(serviceName);
    if (layerConfig == null)
      throw new ConfigurationException("The layer for " + serviceName + " is not specified");
    layerConfig.initLayer();
    Layer layer = layerConfig.getLayerObject(layerMap, internalIfaces);

    /* Associate the service name of this layer with the layer instance. */
    layerMap.put(serviceName, layer);

    /*
     * Adds the *Service.class literal as the key to the service name.
     * This is needed to check the dependency specifications.  See the
     * LayerInfo.getLayerObject() method for details.
     */
    Class serviceIface = ServiceConfig.getServiceIface(serviceName);
    layerMap.put(serviceIface, serviceName);
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Private methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  private void resetLayerObject()
  {
    layer = null;
  }


  @SuppressWarnings("unchecked")
  private Layer getLayerObject(Map layerMap, List<Class> internalIfaces)
    throws JgroupException
  {
    /* If this layer was already constructed, just return it. */
    if (layer != null) {
      if (log.isDebugEnabled())
        log.debug(layerClass + " layer already constructed");
      return layer;
    }

    for (Iterator<String> iter = depends(); iter.hasNext(); ) {
      String dependsOn = iter.next();
      if (!layerMap.containsKey(dependsOn)) {
        if (log.isDebugEnabled())
          log.debug("Creating: " + dependsOn + " since " + service + " depends on it");
        constructLayerObject(dependsOn, layerMap, internalIfaces);
      }
    }

    /*
     * Parse the getLayer() method to check if it contains the GroupManager
     * in the first argument position; if so, pass the group manager reference
     * to the layer.
     */
    Class[] parameters = getLayerMethod.getParameterTypes();
    Object[] obj = new Object[parameters.length];
    int i = 0;
    if (parameters.length > 0 && parameters[i].equals(GroupManager.class)) {
      /*
       * Check that the remaining number of parameters (that is, exclusive
       * the group manager parameter) match the number of depends.
       */
      if (parameters.length-1 != depends.size()) {
        throw new MalformedLayerException("Arguments to the getLayer method of service " 
          + service + " must correspond to the layers on which this layer depends.");
      }
      obj[i++] = layerMap.get("GM");
    } else {
      if (parameters.length != depends.size()) {
        throw new MalformedLayerException("Arguments to the getLayer method of service " 
          + service + " must correspond to the layers on which this layer depends.");
      }
    }

    /*
     * Parse the remaining parameters for the getLayer() method
     */
    for ( ; i < parameters.length; i++) {

      if (!Layer.class.isAssignableFrom(parameters[i]))
        throw new MalformedLayerException("Arguments to the getLayer method must extend jgroup.core.Layer");

      if (!layerMap.containsKey(parameters[i]))
        throw new MalformedLayerException("The layer " + parameters[i] 
          + " is required to initialize the " + service + " layer.");

      /*
       * Get the service name associated with this formal parameter.
       * The parameters array gives the service name in the form
       * *Serivce.class, while the object returned from the layerMap
       * will be a string specifying the service name.  This service
       * name can be used to lookup in the depends map, to determine if
       * it is specified also in the XML file.
       */
      obj[i] = layerMap.get(parameters[i]);
      if (!depends.contains(obj[i]))
        throw new MalformedLayerException("The layer " + parameters[i] 
          + " is not specified in the dependencies for the " + service + " layer.");

      /*
       * Get the layer instance associated with this service name
       * (obtained above using the formal parameter).  The service name
       * is in string form, while the object returned from the layerMap
       * will be of the actual layer instance object.
       */
      obj[i] = layerMap.get(obj[i]);
    }

    try {
      /*
       * Get an instance of this layer, instantiated using the correct
       * formal parameters.
       */
      layer = (Layer) getLayerMethod.invoke(null, obj);
    } catch (ClassCastException e) {
      throw new JgroupException(layerClass + " or its service interface do not implement the Layer interface");
    } catch (IllegalAccessException e) {
      throw new JgroupException(layerClass + " cannot be accessed (check if class has package access)");
    } catch (InvocationTargetException e) {
      throw new JgroupException("Constructing " + layerClass + " has thrown an exception:",
                                e.getTargetException());
    }

    /*
     * Add the newly created layer as a listener to other layers; that
     * is layers on which this layer depends.  Note that a layer may
     * depend on a service, but still not implement the listener
     * interface.
     */
    for (Iterator<String> it = depends(); it.hasNext(); ) {
      String dependsOnService = it.next();

      /*
       * Get the listener interface for the service on which this layer
       * depends.  This will be used below to check if this layer
       * implements a the listener for service on which it depends.
       * Note that a layer may depend on a service, but still not
       * implement the listener interface.
       */
      Class listenerIface = ServiceConfig.getListenerIface(dependsOnService);

      /*
       * Get an interface associated with this layer, that is assignable from
       * the listener interface of the service on which this layer depends.
       */
      Class assignableIface = getAssignableIface(listenerIface);
      if (assignableIface != null) {
        /*
         * Special handling for IGMI interfaces:
         * Find the internal interfaces implemented by this layer.
         */
        if (InternalGMIListener.class.isAssignableFrom(assignableIface)) {
          internalIfaces.add(assignableIface);
          /*
           * Add also the layer interface to the layer map,
           * so that the interfaces assignable from a layer
           * interface can be reached.  This will allow using
           * getService(jgroup.test.gm.InternalRemoteTest.class).
           * The actual layer implementing the IGMI service will
           * be replaced by a proxy for the layer (see getGroupManager).
           */
          layerMap.put(assignableIface, "IGMI");
        }

        /*
         * This layer implements the listener interface for the service
         * on which it depends.  We add this layer as a listener for the
         * layer on which it depends.
         */
        Layer dependOnLayer = (Layer) layerMap.get(dependsOnService);
        dependOnLayer.addListener(layer);
        if (log.isDebugEnabled())
          log.debug(service + ": listen to events from " + dependsOnService);
      } else {
        if (log.isDebugEnabled())
          log.debug(service + ": does not implement listener interface for: "
            + dependsOnService + " (this is not a requirement)");
      }
    }

    /*
     * Replace other layers in the layerMap with this layer, if their
     * service is specified in the replaces tag.  Note that a layer that
     * replaces a service, must also implement the listener interface of
     * the service it is replacing.
     */
    for (Iterator<String> it = replaces(); it.hasNext(); ) {
      String replacesService = it.next();

      /*
       * Get the listener interface for the service that this layer
       * replaces.  This will be used below to check if this layer
       * implements a the listener for service that it replaces.  Note
       * that a layer cannot replace a service without also implementing
       * the listener interface of the service.
       */
      Class listenerIface = ServiceConfig.getListenerIface(replacesService);

      /* Check if this layer implements the listener interface. */
      if (layerIfaces.contains(listenerIface)) {
        /*
         * This layer implements the listener interface for the service
         * that it replaces.  We update the layerMap with this layer in
         * the position of the service that we are replacing.
         */
        layerMap.put(replacesService, layer);
        if (log.isDebugEnabled()) 
          log.debug(service + ": replaces " + replacesService);
      } else
        throw new MalformedLayerException(service + " does not implement the listener " 
          + " interface required to replace the service: " + replacesService);
    }

    /* Return the newly created layer */
    return layer;
  }


  /**
   * Check wheter the layer implements the given interface or
   * a descendand of the given layer.
   * 
   * @param iface Interface to implement
   * 
   * @return true if layer implements the given interface or one of its descendants
   */
  private boolean contains(Class iface)
  {
    return (getAssignableIface(iface) != null);
  }


  /**
   *  Return the provided interface if it is directly implemented by
   *  this layer.  If not, iterate over the interfaces implemented
   *  by this layer to determine if the provided interface is
   *  assignable from one of the interfaces implemented by this layer.
   *  In the latter case, the interface implemented by the layer that
   *  is assignable from the provided interface is returned.
   *
   *  @param iface
   *    The interface to check if is assignable from any of the
   *    interfaces in this layer.
   *  @return
   *    The interface to which the provided interface is assignable.
   *    If no interfaces are assignable from the given interface,
   *    <code>null</code> is returned.
   */
  private Class getAssignableIface(Class<?> iface)
  {
    if (layerIfaces.contains(iface)) {
      return iface;
    }
    for (Class layerIface : layerIfaces) {
      if (iface.isAssignableFrom(layerIface)) {
        return layerIface;
      }
    }
    return null;
  }


  private void addDependency(String dependsOn)
    throws ConfigurationException
  {
    if (!layers.containsKey(dependsOn))
      throw new ConfigurationException("Layer " + service + " depends on an unknown service " + dependsOn);
    if (service.equals(dependsOn))
      throw new ConfigurationException("The layer " + service + " cannot depend on itself.");
    if (depends == null)
      depends = new HashSet<String>(7);
    depends.add(dependsOn);
  }


  private void addReplaces(String replacee)
    throws ConfigurationException
  {
    if (!layers.containsKey(replacee))
      throw new ConfigurationException("Layer " + service + " specifies to replace an unknown service " + replacee);
    if (service.equals(replacee))
      throw new ConfigurationException("The layer " + service + " cannot replace itself.");
    if (replaces == null)
      replaces = new HashSet<String>(4);
    replaces.add(replacee);
  }


  private Iterator<String> depends()
  {
    if (depends == null)
      depends = new HashSet<String>(0);
    return depends.iterator();
  }


  private Iterator<String> replaces()
  {
    if (replaces == null)
      replaces = new HashSet<String>(0);
    return replaces.iterator();
  }

} // END LayerConfig
