/*
 * Copyright (c) 1998-2002 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package jgroup.relacs.daemon;

import java.util.Collection;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;
import java.util.SortedMap;
import java.util.TreeMap;

import jgroup.core.EndPoint;

import org.apache.log4j.Logger;

/**
 *  The <code>Estimate</code> class
 *
 *  @author  Alberto Montresor
 *  @since   Jgroup 0.3
 */
final class Estimate
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Logger
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Obtain logger for this class */
  private static final Logger log = Logger.getLogger(Estimate.class);


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Map from EndPoint to HostData objects */
  private final SortedMap<EndPoint,HostData> estimate = new TreeMap<EndPoint,HostData>();

  /** Set of reachable endpoints that remains to be synchronized */
  private final Set<EndPoint> toSynchronize = new HashSet<EndPoint>();

  /** True if estimate has changed */
  private boolean changed;

  /** Send buffer used to notify acknowledgment */
  private final SendBuffer sbuffer;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructors
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Creates a new estimate by adding the hosts contained in
   *  <code>rset</code>.
   *
   *  @param sbuffer     send buffer used to communicate the list of suspect
   *                     to remove
   *  @param thosts      the total list of hosts
   *  @param rset        the reachable set
   */
  Estimate(SendBuffer sbuffer, Map thosts, EndPoint[] rset)
  {
    /* Initialize the estimate to empty */
    this.sbuffer = sbuffer;
    initEstimate(thosts, rset);
  }


  void reset(Map thosts, EndPoint[] rset)
  {
    estimate.clear();
    toSynchronize.clear();
    initEstimate(thosts, rset);
  }


  private void initEstimate(Map thosts, EndPoint[] rset)
  {
    /* Add the elements in the reachable sets */
    for (int i=0; i < rset.length; i++) {
      if (log.isDebugEnabled()) 
        log.debug("rset[" +i+ "]: " +rset[i]);
      HostData host = (HostData) thosts.get(rset[i]);
      if (host == null) {
        /*
         * Note that the system can contain an endpoint for a node that
         * do not belong to the group; the node may belong to a
         * different group.  We simply ignore those entries.
         */
        continue; 
      }
      if (!host.isLeaving()) {
        estimate.put(rset[i], host);
        if (!host.getHost().isLocal())
          toSynchronize.add(rset[i]);
      } else {
        sbuffer.suspect(host);
      }
    }

    if (log.isDebugEnabled())
      log.debug(toString());
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Returns true if the agreed version numbers of the intersection
   *  between the estimate and endpoints (P) are equal.  Otherwise,
   *  false is returned.
   *
   *  See mrecv(ESTIMATE) in Figure 5.8, page 94 of Albertos thesis;
   *  the elseif part.
   */
  boolean checkAgreed(EndPoint[] endpoints, int[] agreed)
  {
    for (int i=0; i < endpoints.length; i++) {
      HostData host = estimate.get(endpoints[i]);
      if (host != null && host.getAgreed() != agreed[i]) {
        if (log.isDebugEnabled())
          log.debug(endpoints[i] + ": local agreed=" + host.getAgreed() + ", remote agreed=" + agreed[i]);
        return false;
      }
    }
    return true;
  }


  /**
   * Update the agreed values for the given hosts.
   * 
   * @param hosts The hosts whose agreed value to update
   * @param agreed The agreed version numbers received with the hosts
   */
  void updateAgreed(EndPoint[] hosts, int[] agreed)
  {
    if (agreed != null) {
      for (int i = 0; i < hosts.length; i++) {
        HostData host = estimate.get(hosts[i]);
        if (host != null)
          host.setAgreed(agreed[i]);
      }
    }
  }


  /**
   *  Computes the intersection between the current estimate and
   *  the given set of <code>hosts</code>.
   *
   *  @param hosts
   *    The hosts to be intersected with the current estimate
   */
  void intersect(EndPoint[] hosts)
  {
    /* Identify the elements to be removed */
    Set<EndPoint> toRemove = new HashSet<EndPoint>(estimate.keySet());
    for (EndPoint host : hosts)
      toRemove.remove(host);

    /* Remove the elements in toRemove */
    for (EndPoint endpoint : toRemove) {
      changed = true;
      sbuffer.suspect(estimate.get(endpoint));
      estimate.remove(endpoint);
      toSynchronize.remove(endpoint);
    }
  }


  /**
   *  Remove the host identified by <code>endpoint</code> from the
   *  estimate.
   *
   *  @param endpoint
   *    The endpoint to be removed.
   */
  void remove(EndPoint endpoint)
  {
    /* Remove the endpoint */
    HostData host = estimate.remove(endpoint);
    if (host != null) {
      toSynchronize.remove(endpoint);
      sbuffer.suspect(host);
      changed = true;
    }
  }

  
  /**
   *  Remove from estimate, those hosts whose endpoint address is
   *  contained in the given <code>set</code>.
   *  
   *  @param set
   *    The endpoints of the hosts to be removed.
   */
  void remove(EndPoint[] set)
  {
    for (int i=0; i < set.length; i++)
      remove(set[i]);
  }


  /**
   *  Returns true if the specified endpoint is included in the estimate;
   *  false otherwise.
   *  
   *  @param endpoint the endpoint to check
   */
  boolean contains(EndPoint endpoint)
  {
    return estimate.containsKey(endpoint);
  }

  /**
   *  Return the <CODE>HostData</CODE> object associated to the specified endpoint if the
   *  host is contained in the estimate; null otherwise.
   */
  HostData get(EndPoint endpoint)
  {
    return estimate.get(endpoint);
  }

  /**
   *  Returns a collection containing the <CODE>HostData</CODE> objects
   *  associated with the hosts included in the estimate.
   */
  Collection<HostData> getHosts()
  {
    return estimate.values();
  }

  /**
   *  Returns array of the endpoints contained in this estimate.
   */
  EndPoint[] getEndPoints()
  {
    return estimate.keySet().toArray(new EndPoint[estimate.size()]);
  }
  
  /**
   *  Returns the size of this estimate.  
   */
  int size()
  {
    return estimate.size();
  }

  /**
   *  Returns true if this estimate is empty; false otherwise.
   */
  boolean isEmpty()
  {
    return estimate.isEmpty();
  }

  /**
   * Returns true if the estimates are equal.
   */
  public boolean equals(Object o)
  {
    if (o instanceof Estimate) {
      Estimate e = (Estimate) o;
      return (e.estimate.equals(this.estimate) && e.toSynchronize.equals(this.toSynchronize)); 
    }
    return false;
  }

  /**
   *  Add the specified endpoint to the set of synchronized objects.
   */
  void setSynchronized(EndPoint endpoint)
  {
    toSynchronize.remove(endpoint);
  }

  /**
   *  Return true if every element is synchronized 
   */
  boolean isSynchronized()
  {
    return toSynchronize.size() == 0;
  }

  /**
   *  Return true if the last operation on this estimate has changed it.
   */
  boolean hasChanged()
  {
    boolean ret = changed;
    changed = false;
    return ret;
  }

  /**
   *  Return the endpoint of the host that acts as coordinator.
   */
  EndPoint getCoordinator()
  {
    if (estimate.isEmpty()) {
      log.warn("Estimate is empty: " + estimate);
      return null;
    }
    return (EndPoint) estimate.firstKey();
  }


  /**
   *  Returns a string representation of this object
   */
  public String toString()
  {
    StringBuilder buf = new StringBuilder();
    buf.append("[Estimate: size=");
    buf.append(estimate.size());
    buf.append(", comp={");
    for (Iterator<EndPoint> iter = estimate.keySet().iterator(); iter.hasNext();) {
      EndPoint endpoint = iter.next();
      buf.append(endpoint);
      if (iter.hasNext())
        buf.append(", ");
    }
    buf.append("}");
    buf.append(", toSynchronize=");
    buf.append(toSynchronize);
    buf.append("]");
    return buf.toString();
  }
  
} // END Estimate
