/*
 * Copyright (c) 1998-2002 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package jgroup.relacs.daemon;

import java.io.IOException;
import java.io.ObjectInput;

import jgroup.relacs.mss.MssConstants;
import jgroup.relacs.types.GroupId;
import jgroup.relacs.types.GroupIndex;
import jgroup.relacs.types.MessageId;
import jgroup.relacs.types.ViewId;
import jgroup.util.MsgFactory;
import jgroup.util.OutMessage;

/**
 *  The <code>MsgAck</code> class contain methods to encode and decode
 *  ACK messages from remote hosts.  The ACK message structure is the
 *  following:
 *  <p>
 *  <code>
 *  +-------+-------+-------+-------+-------+ ....... +-------+
 *  |  GID  |  VID  |  HPOS | VLEN  |  ACK  |         |  ACK  |
 *  |       |       |       |       |   1   |         |  VLEN |
 *  +-------+-------+-------+-------+-------+ ....... +-------+
 *  </code>
 *
 *  @author  Alberto Montresor
 *  @since   Jgroup 0.1
 */
final class MsgAck 
  implements MssConstants, DaemonMsg
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Position constants
  ////////////////////////////////////////////////////////////////////////////////////////////

  private static final int START       = MSS_HEADER_SIZE;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Unique message instance
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * As these messages are never stored by the daemon, a single message
   * instance is sufficient.  In this way, we save the costs of
   * allocating and garbage collecting messages.
   */
  private static MsgAck msg = new MsgAck();

  /** The marshalled message data (also static for same reason as above) */
  private static OutMessage outmsg;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** View identifier */            
  long vid;      

  /** Host index in the view */
  int hpos;     

  /** Acknowledgement numbers for each host in the current view */
  int[] ack;      


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Returns a <code>MsgAck</code> object that contains the decoding of
   *  the m-received input stream.
   *
   *  @param stream    the message input stream to decode
   */
  static MsgAck unmarshal(ObjectInput stream)
    throws IOException
  {
    msg.vid  = ViewId.unmarshal(stream);
    msg.hpos = GroupIndex.unmarshal(stream);

    /* Read length and content of the ack array */
    int len  = GroupIndex.unmarshal(stream);
    msg.ack  = new int[len];
    for (int i=0; i < len; i++)   // Read ack fields
      msg.ack[i] = MessageId.unmarshal(stream);

    return msg;
  }


  /**
   *  Returns a output stream containing the encoding of a
   *  <code>MsgAck</code> message.
   *
   *  @param     gid     group identifier
   *  @param     vid     view identifier
   *  @param     hpos    host index in the view
   */
  static MsgAck marshal(int gid, long vid, int hpos, int[] acks, int nreceivers)
    throws IOException
  {
    // Allocates data buffer
    int size = START + GroupId.SIZE + ViewId.SIZE + GroupIndex.SIZE*2 + MessageId.SIZE*acks.length;
    outmsg = MsgFactory.get(size, nreceivers);
    outmsg.seek(START);

    GroupId.marshal(outmsg, gid);
    ViewId.marshal(outmsg, vid);
    GroupIndex.marshal(outmsg, hpos);
    GroupIndex.marshal(outmsg, acks.length);
    for (int i=0; i < acks.length; i++)
      MessageId.marshal(outmsg, acks[i]);

    return msg;
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from DaemonMsg
  ////////////////////////////////////////////////////////////////////////////////////////////

  public int size()
  {
    /* Note; it is ok to use the outmsg size since MsgAck will never be
     * forwarded or resent. */
    return outmsg.size();
  }

  public OutMessage getOutMessage()
  {
    return outmsg;
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from Object
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Returns a string representation of this object
   */
  public String toString()
  {
    StringBuilder buffer = new StringBuilder();
    buffer.append("[MsgAck: vid=");
    buffer.append(vid);
    buffer.append(", hpos=");
    buffer.append(hpos);
    buffer.append(", ack={");
    for (int i=0; i < ack.length; i++) {
      buffer.append(ack[i]);
      if (i != ack.length-1)
        buffer.append(", ");
    }
    buffer.append("}]");
    return buffer.toString();
  }

} // END MsgAck
