/*
 * Copyright (c) 1998-2002 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package jgroup.relacs.daemon;


import java.io.IOException;
import java.io.ObjectInput;
import java.util.Collection;
import java.util.Iterator;

import jgroup.core.EndPoint;
import jgroup.relacs.mss.MssConstants;
import jgroup.relacs.types.EndPointImpl;
import jgroup.relacs.types.GroupId;
import jgroup.relacs.types.GroupIndex;
import jgroup.relacs.types.VersionId;
import jgroup.util.MsgFactory;
import jgroup.util.OutMessage;

/**
 * <code>MsgEstim</code> messages are used during the ee-phase (estimate exchange) of
 * the view agreement algorithm. They are sent to daemons belonging to the current
 * estimate whenever this estimate changes.
 *
 * @author  Alberto Montresor
 * @since   Jgroup 0.1
 */
final class MsgEstim
  implements MssConstants, DaemonMsg
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Position constants
  ////////////////////////////////////////////////////////////////////////////////////////////

  private static final int START       = MSS_HEADER_SIZE;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Unique message instance
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * As these messages are never stored by the daemon, a single message instance is 
   * sufficient. In this way, we save the costs of allocating and garbage collecting
   * messages.
   */
  private static MsgEstim msg = new MsgEstim();

  /** The marshalled message data (also static for same reason as above) */
  private static OutMessage outmsg;


 ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Agreed version number of the sender */
  int vsend;      
  
  /** Agreed version number table of next view estimate */
  int[] agreed;
  
  /** Next view estimate */
  EndPoint[] hosts;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Message to send
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Returns a <code>MsgEstim</code> object that contains the decoding
   *  of the m-received input stream.
   *
   *  @param stream    the message input stream to decode
   */
  static MsgEstim unmarshal(ObjectInput stream)
    throws IOException, ClassNotFoundException
  {
    msg.vsend  = VersionId.unmarshal(stream);
    int len = GroupIndex.unmarshal(stream);         // Read length of agreed and hosts
    msg.agreed = new int[len];
    msg.hosts  = new EndPoint[len];
    for (int i=0; i<len; i++) {                   // Read agreed and hosts
      msg.agreed[i] = VersionId.unmarshal(stream);
      msg.hosts[i] = new EndPointImpl();
      msg.hosts[i].readExternal(stream);
    }
    return msg;
  }

  /**
   *  Returns a output stream containing the encoding of a
   *  <code>MsgEstim</code> message.
   *
   *  @param     gid       group identifier
   *  @param     vsend     version number of the sender
   *  @param     estimate  current estimate
   *  @param     elen      actual length of estimate (estimate.length >= elen)
   */
  static MsgEstim marshal(int gid, int vsend, Collection estimate)
    throws IOException
  {
    // Allocates data buffer
    int size = START + GroupId.SIZE + VersionId.SIZE + GroupIndex.SIZE + 
               (VersionId.SIZE + EndPointImpl.SIZE) * estimate.size();
    outmsg = MsgFactory.get(size, estimate.size());
    outmsg.seek(START);

    GroupId.marshal(outmsg,  gid);
    VersionId.marshal(outmsg, vsend);

    // Write length of agreed and hosts
    GroupIndex.marshal(outmsg, estimate.size());

    // Write agreed and hosts
    Iterator iterator = estimate.iterator();
    while (iterator.hasNext()) {
      HostData scan = (HostData) iterator.next();
      VersionId.marshal(outmsg, scan.getAgreed());
      scan.getEndPoint().writeExternal(outmsg);
    }
    return msg;
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from DaemonMsg
  ////////////////////////////////////////////////////////////////////////////////////////////

  public int size()
  {
    /* Note; it is ok to use the outmsg size since MsgEstim will never be
     * forwarded or resent (it is not queued). */
    return outmsg.size();
  }

  public OutMessage getOutMessage()
  {
    return outmsg;
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from Object
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Returns a string representation of this object
   */
  public String toString()
  {
    StringBuilder buffer = new StringBuilder();
    buffer.append("[MsgEstim: vsend=");
    buffer.append(vsend);
    buffer.append(", hosts={");
    for (int i=0; i < hosts.length; i++) {
      buffer.append(hosts[i]);
      if (i < hosts.length-1)
        buffer.append(", ");
    }
    buffer.append("}, agreed={");
    for (int i=0; i < agreed.length; i++) {
      buffer.append(agreed[i]);
      if (i < agreed.length-1)
        buffer.append(", ");
    }
    buffer.append("}]");
    return buffer.toString();
  }

} // END MsgEstim
