/*
 * Copyright (c) 1998-2002 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package jgroup.relacs.daemon;


import java.io.IOException;
import java.io.ObjectInput;
import java.util.Collection;
import java.util.Iterator;
import java.util.Map;

import jgroup.core.EndPoint;
import jgroup.relacs.mss.MssConstants;
import jgroup.relacs.types.Checksum;
import jgroup.relacs.types.EndPointImpl;
import jgroup.relacs.types.GroupId;
import jgroup.relacs.types.GroupIndex;
import jgroup.relacs.types.IncarnationId;
import jgroup.relacs.types.LocalId;
import jgroup.relacs.types.MessageId;
import jgroup.relacs.types.MessageLen;
import jgroup.relacs.types.VersionId;
import jgroup.relacs.types.ViewId;
import jgroup.util.Abort;
import jgroup.util.InMessage;
import jgroup.util.MsgFactory;
import jgroup.util.OutMessage;

// FIXME Reorganize this class in order to be more readable; we should
// divide it in a "Proposal" class and an MsgProp class

/**
 * The <code> MsgProp</code> class
 *
 * @author  Alberto Montresor
 * @since   Jgroup 0.1
 */
final class MsgProp
  implements MssConstants, DaemonMsg
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Position constants
  ////////////////////////////////////////////////////////////////////////////////////////////

  private static final int START  = MSS_HEADER_SIZE;
  private static final int P_DATA = 
    START + GroupId.SIZE + (ViewId.SIZE * 2) + (MessageLen.SIZE * 4) + (Checksum.SIZE * 2);


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  long                  cvid;           // Last complete view identifier
  long                  pvid;           // Last partial view identifier
  int                   last_plen;      // last_prop field length (in bytes)
  int                   host_plen;      // host_prop field length (in bytes)
  int                   memb_plen;      // memb_prop field length (in bytes)
  int                   msgs_plen;      // msgs_prop field length (in bytes)
  int                   last_start;     // last_prop field start position
  int                   host_start;     // host_prop field start position
  int                   memb_start;     // memb_prop field start position
  int                   msgs_start;     // msgs_prop field start position
  int                   host_pchk;      // checksum of host_prop
  int                   msgs_pchk;      // checksum of msgs_prop
  InMessage             stream;         // Message stream of the message
  OutMessage            last_prop;      // Proposal section: last complete view
  OutMessage            host_prop;      // Proposal section: view composition (hosts)
  OutMessage            memb_prop;      // Proposal section: view composition (members)
  OutMessage            msgs_prop;      // Proposal section: delivered messages

  /*
   * The full proposal message stream, as returned through the
   * DaemonMsg interface.
   */
  private OutMessage outmsg;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructors
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * Creates a proposal from local delivery
   */
  MsgProp(OutMessage msg)
    throws IOException
  {
    stream = new InMessage(msg);
    //FIXED HEIN: Added the seek here, to ensure that it reads the right stuff!
    stream.seek(START);

    // Read single fields
    int gid   = GroupId.unmarshal(stream);
    cvid      = ViewId.unmarshal(stream);
    pvid      = ViewId.unmarshal(stream);
    last_plen = MessageLen.unmarshal(stream);
    host_plen = MessageLen.unmarshal(stream);
    memb_plen = MessageLen.unmarshal(stream);
    msgs_plen = MessageLen.unmarshal(stream);
    host_pchk = Checksum.unmarshal(stream);
    msgs_pchk = Checksum.unmarshal(stream);

    // Computes starting positions
    last_start = P_DATA;
    host_start = last_start + last_plen;
    memb_start = host_start + host_plen;
    msgs_start = memb_start + memb_plen;
  }
  
  
  /**
   *  Creates a new instance of <code>MsgProp</code> starting from a
   *  m-received message.
   *
   *  @param stream
   *    The m-received message.
   */
  MsgProp(ObjectInput stream)
    throws IOException
  {
    this.stream = (InMessage) stream;
    /*
     * Note that we assume that the provided stream has unmarshalled the
     * gid value before invoking this constructor.  This is since it is
     * m-received through the daemon, and thus differs from the locally
     * delivered proposals (see constructor above).
     */

    // Read single fields
    cvid      = ViewId.unmarshal(stream);
    pvid      = ViewId.unmarshal(stream);
    last_plen = MessageLen.unmarshal(stream);
    host_plen = MessageLen.unmarshal(stream);
    memb_plen = MessageLen.unmarshal(stream);
    msgs_plen = MessageLen.unmarshal(stream);
    host_pchk = Checksum.unmarshal(stream);
    msgs_pchk = Checksum.unmarshal(stream);

    // Computes starting positions
    last_start = P_DATA;
    host_start = last_start + last_plen;
    memb_start = host_start + host_plen;
    msgs_start = memb_start + memb_plen;
  }


  /**
   *  Creates a new instance of <code>MsgProp</code> acting as a
   *  container for a proposal.
   */
  MsgProp()
  {
    //FIXME HEIN: Should these payload values be static?
    last_prop   = new OutMessage(256);
    host_prop   = new OutMessage(256);
    memb_prop   = new OutMessage(256);
    msgs_prop   = new OutMessage(256);
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Encoding methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  static MsgProp unmarshal(ObjectInput stream)
    throws IOException
  {
    return new MsgProp(stream);
  }
  
  
  /**
   *  Encodes the last complete view and stores it into the specified group.
   *
   *  @param    cview   last complete view
   *  @param    length  actual length of cview
   */
  void encodeLastProp(EndPoint[] hosts)
    throws IOException
  {
    last_prop.reset();
    GroupIndex.marshal(last_prop, hosts.length);
    for (int i = 0; i < hosts.length; i++)
      hosts[i].writeExternal(last_prop);
    last_plen = GroupIndex.SIZE + hosts.length * EndPointImpl.SIZE;
  }

  /**
   *  Encodes the next view estimate (hosts) and stores it into the
   *  specified group.
   *
   *  @param     hdata   estimate array
   *  @param     hsize   actuale length of the estimate array
   */
  void encodeHostProp(Collection estimate)
    throws IOException
  {
    Iterator iterator;
    
    host_prop.reset();
    GroupIndex.marshal(host_prop, estimate.size());
    iterator = estimate.iterator();
    while (iterator.hasNext()) {
      HostData scan = (HostData) iterator.next();
      scan.getEndPoint().writeExternal(host_prop);
    }
    iterator = estimate.iterator();
    while (iterator.hasNext()) {
      HostData scan = (HostData) iterator.next();
      IncarnationId.marshal(host_prop, scan.getHost().getIncarnationId());
    }
    iterator = estimate.iterator();
    while (iterator.hasNext()) {
      HostData scan = (HostData) iterator.next();
      VersionId.marshal(host_prop, scan.getAgreed());
    }
    host_pchk = host_prop.computeChecksum();
    host_plen = GroupIndex.SIZE + estimate.size() * 
      (EndPointImpl.SIZE + IncarnationId.SIZE + VersionId.SIZE);
  }


  /**
   *  Encodes the next view estimate (local members) and stores it into the
   *  specified group.
   *
   *  @param members     list of members
   *  @param msize       number of members in the current estimate
   */
  void encodeMembProp(Map members, int size)
    throws IOException
  {
    memb_prop.reset();
    GroupIndex.marshal(memb_prop, size);
    for (Iterator iter = members.values().iterator(); iter.hasNext();) {
      MemberData md = (MemberData) iter.next();
      if (!md.isLeaving() && !md.isCrashed()) {
        md.getMemberId().getLocalId().writeExternal(memb_prop);
      }
    }
    memb_plen = GroupIndex.SIZE + size * LocalId.SIZE;
  }

  /**
   *  Encodes the information about delivered messages and stores it
   *  into the specified group.
   *
   */
  void encodeMsgsProp(int[] delivered)
    throws IOException
  {
    msgs_prop.reset();
    for (int i=0; i < delivered.length; i++) {
      MessageId.marshal(msgs_prop, delivered[i]);
    }
    msgs_pchk = msgs_prop.computeChecksum();
    msgs_plen = delivered.length * MessageId.SIZE;
  }


  /**
   *  Encodes a MsgProp message starting from raw data.
   *
   *  @param     group   object from which raw data are extracted.
   */
  MsgProp marshal(int gid, long cvid, long pvid)
    throws IOException
  {
    // Allocates data buffer
    int size = P_DATA + last_plen + host_plen + memb_plen + msgs_plen;
    /* A proposal is only sent to the coordinator, thus 1 receiver only. */
    outmsg = MsgFactory.get(size, 1);
    outmsg.seek(START);

    GroupId.marshal(outmsg, gid);
    ViewId.marshal(outmsg, cvid);
    ViewId.marshal(outmsg, pvid);
    MessageLen.marshal(outmsg, last_plen);
    MessageLen.marshal(outmsg, host_plen);
    MessageLen.marshal(outmsg, memb_plen);
    MessageLen.marshal(outmsg, msgs_plen);
    Checksum.marshal(outmsg, host_pchk);
    Checksum.marshal(outmsg, msgs_pchk);

    outmsg.write(last_prop);
    outmsg.write(host_prop);
    outmsg.write(memb_prop);
    outmsg.write(msgs_prop);

    return this;
  }

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Equality methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * Compares the estimates part of two instances of MsgProp
   *
   * @param     msg     the reference MsgProp message with which to compare.
   */
  boolean checkEstimate(MsgProp msg)
  {
    if (msg == null || host_plen != msg.host_plen || host_pchk != msg.host_pchk)
      return false;
    return stream.compare(host_start, msg.stream, msg.host_start, host_plen);
  }

  /**
   * Compares the messages part of two instances of MsgProp
   *
   * @param     msg     the reference MsgProp message with which to compare.
   */
  boolean checkMessages(MsgProp msg)
  {
    if (msg == null)
      return false;
    //FIXME CLEANUP debugging code
    if (pvid == msg.pvid) {
      if (msgs_plen != msg.msgs_plen || msgs_pchk != msg.msgs_pchk) {
        Group.log.debug("MSGS NOT EQUAL");
        return false;
      }
      boolean cmp = stream.compare(msgs_start, msg.stream, msg.msgs_start, msgs_plen);
      if (!cmp)
        Group.log.debug("STREAMS NOT EQUAL");
      else
        Group.log.debug("STREAMS EQUAL");
      return cmp;
//      return stream.compare(msgs_start, msg.stream, msg.msgs_start, msgs_plen);
    }
    Group.log.debug("MSGS EQUAL");
    return true;
  }

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Decoding methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Decodes the last complete view.
   */
  EndPoint[] decodeLastProp()
    throws IOException, ClassNotFoundException
  {
    stream.seek(last_start);
    int len = GroupIndex.unmarshal(stream);
    EndPoint[] ret = new EndPoint[len];
    for (int i=0; i < len; i++) {
      ret[i] = new EndPointImpl();
      ret[i].readExternal(stream);
    }
    return ret;
  }

  /**
   *  Decodes the estimate for the next complete view.
   */
  EndPoint[] decodeHostProp_hosts()
    throws IOException, ClassNotFoundException
  {
    stream.seek(host_start);
    int len = GroupIndex.unmarshal(stream);
    EndPoint[] ret = new EndPoint[len];
    for (int i=0; i < len; i++) {
      ret[i] = new EndPointImpl();
      ret[i].readExternal(stream);
    }
    return ret;
  }


  /**
   * Decodes the agreed version numbers for the next complete view.
   */
  int[] decodeHostProp_incarns()
    throws IOException
  {
    stream.seek(host_start);
    int len = GroupIndex.unmarshal(stream);
    int[] ret = new int[len];
    stream.seek(host_start + GroupIndex.SIZE + len * EndPointImpl.SIZE);
    for (int i=0; i < len; i++)
      ret[i] = IncarnationId.unmarshal(stream);
    return ret;
  }


  /**
   * Decodes the agreed version numbers for the next complete view.
   */
  int[] decodeHostProp_agreed()
    throws IOException
  {
    stream.seek(host_start);
    int len = GroupIndex.unmarshal(stream);
    int[] ret = new int[len];
    stream.seek(host_start + GroupIndex.SIZE + len * (EndPointImpl.SIZE + IncarnationId.SIZE));
    for (int i=0; i < len; i++)
      ret[i] = VersionId.unmarshal(stream);
    return ret;
  }

  /**
   * Decodes the list of local members contained in the proposal.
   */
  LocalId[] decodeMembProp()
    throws IOException, ClassNotFoundException
  {
    stream.seek(memb_start);
    int len = GroupIndex.unmarshal(stream);
    LocalId[] ret = new LocalId[len];
    for (int i=0; i < len; i++) {
      ret[i] = new LocalId();
      ret[i].readExternal(stream);
    }
    return ret;
  }

  /**
   * Decodes the list of messages delivered by the member who prepared
   * this proposal.
   */
  private int[] decodeMsgsProp()
    throws IOException
  {
    stream.seek(msgs_start);
    int len = msgs_plen / MessageId.SIZE;
    int[] delivered = new int[len];
    for (int i = 0; i < delivered.length; i++) {
      delivered[i] = MessageId.unmarshal(stream);
    }
    return delivered;
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from DaemonMsg
  ////////////////////////////////////////////////////////////////////////////////////////////

  public int size()
  {
    /* Note; it is ok to use the outmsg size since MsgProp will never be
     * forwarded or resent (it is not queued). */
    return outmsg.size();
  }

  public OutMessage getOutMessage()
  {
    return outmsg;
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from Object
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Returns a string representation of this object
   */
  public String toString()
  {
    StringBuilder buffer = new StringBuilder();
    buffer.append("[Proposal:");
    try {
      EndPoint[] endpoints = decodeHostProp_hosts();
      buffer.append(" hosts=[");
      buffer.append(endpoints.length);
      buffer.append("]{");
      for (int i=0; i < endpoints.length; i++) {
        buffer.append(endpoints[i]);
        if (i < endpoints.length-1)
          buffer.append(", ");
      }
    } catch (Exception e) {
      Abort.exit("Error unmarshalling MsgProp; host proposal host values", e);
    }
    try {
      int[] incarns = decodeHostProp_incarns();
      buffer.append("}, incarns=[");
      buffer.append(incarns.length);
      buffer.append("]{");
      for (int i=0; i < incarns.length; i++) {
        buffer.append(incarns[i]);
        if (i < incarns.length-1)
          buffer.append(", ");
      }
    } catch (Exception e) {
      Abort.exit("Error unmarshalling MsgProp; host proposal incarnation values", e);
    }
    try {
      int[] agreed = decodeHostProp_agreed();
      buffer.append("}, agreed=[");
      buffer.append(agreed.length);
      buffer.append("]{");
      for (int i=0; i < agreed.length; i++) {
        buffer.append(agreed[i]);
        if (i < agreed.length-1)
          buffer.append(", ");
      }
    } catch (Exception e) {
      Abort.exit("Error unmarshalling MsgProp; host proposal agreed values", e);
    }
    try {
      int[] delivered = decodeMsgsProp();
      buffer.append("}, delivered=[");
      buffer.append(delivered.length);
      buffer.append("]{");
      for (int i=0; i < delivered.length; i++) {
        buffer.append(delivered[i]);
        if (i < delivered.length-1)
          buffer.append(", ");
      }
    } catch (Exception e) {
      Abort.exit("Error unmarshalling MsgProp; host proposal agreed values", e);
    }
    buffer.append("}, cvid=" + cvid);
    buffer.append(", pvid=" + pvid);
    buffer.append(", last_plen=" + last_plen);
    buffer.append(", host_plen=" + host_plen);
    buffer.append(", memb_plen=" + memb_plen);
    buffer.append(", msgs_plen=" + msgs_plen);
    buffer.append(", last_start=" + last_start);
    buffer.append(", host_start=" + host_start);
    buffer.append(", memb_start=" + memb_start);
    buffer.append(", msgs_start=" + msgs_start);
    buffer.append(", host_phck=" + host_pchk);
    buffer.append(", msgs_phck=" + msgs_pchk);
    try {
      EndPoint[] endpoints = decodeLastProp();
      buffer.append(", lastprop=[");
      buffer.append(endpoints.length);
      buffer.append("]{");
      for (int i=0; i < endpoints.length; i++) {
        buffer.append(endpoints[i]);
        if (i < endpoints.length-1)
          buffer.append(", ");
      }
    } catch (Exception e) {
      Abort.exit("Error unmarshalling MsgProp; last proposal", e);
    }
    buffer.append("}");
    return buffer.toString();
  }

} // END MsgProp
