/*
 * Copyright (c) 1998-2002 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package jgroup.relacs.daemon;


import java.io.IOException;
import java.io.ObjectInput;
import java.util.Iterator;
import java.util.List;

import jgroup.core.EndPoint;
import jgroup.relacs.mss.MssConstants;
import jgroup.relacs.types.EndPointImpl;
import jgroup.relacs.types.GroupId;
import jgroup.relacs.types.GroupIndex;
import jgroup.relacs.types.VersionId;
import jgroup.util.MsgFactory;
import jgroup.util.OutMessage;


/**
 * The <code> MsgSymm</code> class
 *
 * @author  Alberto Montresor
 * @since   Jgroup 0.1
 */
final class MsgSymm
  implements MssConstants, DaemonMsg
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Position constants
  ////////////////////////////////////////////////////////////////////////////////////////////

  private static final int START  = MSS_HEADER_SIZE;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Unique message instance
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * As these messages are never stored by the daemon, a single message
   * instance is sufficient.  In this way, we save the costs of
   * allocating and garbage collecting messages.
   */
  private static MsgSymm msg = new MsgSymm();

  /** The marshalled message data (also static for same reason as above) */
  private static OutMessage outmsg;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Sender version number */
  int vsend;

  /** Host identifiers */
  EndPoint[] hosts;

  /** Corresponding version numbers */
  int[] version;

  /** Reachable set */
  EndPoint[] rset;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Decoding methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * Creates a new instance of MsgSymm starting from a m-received msg
   * input stream
   */
  static MsgSymm unmarshal(ObjectInput msgstream)
    throws IOException, ClassNotFoundException
  {
    /* Read sender version number */
    msg.vsend = VersionId.unmarshal(msgstream);
    
    /* Read host identifiers and version numbers of the nrset */
    int len  = GroupIndex.unmarshal(msgstream);
    msg.version = new int[len];
    msg.hosts   = new EndPoint[len];
    for (int i=0; i < len; i++) {   
      msg.hosts[i] = new EndPointImpl();
      msg.hosts[i].readExternal(msgstream);
      msg.version[i] = VersionId.unmarshal(msgstream);
    }
    
    /* Read the rset array */
    len = GroupIndex.unmarshal(msgstream);
    msg.rset = new EndPoint[len];
    for (int i=0; i < len; i++) {
      msg.rset[i] = new EndPointImpl();
      msg.rset[i].readExternal(msgstream);
    }
    return msg;
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Encoding methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * Copies informations into Message structure
   */
  static MsgSymm marshal(int gid, int vsend, List nrset, EndPoint[] rset, int nreceivers)
    throws IOException
  {
    /* Allocates data buffer */
    int size = START + GroupId.SIZE + VersionId.SIZE + 2 * GroupIndex.SIZE + nrset.size() * 
               (VersionId.SIZE + EndPointImpl.SIZE) + rset.length * EndPointImpl.SIZE;
    outmsg = MsgFactory.get(size, nreceivers);
    outmsg.seek(START);

    /* Write group identifier and sender version number */
    GroupId.marshal(outmsg, gid);
    VersionId.marshal(outmsg, vsend);

    /*
     * Writes the identifier of the endpoints contained in nrset,
     * together with their version numbers.
     */
    GroupIndex.marshal(outmsg, nrset.size());
    Iterator iterator = nrset.iterator();
    while (iterator.hasNext()) {
      HostData data = (HostData) iterator.next();
      data.getEndPoint().writeExternal(outmsg);
      VersionId.marshal(outmsg, data.getVersion());
    }

    /*
     * Writes the identifier of the endpoints contained in rset
     */
    GroupIndex.marshal(outmsg, rset.length);
    for (int i=0; i < rset.length; i++)
      rset[i].writeExternal(outmsg);

    return msg;
  }


  /**
   * Copies informations into Message structure
   */
  static MsgSymm marshal(int gid, int vsend, HostData dest, EndPoint[] rset, int nreceivers)
    throws IOException
  {
    /* Allocates data buffer */
    int size = START + GroupId.SIZE + VersionId.SIZE + 2 * GroupIndex.SIZE +
               VersionId.SIZE + EndPointImpl.SIZE + rset.length * EndPointImpl.SIZE;
    outmsg = MsgFactory.get(size, nreceivers);
    outmsg.seek(START);

    /* Write group identifier and sender version number */
    GroupId.marshal(outmsg, gid);
    VersionId.marshal(outmsg, vsend);

    /* Write dest identifier and version number */
    GroupIndex.marshal(outmsg, 1);
    dest.getEndPoint().writeExternal(outmsg);
    VersionId.marshal(outmsg, dest.getVersion());

    /*
     * Writes the identifier of the endpoints contained in rset
     */
    GroupIndex.marshal(outmsg, rset.length);
    for (int i=0; i < rset.length; i++)
      rset[i].writeExternal(outmsg);

    return msg;
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from DaemonMsg
  ////////////////////////////////////////////////////////////////////////////////////////////

  public int size()
  {
    /* Note; it is ok to use the outmsg size since MsgSymm will never be
     * forwarded or resent (it is not queued). */
    return outmsg.size();
  }

  public OutMessage getOutMessage()
  {
    return outmsg;
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from Object
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Returns a string representation of this object
   */
  public String toString()
  {
    StringBuilder buffer = new StringBuilder();
    buffer.append("[MsgSymm: vsend=");
    buffer.append(vsend);
    buffer.append(", rset={");
    for (int i=0; i < rset.length; i++) {
      buffer.append(rset[i]);
      if (i < rset.length-1)
        buffer.append(", ");
    }
    buffer.append("}, hosts={");
    for (int i=0; i < hosts.length; i++) {
      buffer.append(hosts[i]);
      if (i < hosts.length-1)
        buffer.append(", ");
    }
    buffer.append("}, versions={");
    for (int i=0; i < version.length; i++) {
      buffer.append(version[i]);
      if (i < version.length-1)
        buffer.append(", ");
    }
    buffer.append("}]");
    return buffer.toString();
  }

} // END MsgSymm
