/*
 * Copyright (c) 1998-2002 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package jgroup.relacs.daemon;

import java.util.Collection;
import java.util.Iterator;
import java.util.Map;

import jgroup.core.EndPoint;
import jgroup.core.View;
import jgroup.relacs.events.DeliveryAck;
import jgroup.relacs.events.DeliveryEvent;
import jgroup.relacs.events.InstallAck;
import jgroup.relacs.events.InstallEvent;
import jgroup.relacs.mss.Mss;
import jgroup.relacs.types.LocalId;
import jgroup.util.InMessage;
import jgroup.util.Queue;
import jgroup.util.Util;

import org.apache.log4j.Logger;

/**
 *  The <code>ViewDescription</code> class describes a view.
 *
 *  @author Alberto Montresor
 *  @since Jgroup 0.3
 */
final class ViewDescription
  implements Tag
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Logger
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Obtain logger for this class */
  private static final Logger log = Logger.getLogger(ViewDescription.class);


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Hosts that are members of this view */
  private HostData[] hosts;

  /** Data associated with hosts included in the current view */
  private HostViewData[] data;
  
  /** Endpoints of the hosts in this view */
  private EndPoint[] endpoints;

  /** 
   * True if the view is stable, i.e. if its current host membership
   * is equal to the current reachability set. If a view is not
   * stable, a new agreement protocol must be launched.
   */
  private boolean stable;
  
  /** True if the view contains local members that have crashed. */
  private boolean containsCrashed;

  /** Members in this view */
  private MemberData[] members;

  /** 
   * Array of boolean flags indicating whether the view described
   * in this section has been installed or not.
   */
  private boolean[] hasInstalled;
  
  /** Position of the local host in the view description */
  private int localIndex;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructors
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Builds a <CODE>ViewDescription</CODE> containing a single member
   *  on a single host.
   *  
   *  @param hd the host to be inserted
   *  @param md the member to be inserted
   */
  ViewDescription(HostData hd, MemberData md)
  {
    hosts = new HostData[] { hd };
    data = new HostViewData[] { new HostViewData(1, 1) };
    endpoints = new EndPoint[] { hd.getEndPoint() };
    members = new MemberData[] { md };
    hasInstalled = new boolean[1];
    localIndex = 0;
    stable = true;
    containsCrashed = false;
  }


  /**
   *  Builds a <CODE>ViewDescription</CODE> containing the hosts
   *  identified by <CODE>endpoints</CODE>.
   *
   */
  ViewDescription(Map thosts, EndPoint[] endpoints, EndPoint[] rset, 
    Map tmembers, LocalId[] localMembers, int localIndex)
  {
    this.stable = true;

    /*
     *  Build host composition 
     */
    this.hosts = new HostData[endpoints.length];
    this.endpoints = new EndPoint[endpoints.length];
    this.data = new HostViewData[endpoints.length];
    this.hasInstalled = new boolean[tmembers.size()];
    for (int i=0; i < hosts.length; i++) {
      this.endpoints[i] = endpoints[i];
      hosts[i] = (HostData) thosts.get(endpoints[i]);
      if (hosts[i] == null)
        throw new IllegalStateException("Error: installing inconsistent view");

      /*
       *  FIXED ALB: tmembers contains the current set of local hosts
       *  while memberIds contains the identifiers of the local
       *  members that have been included in the view. Their size
       *  may differ, as new members may join the group when a view
       *  is installed. Since an HostViewData object contains the
       *  acknowledgement information for the current view, to
       *  initialize it we need to use memberIds.
       *  Previous code:
       *    data[i] = new HostViewData(endpoints.length, tmembers.size());
       */
      data[i] = new HostViewData(endpoints.length, localMembers.length);
      
      if (hosts[i].getVersion() > hosts[i].getAgreed()) {
        if (log.isDebugEnabled())
          log.debug(endpoints[i] + " version " + hosts[i].getVersion()
              + " > agreed " + hosts[i].getAgreed() + "(unstable view)");
        stable = false;
      }
    }

    for (Iterator iterator = thosts.values().iterator(); iterator.hasNext(); ) {
      HostData host = (HostData) iterator.next();
      EndPoint endpoint = host.getEndPoint();
      if (!host.isLeaving() && Util.in(rset, endpoint)) {
        boolean found = false;
        for (int i=0; i < endpoints.length && !found; i++) {
          if (endpoints[i].equals(endpoint)) {
            found = true;
          }
        }
        if (!found) {
          if (log.isDebugEnabled())
            log.debug(endpoint + " excluded, but is in the reachable set (unstable view)");
          stable = false;
        }
      }
      if (log.isDebugEnabled()) {
        log.debug(endpoint + ": is " + (host.isLeaving() ? "" : "not ")
          + "leaving, and is " + (Util.in(rset, endpoint) ? "" : "not ")
          + "in the reachable set.");
      }
    }
    if (log.isDebugEnabled()) {
      StringBuilder buf = new StringBuilder("endpoints: ");
      for (int i = 0; i < endpoints.length; i++) {
        buf.append(endpoints[i]);
        buf.append(" ");
      }
      log.debug(buf);
      buf = new StringBuilder(" reachset: ");
      for (int i = 0; i < rset.length; i++) {
        buf.append(rset[i]);
        buf.append(" ");
      }
      log.debug(buf);
    }

    /*
     *  Build member composition 
     */
    this.members = new MemberData[localMembers.length];
    for (int i=0; i < localMembers.length; i++) {
      members[i] = (MemberData) tmembers.get(localMembers[i]);
      if (members[i] == null)
        throw new IllegalStateException("Error: installing inconsistent view");
    }

    this.localIndex = localIndex;
  }

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Returns the number of hosts composing this view.
   */
  int hostSize()
  {
    return hosts.length;
  }
  
  /**
   *  Returns the number of members composing this view.
   */
  int memberSize()
  {
    return members.length;
  }
  
  /**
   *  Returns the i-th host of the view 
   */
  HostData getHost(int index)
  {
    return hosts[index];
  }
  
  /**
   *  Returns the i-th member of the view 
   */
  MemberData getMember(int index)
  {
    return members[index];
  }

  /**
   *  Returns the set of endpoints composing this view.
   */
  EndPoint[] getEndPoints()
  {
    return endpoints;
  }
  
  /**
   *  Returns the size of this view.
   */
  int size()
  {
    return endpoints.length;
  }

  /**
   * Returns true if the provided view composition equals the
   * view composition represented by this <code>ViewDescription</code>
   * object.
   *
   * @param prev_cvcomp
   *   The previous view composition to compare against this view
   * @return
   *   True if the views are equal, false otherwise.
   */
  boolean isDuplicate(EndPoint[] prev_cvcomp)
  {
    if (endpoints.length == prev_cvcomp.length) {
      for (int i = 0; i < prev_cvcomp.length; i++) {
        if (!prev_cvcomp[i].equals(endpoints[i])) {
          return false;
        }
      }
      return true;
    } else {
      return false;
    }
  }

  /**
   * Returns true if this view is stable, i.e. if its current host membership
   * is equal to the current reachability set. If a view is not
   * stable, a new agreement protocol must be launched.
   */
  boolean isStable()
  {
    return stable;
  }
  
  /**
   * Method <code>containsCrashedMembers</code> returns true if some of
   * the members contained in the current view have crashed. 
   * @return true if the current view contains crashed members.
   */
  boolean containsCrashedMembers()
  {
    return containsCrashed;
  }
  
  int getLocalIndex()
  {
    return localIndex;
  }
  
  int getDelivered(int index)
  {
    log.assertLog(data.length > index, "Provided host index=" + index
      + " is greater than data.length=" + data.length);
    return data[index].delivered;
  }
  
  /**
   *  
   */
  int[] getAcks()
  {
    int[] acks = new int[data.length];
    for (int i=0; i < acks.length; i++) {
      acks[i] = data[i].hack[localIndex];
    }
    return acks;
  }

  /**
   *  
   */
  int[] getDelivered(Collection estimate)
  {
    int[] delivered = new int[estimate.size()];
    int i = 0;
    for (Iterator iter = estimate.iterator(); iter.hasNext(); i++) {
      HostData host = (HostData) iter.next();
      if (host.hasValidViewIndex())
        delivered[i] = data[host.getViewIndex()].delivered;
      else
        delivered[i] = 0;
      if (log.isDebugEnabled())
        log.debug(host + ": delivered=" + delivered[i]);
    }
    if (log.isDebugEnabled()) {
      StringBuilder buffer = new StringBuilder();
      buffer.append("Delivered: ");
      for (i=0; i < delivered.length; i++)
        buffer.append(delivered[i] + " ");
      log.debug(buffer.toString());
    }
    return delivered;
  }
    

  /**
   *  Returns true if the message has not already been delivered
   *  locally.
   *  
   *  @param index the index of the sender
   *  @param mid the message identifier
   */
  boolean isNew(int index, int mid)
  {
    return (mid > data[index].hack[localIndex]);
  }


  /**
   *  
   */
  boolean localViewAck(InstallAck ack) 
  {
    boolean viewStable = true;
    hasInstalled[ack.getMemberIndex()] = true;
    for (int i=0; i < members.length && viewStable; i++)
      viewStable = (members[i].isLeaving() || 
        members[i].isCrashed() || hasInstalled[i]);
    return viewStable;
  }

  /**
   *  
   */
  void remoteMessageAck(MsgAck msg)
  {
    /* Sanity check */
    log.assertLog(msg.ack.length == hosts.length, "Received incorrect ACK message");

    /* Check ack */
    for (int i=0; i < hosts.length; i++) {
      if (data[i].hack[msg.hpos] < msg.ack[i]) {
        data[i].hack[msg.hpos] = msg.ack[i];
        data[i].checkAck();
      }
    }
  }

  /**
   *  
   */
  boolean localDeliveryAck(DeliveryAck ack)
  {
    int mid = ack.getMessageId();
    int hostIndex = ack.getHostIndex();
    HostViewData sender = data[hostIndex];
    sender.mack[ack.getMemberIndex()] = mid;
    if (log.isDebugEnabled()) {
      log.debug("DeliveryAck from " + hosts[hostIndex]
          + " with mid=" + mid + ", sender.delivered=" + sender.delivered);
    }
    boolean carryon = true;

    /*
     * Maybe the number of delivered msgs have increased
     */
//FIXME implement buffer instead
    if (mid >= sender.delivered+1) {
      carryon = updateDelivered(sender, hostIndex, mid);
    }
    if (log.isDebugEnabled()) {
      log.debug("carryon=" + carryon);
    }
    return carryon;
  }

  /**
   * 
   * @param sender
   * @param hostIndex
   * @param mid
   */
  private boolean updateDelivered(HostViewData sender, int hostIndex, int mid)
  {
    boolean carryon = true;
    /*
     * Check that we have received an appropriate ACK from all members.
     * Note that, members.length must be is equal to sender.mack.length.
     */
    for (int i=0; i<members.length && carryon; i++)
      carryon = (members[i].isCrashed() || members[i].isLeaving() || sender.mack[i] >= mid);
    if (carryon) {
      sender.delivered = mid;
      if (log.isDebugEnabled())
        log.debug("Updated delivered for " + hosts[hostIndex] + ": " + mid);
      /*
       * Check that we have received an appropriate ACK from all hosts
       */
      for (int i=0; i < data.length && carryon; i++) {
        carryon = (data[i].hack[localIndex] == data[i].delivered);
      }
    }
    return carryon;
  }

  /**
   *  Forward the messages contained in the buffer
   *  to the hosts in the view.
   */
  void forwardMessages(Mss mss)
  {
    /**
     *  If the current view size is equal to one, it contains just the
     *  local host; there is no need to send this message to it
     */
    if (endpoints.length > 1) {
      for (int i = 0; i < data.length; i++) {
        for (Iterator iter = data[i].buffer.iterator(); iter.hasNext(); ) {
          MsgMcast msg = (MsgMcast) iter.next();
          if (log.isDebugEnabled())
            log.debug("Forwarding msg " + msg.mid + " to group members");
          mss.msend(R_FORWARD, msg, endpoints);
        }
      }
    }
  }


  /**
   *  This method performs the local installation of a view.
   *  Returns true if all members have received the InstallEvent;
   *  return false if some of them have crashed. Having received
   *  a InstallEvent does not mean that the view has been installed;
   *  the members will send an InstallAck message when this happens.
   * 
   * @param gid
   *   the identifier of the group for which this view has been generated
   * @param view
   *   the view to be installed
   */
  void notifyView(int gid, View view)
  {
    if (log.isDebugEnabled()) {
      log.debug(view);
    }
    for (int i=0; i < members.length; i++) {
      if (!members[i].isLeaving() && !members[i].isCrashed()) {
        InstallEvent event = new InstallEvent(gid, view, localIndex, i);
        boolean success = DaemonInteraction.addEvent(members[i], event);
        if (!success) {
          members[i].setCrashed();
          containsCrashed = true;
        }
      }
    }
  }


  /**
   *  Performs the local delivery of a multicast message
   */
  boolean notifyMessage(int gid, MsgMcast msg)
  {
    boolean newCrashed = false;
    if (log.isDebugEnabled())
      log.debug("notifyMessage for group: " + gid + ", msg (" + msg.mid + ")" );

    /* Notify the message to the members */
    for (int i=0; i < members.length; i++) {
      if (!members[i].isLeaving() && !members[i].isCrashed()) {
        InMessage inmsg = msg.getInMessage();
        boolean success;
        synchronized (inmsg) {
          DeliveryEvent event = new DeliveryEvent(gid, msg.isObject, msg.mid, 
            msg.hpos, msg.sender, msg.ackr, inmsg);
          success = DaemonInteraction.addEvent(members[i], event);
        }
        if (!success) {
          members[i].setCrashed();
          newCrashed = true;
          containsCrashed = true;          
        } else {
          members[i].setAck(msg.mid);
        }
      }
    }

    /* Update ack data */  
    data[msg.hpos].hack[localIndex] = msg.mid;
    data[msg.hpos].hack[msg.hpos] = msg.mid;
    data[msg.hpos].buffer.insert(msg);
    
    data[msg.hpos].checkAck();
    return newCrashed;
  }


  public String toString()
  {
    StringBuilder b = new StringBuilder("[ViewDescription: ");
    b.append((stable ? "stable" : "not stable"));
    b.append(", hosts={");
    for (int i = 0; i < hosts.length; i++) {
      if (i != 0)
        b.append(", ");
      b.append(hosts[i]);
    }
    b.append("} delivered={");
    for (int i = 0; i < data.length; i++) {
      if (i != 0)
        b.append(", ");
      b.append(data[i].delivered);
    }
    b.append("}, acks={");
    int[] acks = getAcks();
    for (int i = 0; i < acks.length; i++) {
      if (i != 0)
        b.append(", ");
      b.append(acks[i]);
    }
    b.append("}");
    return b.toString();
  }


  private class HostViewData 
  {
    
    /** Hosts acknowledgement */ 
    int[] hack;
  
    /** Members acknowledgement */ 
    int[] mack;
  
    /** Minimal value among host acknowledgement */ 
    int hmin;
  
    /** Messages sent to members */ 
    int delivered;
  
    /** Message buffer */ 
    Queue buffer;

    /**
     *  Builds an HostViewData for a view of size <CODE>size</CODE>.
     */
    HostViewData(int hostSize, int memberSize)
    {
      hack        = new int[hostSize];
      mack        = new int[memberSize];
      hmin        = 0;
      delivered   = 0;
      buffer      = new Queue();
    }
    
    /**
     *
     */
    void checkAck()
    {
      int min = Integer.MAX_VALUE;
      for (int i=0; i < hack.length; i++) 
        if (min > hack[i])
          min = hack[i];
      if (min > hmin) {
        for (int i=hmin; i < min; i++) 
          buffer.removeFirst();
        hmin = min;
      }
    }
    
  } // END HostViewData

} // END ViewDescription
