/*
 * Copyright (c) 1998-2002 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package jgroup.relacs.events;

import java.io.IOException;
import java.io.ObjectInput;
import java.io.ObjectOutput;

import jgroup.relacs.types.GroupIndex;

/**
 *  The <code>DeliveryAck</code> class represents a local message 
 *  exchanged between members and the daemon; it is used by the member 
 *  to notify the daemon that a view has been successfully installed.
 *
 *  @author Alberto Montresor
 *  @since  Jgroup 0.1
 */
public final class InstallAck
  extends Event
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  private static final long serialVersionUID = -6124650592732535593L;

  /** Member position */
  private int memberIndex;
  

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructors
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Default constructor for externalization.
   */
  public InstallAck()
  {
  }

  /**
   *  Creates an <code>InstallAck</code> object storing the local
   *  identifier of the last installed view, and the member's
   *  position index in the array containing the view members.
   * 
   *  @param gid group identifier
   *  @param memberIndex index of the member position in the 
   *    array containing the members of the last installed view
   *  @param lastInstalled local identifier of the last installed
   *    view
   */
  public InstallAck(int gid, int memberIndex)
  {
    super(INSTALL_ACK, gid);
    this.memberIndex = memberIndex;
  }

  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Accessor methods
  ////////////////////////////////////////////////////////////////////////////////////////////
  
  /**
   *  Returns the position index of the member generating this <CODE>InstallAck</CODE>
   *  event in the array containing the members of the last installed view.
   */
  public int getMemberIndex()
  {
    return memberIndex;
  }
  
  
  ////////////////////////////////////////////////////////////////////////////////////////////
  //  Methods from Externalizable
  ////////////////////////////////////////////////////////////////////////////////////////////
  
  /**
   *  Restores the content of this object from the marshalled data contained
   *  in the specified input stream.
   * 
   *  @param in the stream to be read
   */
  public void readExternal(ObjectInput in)
    throws IOException, ClassNotFoundException
  {
    super.readExternal(in);
    memberIndex = GroupIndex.unmarshal(in);
  }
  
  /**
   *  Marshals the content of this object to the specified output stream.
   * 
   *  @param out the stream to be written
   */
  public void writeExternal(ObjectOutput out)
    throws IOException
  {
    super.writeExternal(out);
    GroupIndex.marshal(out, memberIndex);
  }

} // END InstallAck

