/*
 * Copyright (c) 1998-2002 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package jgroup.relacs.events;

import java.io.IOException;
import java.io.ObjectInput;
import java.io.ObjectOutput;

import jgroup.core.View;
import jgroup.relacs.types.GroupIndex;
import jgroup.relacs.types.ViewImpl;

/**
 *  The <code>InstallEvent</code> class represents a local message 
 *  exchanged between members and the daemon; it is used by a daemon 
 *  to notify members about a view to be installed.
 *
 *  @author Alberto Montresor
 *  @since  Jgroup 1.2
 */
public final class InstallEvent
  extends Event
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  private static final long serialVersionUID = -8820607837570501729L;

  /** View to be installed */
  private View view;
  
  /**
   *  Position index of the local host in the array 
   *  containing the hosts of the current view.
   */
  private int hostIndex;
  
  /**
   *  Position index of the member that generated this
   *  event in the array containing the members of the current view.
   */
  private int memberIndex;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructors
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Default constructor for externalization.
   */
  public InstallEvent()
  {
  }

  /**
   *  Constructs a new view to be installed.
   */
  public InstallEvent(int gid, View view, int hostIndex, int memberIndex)
  {
    super(INSTALL_EVENT, gid);
    this.view = view;
    this.hostIndex = hostIndex;
    this.memberIndex = memberIndex;
  }

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Returns the position index of the local host in the array 
   *  containing the hosts of the current view.
   */
  public int getHostIndex()
  {
    return hostIndex;
  }

  /**
   *  Returns the position index of the member that generated this
   *  event in the array containing the members of the current view.
   */
  public int getMemberIndex()
  {
    return memberIndex;
  }
  
  /**
   *  Returns the view to be installed
   */
  public View getView()
  {
    return view;
  }
  
  ////////////////////////////////////////////////////////////////////////////////////////////
  //  Methods from Externalizable
  ////////////////////////////////////////////////////////////////////////////////////////////
  
  /**
   *  Restores the content of this object from the marshalled data contained
   *  in the specified input stream.
   * 
   *  @param in the stream to be read
   */
  public void readExternal(ObjectInput in)
    throws IOException, ClassNotFoundException
  {
    super.readExternal(in);
    memberIndex = GroupIndex.unmarshal(in);
    hostIndex = GroupIndex.unmarshal(in);
    view = new ViewImpl();
    view.readExternal(in);
  }
  
  /**
   *  Marshals the content of this object to the specified output stream.
   * 
   *  @param out the stream to be written
   */
  public void writeExternal(ObjectOutput out)
    throws IOException
  {
    super.writeExternal(out);
    GroupIndex.marshal(out, memberIndex);
    GroupIndex.marshal(out, hostIndex);
    view.writeExternal(out);
  }

} // END InstallEvent
