/*
 * Copyright (c) 1998-2002 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package jgroup.relacs.gm;

import java.rmi.RemoteException;

import jgroup.core.JgroupException;
import jgroup.core.registry.RegistryService;
import jgroup.relacs.config.AppConfig;


/**
 *  The <code>LeaseLayer</code> implements a lease based refresh mechanism
 *  towards the dependable registry.  This layer must be configured within
 *  a server group, allowing each server to perform refresh calls towards
 *  the dependable registry.
 *
 *  @author Henning Hommeland
 *  @author Jo Andreas Lind
 *  @author Hein Meling
 *  @since Jgroup 2.1
 */
public class LeaseLayer
  implements LeaseService, Runnable
{

  //////////////////////////////////////////////////////////////////////////////////////////
  // Private fields
  //////////////////////////////////////////////////////////////////////////////////////////	

  /* The RegistryService of the local group manager */
  private RegistryService registryService;
  
  /* The refresh rate used by this layer to update the dependable registry */
  private long refreshRate = DEFAULT_REFRESH_RATE;


  //////////////////////////////////////////////////////////////////////////////////////////
  // Constructors
  //////////////////////////////////////////////////////////////////////////////////////////	
	
  /**
   *  Initializes a LeaseLayer.
   */
  private LeaseLayer(RegistryService regs)
  {	
    /* Keep the RegistryService for later use. */
    registryService = regs;
  }
  
  
  //////////////////////////////////////////////////////////////////////////////////////////
  // Static Factory
  //////////////////////////////////////////////////////////////////////////////////////////

  public static LeaseLayer getLayer(RegistryService regs)
    throws RemoteException
  {
    return new LeaseLayer(regs);
  }


  //////////////////////////////////////////////////////////////////////////////////////////
  // Methods from LeaseService
  //////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  We obtain the refresh rate for the given server and start
   *  the lease refresh thread.
   */
  public void addListener(Object listener)
  {
    AppConfig app = AppConfig.getApplication(listener);
    refreshRate = app.getIntParam("LeaseRefresh.refreshRate", DEFAULT_REFRESH_RATE);

    /*
     * Create a new thread and start it once the iid in
     * the RegistryLayer is ready.
     */
    Thread thrd = new Thread(this, "LeaseThread");
    thrd.setDaemon(true);
    thrd.start();
  }


  //////////////////////////////////////////////////////////////////////////////////////////
  // Methods from Runnable
  //////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  This method will execute in a new thread, 
   *  started from the constructor. It will preform
   *  a refresh every <i>refreshrate</i> seconds.
   */
  public void run()
  {		
    while (true) {
      /* Run a refresh on the registry */     
      try {
        registryService.refreshLease();
      } catch (RemoteException re) {
        System.out.println(re.getMessage());
      } catch (JgroupException je) {
        System.out.println(je.getMessage());
      }

      /* Sleep for refresh rate milliseconds */
      try {
        Thread.sleep(refreshRate);
      } catch (InterruptedException e) { /* ignore it */ }
    }
  }

} // End LeaseLayer
