/*
 * Copyright (c) 1998-2002 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package jgroup.relacs.gm;

import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.OutputStream;
import java.util.HashMap;
import java.util.Map;

import jgroup.core.JgroupException;
import jgroup.core.MemberId;
import jgroup.core.MembershipService;
import jgroup.core.multicast.AckListener;
import jgroup.core.multicast.ChainIdentifier;
import jgroup.core.multicast.MulticastListener;
import jgroup.core.multicast.MulticastService;
import jgroup.relacs.daemon.DaemonInteraction;
import jgroup.relacs.daemon.MsgMcast;
import jgroup.relacs.events.DeliveryAck;
import jgroup.relacs.events.DeliveryEvent;
import jgroup.relacs.events.Event;
import jgroup.relacs.events.EventTags;
import jgroup.relacs.events.MulticastRequest;
import jgroup.util.InMessage;
import jgroup.util.Network;
import jgroup.util.OutMessage;

import org.apache.log4j.Logger;
import org.apache.log4j.NDC;

/**
 *  The <code>MulticastLayer</code> class implements the multicast service.
 *
 *  @author Alberto Montresor
 *  @author Hein Meling
 *  @since Jgroup 2.2
 */
public final class MulticastLayer
  implements MulticastService, DispatcherListener, EventTags
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Logger
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Obtain logger for this class */
  private static final Logger log = Logger.getLogger(MulticastLayer.class);


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Dynamic map of multicast listeners; protocol name as key. */
  private final Map<String,MulticastListener> multicastListeners =
    new HashMap<String,MulticastListener>();

  /** The membership service */
  private final MembershipService pgms;

  /** Member identifier */
  private final MemberId me;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructor
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Initialize a new <code>MulticastLayer</code> object.
   *
   *  @param pgms
   *    The membership service.
   *  @exception JgroupException
   *    Raised if the distributed system configuration was not available.
   */
  private MulticastLayer(MembershipService pgms)
    throws JgroupException
  {
    this.pgms = pgms;
    this.me   = pgms.getMyIdentifier();
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Static factory
  ////////////////////////////////////////////////////////////////////////////////////////////

  public static MulticastLayer getLayer(DispatcherService ds, MembershipService pgms)
    throws JgroupException
  {
    return new MulticastLayer(pgms);
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Layer interface methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Add a listener object for this layer to provide upcalls to in
   *  response to multicast events.
   */
  public void addListener(Object listener)
  {
    if (listener instanceof MulticastListener) {
      MulticastListener mListener = (MulticastListener) listener;
      String protocol = mListener.getProtocolName();
      Object currentListener = multicastListeners.get(protocol);
      if (currentListener == null)
        multicastListeners.put(protocol, mListener);
      else if (!currentListener.equals(listener))
        throw new IllegalArgumentException("Multiple multicast listeners for the same protocol is illegal: " + protocol);
    } else {
      throw new IllegalArgumentException("Specified listener does not implement a MulticastListener");
    }
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from MulticastService
  ////////////////////////////////////////////////////////////////////////////////////////////


  /* (non-Javadoc)
   * @see jgroup.core.multicast.MulticastService#mcast(java.io.OutputStream, jgroup.core.multicast.AckListener, jgroup.core.multicast.ChainIdentifier)
   */
  public void mcast(OutputStream out, AckListener ackl, ChainIdentifier chId)
    throws JgroupException
  {
    doChecks();
    if (!(out instanceof OutMessage))
      throw new JgroupException("Only OutputStream obtained through getMessage(protocol) can be multicast");

    if (log.isDebugEnabled())
      log.debug("mcast output stream: " + out);
    /* Send the multicast request to the daemon */
    DaemonInteraction.addEvent(
      new MulticastRequest(pgms.getGid(), false, (OutMessage) out, me, ackl, chId));
  }

  /* (non-Javadoc)
   * @see jgroup.core.multicast.MulticastService#mcast(java.io.OutputStream, jgroup.core.multicast.AckListener)
   */
  public void mcast(OutputStream stream, AckListener ackl)
    throws JgroupException
  {
    doChecks();
    if (!(stream instanceof OutMessage))
      throw new JgroupException("Only OutputStream obtained through getMessage(protocol) can be multicast");

    if (log.isDebugEnabled())
      log.debug("mcast output stream: " + stream);
    /* Send the multicast request to the daemon */
    DaemonInteraction.addEvent(
      new MulticastRequest(pgms.getGid(), false, (OutMessage) stream, me, ackl, null));
  }

  /* (non-Javadoc)
   * @see jgroup.core.multicast.MulticastService#mcast(java.lang.String, java.lang.Object, jgroup.core.multicast.AckListener)
   */
  public void mcast(String protocol, Object obj, AckListener ackl)
    throws JgroupException, IOException
  {
    doChecks();
    if (!multicastListeners.containsKey(protocol))
      throw new UnsupportedOperationException("Unsupported protocol: " + protocol);

    if (log.isDebugEnabled())
      log.debug("mcast object (" + protocol + ")");
    OutMessage bout = (OutMessage) getMessage(protocol);
    ObjectOutputStream out = new ObjectOutputStream(bout);
    out.writeObject(obj);
    out.close();

    /* Send the multicast request to the daemon */
    DaemonInteraction.addEvent(
      new MulticastRequest(pgms.getGid(), true, bout, me, ackl, null));
  }

  /* (non-Javadoc)
   * @see jgroup.core.multicast.MulticastService#getMessage(java.lang.String)
   */
  public OutputStream getMessage(String protocol)
    throws IOException
  {
    if (multicastListeners.isEmpty())
      throw new UnsupportedOperationException("Multicast not supported");
    if (!multicastListeners.containsKey(protocol))
      throw new UnsupportedOperationException("Unsupported protocol: " + protocol);

    OutMessage outmsg = (OutMessage) MsgMcast.createOutputStream();
    outmsg.writeUTF(protocol);
    return outmsg;
  }

  private void doChecks()
    throws JgroupException
  {
    if (multicastListeners.isEmpty())
      throw new UnsupportedOperationException("Multicast not supported");
    if (!pgms.isMemberOrJoining())
      throw new JgroupException("The gm is not member of any group (may be leaving)");
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from DispatcherListener
  ////////////////////////////////////////////////////////////////////////////////////////////

  /* (non-Javadoc)
   * @see jgroup.relacs.gm.DispatcherListener#eventTypes()
   */
  public int[] eventTypes()
  {
    return new int[] {
      DELIVERY_EVENT
    };
  }


  /* (non-Javadoc)
   * @see jgroup.relacs.gm.DispatcherListener#notifyEvent(jgroup.relacs.events.Event)
   */
  public void notifyEvent(Event event)
  {
    if (event.getTag() == DELIVERY_EVENT) {
      handleDeliveryEvent((DeliveryEvent) event);
    }
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Private methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * Delivers the message to the application layer. Note that the
   * byte structure are condivided up by all of the members.
   * 
   * Note that care must be taken regarding the update of state variables in
   * a server (or layer) using both MulticastListener and MembershipListener.
   * Such updates must be syncrhonized since there is no protection from
   * any of these layers.
   */
  private void handleDeliveryEvent(final DeliveryEvent event)
  {
    if (multicastListeners.isEmpty()) { 
      log.warn("Received a multicast message addressed to a non-multicast member");
      return;
    }
    if (log.isDebugEnabled()) {
      NDC.push(Network.getMachineName(event.getSender().getCanonicalHostName()));
      log.debug("Handling " + event);
    }

    if (pgms.isMemberOrJoining()) {
      final InMessage msg = event.getPayload();
      final MemberId sender = event.getSender();
      final int seqNo = event.getMessageId();
      try {
        Object result = null;
        if (event.isObject()) {
          ObjectInputStream in = new ObjectInputStream(msg);
          String protocol = in.readUTF();
          MulticastListener mcastListener = getListener(protocol);
          Object obj = in.readObject();
          result = mcastListener.deliverObject(obj, sender, seqNo);
        } else {
          msg.mark(0);
          String protocol = msg.readUTF();
          MulticastListener mcastListener = getListener(protocol);
          result = mcastListener.deliverStream(msg, sender, seqNo);
        }
        if (!event.isAckRequired())
          result = null;

        /*
         * Returns a local ack to the daemon, to confirm that the
         * message has been delivered to the application. Note that
         * this ack is a local ack for the daemon, different from
         * the remote ack that later may be sent to the other members
         * of the group, depending whether the ack is required or not.
         */
        DeliveryAck deliveryAck = new DeliveryAck(
          pgms.getGid(), event.getHostIndex(), pgms.getMemberIndex(), pgms.getViewIndex(),
          seqNo, sender, result, me);
        DaemonInteraction.addEvent(deliveryAck);

      } catch (Exception e) {
        /*
         * An exception was thrown during message delivery; we just
         * print a warning message together with the exception.
         */ 
        log.warn("Exception caught during message unmarshaling", e);
      } finally {
        if (log.isDebugEnabled()) {
          log.debug("handleDeliveryEvent end");
          NDC.pop();
        }
      }
    }
  }


  /**
   * Returns the <code>MulticastListener</code> for the given protocol so that
   * it can be invoked appropriately.
   * 
   * @param protocol the protocol whose listener to obtain.
   * @return the <code>MulticastListener</code> for the provided protocol.
   * @throws JgroupException if the provided protocol has no registered listener.
   */
  private MulticastListener getListener(String protocol)
    throws JgroupException
  {
    if (log.isDebugEnabled())
      log.debug("Looking for protocol: " + protocol);
    MulticastListener mcastListener = multicastListeners.get(protocol);
    if (mcastListener == null) {
      throw new JgroupException("Received message with unknown protocol: " + protocol);
    } else {
      return mcastListener;
    }
  }

} // END MulticastLayer
