/*
 * Copyright (c) 1998-2002 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package jgroup.relacs.gm;

import java.io.Externalizable;
import java.io.IOException;
import java.io.InputStream;
import java.io.ObjectInput;
import java.io.ObjectInputStream;
import java.io.ObjectOutput;
import java.io.ObjectOutputStream;
import java.io.OutputStream;
import java.rmi.RemoteException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.SortedMap;
import java.util.TreeMap;

import jgroup.core.JgroupException;
import jgroup.core.MemberId;
import jgroup.core.MembershipListener;
import jgroup.core.MembershipService;
import jgroup.core.View;
import jgroup.core.multicast.AckListener;
import jgroup.core.multicast.ChainIdentifier;
import jgroup.core.multicast.MulticastListener;
import jgroup.core.multicast.MulticastService;
import jgroup.relacs.types.EndPointImpl;
import jgroup.relacs.types.MemberIdImpl;
import jgroup.util.InMessage;
import jgroup.util.OutMessage;

import org.apache.log4j.Logger;

/**
 *  The <code>TotalOrderLayer</code> class implements the multicast service,
 *  and uses the underlying reliable multicast service.
 *
 *  Total ordering is accomplished using distributed agreement protocol similar
 *  to that of [Birman and Joseph 1987].
 *
 *  @author Hein Meling
 *  @since Jgroup 2.2
 */
public final class TotalOrderLayer
  implements MulticastService, MulticastListener, MembershipListener
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Logger
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Obtain logger for this class */
  private static final Logger log = Logger.getLogger(TotalOrderLayer.class);


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constants
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Protocol name used to distinguish messages from other protocols. */
  private static final String PROTOCOL_NAME = "Total";

  /** Message identifiers for the distributed agreement protocol */
  private static final int REQUEST_SEQ_NO  = 1;
  private static final int AGREED_SEQ_NO   = 2;

  /** Indicator used to indicate that the stream contains a single object */
  private static final boolean OBJECT_TYPE = true;

  /** Indicator used to indicate that the stream contains a stream */
  private static final boolean STREAM_TYPE = false;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Dynamic map of multicast listeners; protocol name as key. */
  private Map<String,MulticastListener> multicastListeners = new HashMap<String,MulticastListener>();

  /** The multicast service */
  private MulticastService mcast;

  /** This member's identifier */
  private MemberId me;

  /** The largest agreed sequence number */
  private SequenceNo largestAgreedSeqNo;

  /** The largest proposed sequence number */
  private SequenceNo largestProposedSeqNo;

  /** Map of proposals: MessageID -> SequenceNo (Proposal) */
  private Map<MessageID,SequenceNo> proposalsMap =
    Collections.synchronizedMap(new HashMap<MessageID,SequenceNo>());

  /** The hold-back queue of messages (mapping: SequenceNo -> InputStream) */
  private SortedMap<SequenceNo,InputStream> holdBackQueue =
    Collections.synchronizedSortedMap(new TreeMap<SequenceNo,InputStream>());

  /** The delivery queue of messages to be delivered */
//  private List deliveryQueue = Collections.synchronizedList(new LinkedList());


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructor
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Initialize a new <code>TotalOrderLayer</code> object.
   *
   *  @param pgms
   *    The membership service.
   *  @param mcast
   *    The multicast service.
   *  @exception JgroupException
   *    Raised if the distributed system configuration was not available.
   */
  private TotalOrderLayer(MembershipService pgms, MulticastService mcast)
    throws JgroupException
  {
    me = pgms.getMyIdentifier();
    largestAgreedSeqNo = new SequenceNo(me);
    largestProposedSeqNo = new SequenceNo(me);
    this.mcast = mcast;
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Static factory
  ////////////////////////////////////////////////////////////////////////////////////////////

  public static TotalOrderLayer getLayer(MembershipService pgms, MulticastService mcast)
    throws JgroupException
  {
    return new TotalOrderLayer(pgms, mcast);
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Layer interface methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Add a listener object for this layer to provide upcalls to in
   *  response to multicast events.
   */
  public void addListener(Object listener)
  {
    if (listener instanceof MulticastListener) {
      MulticastListener mListener = (MulticastListener) listener;
      String protocol = mListener.getProtocolName();
      Object currentListener = multicastListeners.get(protocol);
      if (currentListener == null)
        multicastListeners.put(protocol, mListener);
      else if (!currentListener.equals(listener))
        throw new IllegalArgumentException("Multiple multicast listeners for the same protocol is illegal: " + protocol);
    } else {
      throw new IllegalArgumentException("Specified listener does not implement a MulticastListener");
    }
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from MembershipListener
  ////////////////////////////////////////////////////////////////////////////////////////////
//FIXME remove this interface.
  public void viewChange(View view)
  {
//    log.debug("View: " + view);
  }

  public void prepareChange()
  {
    log.debug("prepareChange");
  }

  public void hasLeft() { }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from MulticastService
  ////////////////////////////////////////////////////////////////////////////////////////////

  /* (non-Javadoc)
   * @see jgroup.core.multicast.MulticastService#mcast(java.io.OutputStream, jgroup.core.multicast.AckListener, jgroup.core.multicast.ChainIdentifier)
   */
  public void mcast(OutputStream out, AckListener ackl, ChainIdentifier chId)
    throws JgroupException
  {
    throw new UnsupportedOperationException();
  }


  /* (non-Javadoc)
   * @see jgroup.core.multicast.MulticastService#mcast(java.io.OutputStream, jgroup.core.multicast.AckListener)
   */
  public void mcast(OutputStream stream, AckListener ackl)
    throws JgroupException
  {
    ProposalCollector pc = new ProposalCollector();
    mcast.mcast(stream, pc);
    // Wait for all proposals and compute the agreed sequence number
    SequenceNo agreed = pc.getGreatestProposal();

    try {
      /* Send the agreed sequence number to all members */
      OutMessage out = (OutMessage) mcast.getMessage(PROTOCOL_NAME);
      // write common part: message type
      out.writeInt(AGREED_SEQ_NO);
      // write the agreed sequence number
      agreed.writeExternal(out);
      out.flush();
      ResultCollector rc = new ResultCollector(agreed, ackl);
      acklMap.put(agreed, rc);
      // send out the agreed message
      mcast.mcast(out, rc);
    } catch (IOException e) {
      log.warn("Could not send out agreed sequence numbers.", e);
    }
  }

  /* (non-Javadoc)
   * @see jgroup.core.multicast.MulticastService#mcast(java.lang.String, java.lang.Object, jgroup.core.multicast.AckListener)
   */
  public void mcast(String protocol, Object obj, AckListener ackl)
    throws JgroupException, IOException
  {
    OutMessage bout = getMessage(protocol, OBJECT_TYPE);
    ObjectOutputStream out = new ObjectOutputStream(bout);
    out.writeObject(obj);
    out.close();
    mcast(bout, ackl);
  }

  /* (non-Javadoc)
   * @see jgroup.core.multicast.MulticastService#getMessage(java.lang.String)
   */
  public OutputStream getMessage(String protocol)
    throws IOException
  {
    return getMessage(protocol, STREAM_TYPE);
  }

  /**
   * Obtain a message for the given protocol and message type.
   *
   * @see jgroup.core.multicast.MulticastService#getMessage(java.lang.String)
   */
  private OutMessage getMessage(String protocol, boolean msgType)
    throws IOException
  {
    OutMessage msg = (OutMessage) mcast.getMessage(PROTOCOL_NAME);
    msg.writeInt(REQUEST_SEQ_NO);
    /* Create a message id for this message and write it to the stream header */
    MessageID msgId = new MessageID(me);
    msgId.writeExternal(msg);
    msg.writeUTF(protocol);
    msg.writeBoolean(msgType);
    return msg;
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from MulticastListener
  ////////////////////////////////////////////////////////////////////////////////////////////

  /* (non-Javadoc)
   * @see jgroup.core.multicast.MulticastListener#getProtocolName()
   */
  public String getProtocolName()
  {
    return PROTOCOL_NAME;
  }


  /* (non-Javadoc)
   * @see jgroup.core.multicast.MulticastListener#deliverStream(java.io.InputStream)
   */
  public synchronized Object deliverStream(InputStream msg, MemberId sender, int seq) 
  {
    InMessage inmsg = (InMessage) msg;
    try {
      /* Read message type from input stream */
      int messageType = inmsg.readInt();

      /* Multiplex on the various message types */
      switch (messageType) {

        case REQUEST_SEQ_NO: 
        {
          /* Read the message identifier from the input stream */
          MessageID msgId = new MessageID();
          msgId.readExternal(inmsg);
          /* Compute a new proposed sequence number */
          largestProposedSeqNo = SequenceNo.max(largestAgreedSeqNo, largestProposedSeqNo);
          largestProposedSeqNo.sequenceNo++;
          largestProposedSeqNo.setMsgId(msgId);
          proposalsMap.put(msgId, largestProposedSeqNo);
          /*
           * Note that the current position of the inmsg stream is just
           * before the position of the protocol name, allowing the protocol
           * name to be extracted when getting the message back from the
           * holdBackQueue in the delivery phase.
           */
          holdBackQueue.put(largestProposedSeqNo, inmsg);
//          if (log.isDebugEnabled())
//            log.debug("Sending PROPOSAL: " + largestProposedSeqNo + " for msg " + msgId);
          /* Respond with a proposal */
          return largestProposedSeqNo;
        } // case REQUEST_SEQ_NO

        case AGREED_SEQ_NO:
        {
          SequenceNo agreed = new SequenceNo();
          agreed.readExternal(inmsg);
//          if (log.isDebugEnabled())
//            log.debug("Received AGREED: " + agreed + " for msg " + agreed.msgId + " from " + sender);
          largestAgreedSeqNo = SequenceNo.max(largestAgreedSeqNo, agreed);

          // Get my proposal for the msgId currently being ordered
          SequenceNo myProposal = (SequenceNo) proposalsMap.remove(agreed.msgId);
          if (!agreed.equals(myProposal)) {
            // Reorder the hold-back queue
            log.debug("reordering hold-back queue: myProposal=" + myProposal
              + " is different from agreed=" + agreed);
            InputStream stream = holdBackQueue.remove(myProposal);
            if (stream == null) {
              log.warn("MESSAGE NOT FOUND IN HOLD-BACK QUEUE: " + myProposal);
              break;
            }
            holdBackQueue.put(agreed, stream);
          }
          log.debug("holdBackQueue=" + holdBackQueue.keySet());
          log.debug(" proposalsMap=" + proposalsMap);

          List<SequenceNo> resultValues = new ArrayList<SequenceNo>(holdBackQueue.size());
          for (Iterator iter = holdBackQueue.entrySet().iterator(); iter.hasNext();) {
            Map.Entry entry = (Map.Entry) iter.next();
            SequenceNo seqno = (SequenceNo) entry.getKey();
            // if this is a newly agreed seqno, tag the message as deliverable
            if (seqno.equals(agreed))
              seqno.msgId.markAsDeliverable();
            resultValues.add(seqno);
            if (seqno.isDeliverable()) {
//              if (log.isDebugEnabled())
//                log.debug("Delivering message: " + seqno.msgId + ", assigned seqno: " + seqno);
              InMessage stream = (InMessage) entry.getValue();
              iter.remove();
              String protocol = stream.readUTF();
              boolean msgType = stream.readBoolean();
              MulticastListener listener =
                (MulticastListener) multicastListeners.get(protocol);
              MemberId theSender = seqno.msgId.getSender();
              if (msgType == STREAM_TYPE) {
                seqno.result = listener.deliverStream(stream, theSender, seqno.sequenceNo);
              } else {
                seqno.result = listener.deliverObject(stream, theSender, seqno.sequenceNo);
              }
              if (log.isDebugEnabled())
                log.debug("Result of message: " + seqno.msgId + ", assigned seqno: " + seqno);
            } else {
              // Stop moving messages to the delivery queue
              if (log.isDebugEnabled())
                log.debug("Stopping message delivery. " + seqno.msgId + " is undeliverable: " + seqno);
              break;
            }
          }
          log.debug("resultValues.size=" + resultValues.size());
          return resultValues;
        } // case AGREED_SEQ_NO

      } // END of switch (messageType)

    } catch (Exception e) {
      log.warn("Could not read data from stream.", e);
    }
    return null;
  }


  /* (non-Javadoc)
   * @see jgroup.core.multicast.MulticastListener#deliverObject(java.lang.Object)
   */
  public Object deliverObject(Object msg, MemberId id, int seq)
  {
    /*
     * This should never happen since we use the stream method also
     * when dealing with multicasting simple objects.
     */
    throw new UnsupportedOperationException("Should never happen");
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Nested class implementing the AckListener interface
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * An <code>AckListener</code> for collecting proposals for the
   * distributed agreement protocol.
   */
  private class ProposalCollector
    implements AckListener
  {

    /** States */
    private static final int STARTING  = 1;
    private static final int WAITING   = 2;
    private static final int COMPLETED = 3;

    /** Current state */
    private volatile int state = STARTING;

    /** Number of unacknowledge messages */
    private int missing;

    /** Proposals from servers */
    private SequenceNo[] proposals;

    /** Flag completed servers */
    private boolean[] completed;

    /** Update internal state */
    private synchronized void updateState()
    {
      if (missing==0) {
        state = COMPLETED;
        notifyAll();
      }
    }

    /* (non-javadoc)
     * @see jgroup.core.multicast.AckListener#ack(MemberId, int, Object)
     */
    public void ack(MemberId id, int pos, Object obj)
      throws RemoteException
    {
//      if (log.isDebugEnabled())
//        log.debug("Recieved proposal from " + id + ", pos=" + pos + ": " + obj);

      proposals[pos] = (SequenceNo) obj;
      completed[pos] = true;
      missing--;
      updateState();
    }

    /* (non-javadoc)
     * @see jgroup.core.multicast.AckListener#notifyView(View)
     */
    public void notifyView(View view)
      throws RemoteException
    {
      missing = view.size();
      proposals = new SequenceNo[missing];
      completed = new boolean[missing];
      state = WAITING;
    }

    /* (non-javadoc)
     * @see jgroup.core.multicast.AckListener#viewChange()
     */
    public void viewChange()
      throws RemoteException
    {
      log.warn("ProposalCollector: ViewChange during invocation handling");
      //FIXME we should handle this differently.
      missing = 0;
      updateState();
    }

    /**
     * Returns the greateset proposal obtained from the members.
     * This method might return <code>null</code> in case not all
     * members responded with a <i>proposal</i>.
     */
    public synchronized SequenceNo getGreatestProposal()
    {
      while (state != COMPLETED) {
        try { wait(); } catch (InterruptedException iex) {}
      }

      // Select the greatest proposal
      SequenceNo greatestProposal = proposals[0];
      for (int i = 0; i < proposals.length; ++i) {
        if (completed[i]) {
          greatestProposal = SequenceNo.max(proposals[i], greatestProposal);
        } else {
          log.warn("Not all members have provided a PROPOSAL");
          break;
        }
      }
      return greatestProposal;
    }

  } // END ProposalCollector


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Nested class implementing the AckListener interface
  ////////////////////////////////////////////////////////////////////////////////////////////

  private static final Map<SequenceNo,ResultCollector> acklMap =
    new HashMap<SequenceNo,ResultCollector>();

  /**
   * An <code>AckListener</code> for collecting results from multiple
   * totally ordered invocations.
   */
  private class ResultCollector
    implements AckListener
  {

    /** Number of unacknowledge messages */
    private int missing;

    /** Result values from servers */
    private List[] results;

    /** The agreed sequence number associated with this AckListener instance */
    private SequenceNo agreed;

    /** The application level AckListener that is being wrapped */
    private AckListener ackl;


    public ResultCollector(SequenceNo agreed, AckListener ackl)
    {
      this.agreed = agreed;
      this.ackl = ackl;
    }

    /* (non-javadoc)
     * @see jgroup.core.multicast.AckListener#ack(MemberId, int, Object)
     */
    public void ack(MemberId id, int pos, Object obj)
      throws RemoteException
    {
      if (log.isDebugEnabled())
        log.debug("Recieved result from " + id + ", pos=" + pos + ": " + obj);

//      ackl.ack(id, pos, obj);
      List res = (List) obj;
      for (Iterator iter = res.iterator(); iter.hasNext();) {
        SequenceNo seqno = (SequenceNo) iter.next();
        if (seqno.hasResult()) {
          ResultCollector rc = (ResultCollector) acklMap.get(seqno);
          log.debug("seqno=" + seqno + ", agreed=" + agreed);
          rc.missing--;
          if (rc.missing == 0)
            acklMap.remove(seqno);
          log.debug("delivering: " + seqno);
          rc.ackl.ack(id, pos, seqno.result);
        } else {
          log.debug("no result for seqno: " + seqno);
        }
      }
    }

    /* (non-javadoc)
     * @see jgroup.core.multicast.AckListener#notifyView(View)
     */
    public void notifyView(View view)
      throws RemoteException
    {
//      for (Iterator iter = acklMap.values().iterator(); iter.hasNext();) {
//        ResultCollector rc = (ResultCollector) iter.next();
//        rc.ackl.notifyView(view);
//      }
      ackl.notifyView(view);
      missing = view.size();
      results = new List[missing];
    }

    /* (non-javadoc)
     * @see jgroup.core.multicast.AckListener#viewChange()
     */
    public void viewChange()
      throws RemoteException
    {
      if (log.isDebugEnabled())
        log.debug("ResultCollector: ViewChange during invocation handling: ackl="
          + ackl + ", agreed=" + agreed);
      for (Iterator iter = acklMap.values().iterator(); iter.hasNext();) {
        ResultCollector rc = (ResultCollector) iter.next();
        log.debug("rc.agreed=" + rc.agreed);
        rc.ackl.viewChange();
        rc.missing = 0;
      }
//      ackl.viewChange();
//      missing = 0;
    }

  } // END ResultCollector


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Nested static class for total order sequence numbers
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * Sequence number class used to define a total order of messages.
   */
  private static final class SequenceNo
    implements Externalizable, Comparable
  {

    private static final long serialVersionUID = -6394934923836273234L;

    /** The sequence number value */
    private int sequenceNo = 100;

    /** The member that generated this sequence number */
    private MemberId member;

    /**
     * The message identifier associated with this sequence number.
     * Note that the message identifier is not used when comparing
     * sequence numbers or when computing the hash code.
     */
    private MessageID msgId = null;

    /**
     * Result object associated with this sequence number.
     */
    private Object result = null;


    /** Default constructor for externalization */
    public SequenceNo() {}

    public SequenceNo(MemberId me)
    {
      member = me;
    }

    public SequenceNo(SequenceNo oldSeq)
    {
      this.sequenceNo = oldSeq.sequenceNo;
      this.member = oldSeq.member;
      this.msgId = oldSeq.msgId;
    }

    public static SequenceNo max(SequenceNo s1, SequenceNo s2)
    {
      if (s1.compareTo(s2) > 0) {
        return new SequenceNo(s1);
      } else {
        return new SequenceNo(s2);
      }
    }

    /**
     * Set the message identifier associated with this sequence number.
     */
    public void setMsgId(MessageID msgId)
    {
      this.msgId = msgId;
    }

    public boolean hasResult()
    {
      return result != null;
    }

    public boolean isDeliverable()
    {
      return msgId.deliverable;
    }

    public int hashCode()
    {
      return sequenceNo ^ member.hashCode();
    }

    public boolean equals(Object obj)
    {
      if (obj instanceof SequenceNo) {
        SequenceNo sno = (SequenceNo) obj;
        return (sequenceNo == sno.sequenceNo && member.equals(sno.member));
      } else {
        return false;
      }
    }

    public String toString()
    {
      StringBuilder buf = new StringBuilder();
      buf.append("[");
      String a = member.getEndPoint().getAddress().getHostAddress().toString();
      String[] b = a.split("\\.");
      if (b.length > 0)
        buf.append(b[b.length-1]);
      buf.append(".");
      buf.append(sequenceNo);
      if (result == null) {
        buf.append("]");
      } else {
        buf.append("]: ");
        buf.append(result);
      }
      return buf.toString();
    }

    /* (non-Javadoc)
     * @see java.io.Externalizable#readExternal(java.io.ObjectInput)
     */
    public void readExternal(ObjectInput in)
      throws IOException, ClassNotFoundException
    {
      sequenceNo = in.readInt();
      member = new MemberIdImpl();
      member.readExternal(in);
      msgId = new MessageID();
      msgId.readExternal(in);
      ObjectInputStream ois = new ObjectInputStream((InputStream) in);
      result = ois.readObject();
    }

    /* (non-Javadoc)
     * @see java.io.Externalizable#writeExternal(java.io.ObjectOutput)
     */
    public void writeExternal(ObjectOutput out)
      throws IOException
    {
      out.writeInt(sequenceNo);
      member.writeExternal(out);
      msgId.writeExternal(out);
      // HACK required since OutMessage does not support writeObject()
      ObjectOutputStream oos = new ObjectOutputStream((OutputStream) out);
      oos.writeObject(result);
    }

    /**
     * Compare two <code>SequenceNo</code> instances.
     *
     * For equal sequence numbers we use the <code>MemberId</code>
     * to distinguish them.  Otherwise, we use only the sequence
     * number.  That is, the sequence number has precedence when
     * comparing two <code>SequenceNo</code> instances.
     *
     * @see java.lang.Comparable#compareTo(java.lang.Object)
     */
    public int compareTo(Object o)
    {
      if (this == o)
        return 0;
      SequenceNo seqno = (SequenceNo) o;
      if (this.sequenceNo == seqno.sequenceNo)
        return this.member.compareTo(seqno.member);
      else if (this.sequenceNo > seqno.sequenceNo)
        return 1;
      else
        return -1;
    }

  } // END SequenceNo


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Nested static class for message identifiers
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * A message identifier is a unique identifier for messages
   * generated within the system.  This is implement by a locally
   * unique sequence number and a member identifier.
   */
  private static final class MessageID
    implements Externalizable
  {

    private static final long serialVersionUID = -6073546790683304835L;

    private static long nextSequenceNo = 0;
    private long sequenceNo;
    private MemberId memberId;

    /** Set to true when the message with this message can be delivered */
    private transient boolean deliverable = false;


    /** Default constructor for externalization */
    public MessageID() {}

    public MessageID(MemberId id)
    {
      this.sequenceNo = nextSequenceNo++;
      this.memberId = id;
    }

    public void markAsDeliverable()
    {
      deliverable = true;
    }

    public boolean isDeliverable()
    {
      return deliverable;
    }

    public MemberId getSender()
    {
      return memberId;
    }

    public boolean equals(Object obj)
    {
      if (obj instanceof MessageID) {
        MessageID msgId = (MessageID) obj;
        return (sequenceNo == msgId.sequenceNo && memberId.equals(msgId.memberId));
      } else {
        return false;
      }
    }

    public int hashCode()
    {
      return (int)(sequenceNo ^ (sequenceNo >>> 32)) ^ memberId.hashCode();
    }

    public String toString()
    {
      StringBuilder buf = new StringBuilder();
      buf.append("[");
      String a = memberId.getEndPoint().getAddress().getHostAddress().toString();
      String[] b = a.split("\\.");
      if (b.length > 0)
        buf.append(b[b.length-1]);
      buf.append("*");
      buf.append(sequenceNo);
      buf.append("]");
      return buf.toString();
    }

    /* (non-Javadoc)
     * @see java.io.Externalizable#readExternal(java.io.ObjectInput)
     */
    public void readExternal(ObjectInput in)
      throws IOException, ClassNotFoundException
    {
      sequenceNo = in.readLong();
      memberId = new MemberIdImpl();
      memberId.readExternal(in);
    }

    /* (non-Javadoc)
     * @see java.io.Externalizable#writeExternal(java.io.ObjectOutput)
     */
    public void writeExternal(ObjectOutput out)
      throws IOException
    {
      out.writeLong(sequenceNo);
      memberId.writeExternal(out);
    }

  } // END MessageID


  public static void main(String[] args)
  {
    MemberId m1 = new MemberIdImpl(new EndPointImpl(239239, 1239), 12, new EndPointImpl(122323, 2323), 0);
    MemberId m2 = new MemberIdImpl(new EndPointImpl(339239, 1239), 12, new EndPointImpl(132323, 2323), 0);
    MemberId m3 = new MemberIdImpl(new EndPointImpl(439239, 1239), 12, new EndPointImpl(142323, 2323), 0);
    SequenceNo s1 = new SequenceNo(m1);
    SequenceNo s2 = new SequenceNo(m2);
    SequenceNo s3 = new SequenceNo(m3);
    s1.sequenceNo++; s2.sequenceNo++;
    System.out.println("s1.compareTo(s1)=" + s1.compareTo(s1));
    System.out.println("s2.compareTo(s1)=" + s2.compareTo(s1));
    System.out.println("s3.compareTo(s1)=" + s3.compareTo(s1));
    System.out.println("s1.compareTo(s2)=" + s1.compareTo(s2));
    System.out.println("s1.compareTo(s3)=" + s1.compareTo(s3));
    System.out.println("s2.compareTo(s3)=" + s2.compareTo(s3));
    System.out.println("s1=" + s1);
    System.out.println("s2=" + s2);
    System.out.println("s3=" + s3);
    System.out.println("max(s1, s2)=s2=" + SequenceNo.max(s1, s2));
    System.out.println("max(s1, s3)=s1=" + SequenceNo.max(s1, s3));
    System.out.println("max(s2, s3)=s2=" + SequenceNo.max(s2, s3));
  }

} // END TotalOrderLayer
