/*
 * Copyright (c) 1998-2003 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */
package jgroup.relacs.gmi;

import java.io.IOException;
import java.io.ObjectInput;
import java.io.ObjectOutput;
import java.lang.reflect.InvocationHandler;
import java.rmi.RemoteException;
import java.util.Iterator;

import jgroup.core.MemberId;
import jgroup.core.registry.DependableRegistry.RegistryEntry;

/**
 * The registry entry type required by the JeriEGMI layer.
 *
 * @author Tor Arve Stangeland
 * @author Hein Meling
 */
public class JeriRegistryEntry
  implements RegistryEntry
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constants
  ////////////////////////////////////////////////////////////////////////////////////////////

  private static final long serialVersionUID = -1665657844632204570L;


  public final static String TYPE = "JeriEGMI";


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructors
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** The member containing the Endpoint */
  private MemberId me;

  /** Service name */
  private String service;

  /** The table of methods associated with the service */
  private MethodTable methodTable;
  

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructors
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Create empty entry, for externalization */
  public JeriRegistryEntry() {}

  
  /**
   * Create a new <code>JeriRegistryEntry</code> with the given service name
   * and <code>Endpoint</code>.
   * @param service Service name
   * @param me <code>MemberId</code> containing the <code>Endpoint</code> to 
   * communicate through
   * @param methodTable the table of methods associated with the service
   */
  JeriRegistryEntry(String service, MemberId me, MethodTable methodTable)
  {
    this.service = service;
    this.me = me;
    this.methodTable = methodTable;
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Access methods for JeriRegistryEntry
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * Returns the member.
   */
  public MemberId getMember()
  {
    return me;
  }

  /**
   * Returns the service.
   */
  public String getService()
  {
    return service;
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from DependableRegistry.RegistryEntry
  ////////////////////////////////////////////////////////////////////////////////////////////

  /* (non-Javadoc)
   * @see jgroup.core.registry.DependableRegistry.RegistryEntry#getType()
   */
  public String getType()
  {
    return TYPE;
  }


  /* (non-Javadoc)
   * @see jgroup.core.registry.DependableRegistry.RegistryEntry#createInvocationHandler(java.lang.String, java.util.Iterator)
   */
  public InvocationHandler createInvocationHandler(String serviceName, Iterator entries)
    throws RemoteException
  {
    GroupEndPoint gep = new GroupEndPoint(serviceName);
    while (entries.hasNext()) {
      JeriRegistryEntry entry = (JeriRegistryEntry) entries.next();
      gep.addMember(entry.getMember());
    }
    GroupInvocationHandler handler = new GroupInvocationHandler(gep, methodTable);
    return handler;
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from Externalizable inherited through DependableRegistry.RegistryEntry
  ////////////////////////////////////////////////////////////////////////////////////////////

  /* (non-javadoc)
   * @see java.io.Externalizable#readExternal(ObjectInput)
   */
  public void readExternal(ObjectInput in)
    throws IOException, ClassNotFoundException
  {
    service = (String) in.readObject();
    me = (MemberId) in.readObject();
    methodTable = (MethodTable) in.readObject();
  }

  /* (non-javadoc)
   * @see java.io.Externalizable#writeExternal(ObjectOutput)
   */
  public void writeExternal(ObjectOutput out)
    throws IOException
  {
    out.writeObject(service);
    out.writeObject(me);
    out.writeObject(methodTable);
  }

} // END JeriRegistryEntry
