/*
 * Copyright (c) 1998-2006 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package jgroup.relacs.gmi.protocols;

import java.io.IOException;
import java.io.InputStream;
import java.io.ObjectInputStream;
import java.util.Collections;
import java.util.HashMap;
import java.util.Map;

import jgroup.core.IID;
import jgroup.relacs.gm.TimestampService.Timestamp;
import jgroup.relacs.gmi.GroupInvocationDispatcher;
import jgroup.relacs.gmi.InvocationResult;
import jgroup.util.CacheMap;

import org.apache.log4j.Logger;

/**
 * BasicDispatcher is the super class for all protocols, and must be subclassed.
 *
 * @author Hein Meling <hein.meling@uis.no>
 */
public class BasicDispatcher
  implements ProtocolDispatcher
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Logger
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Obtain logger for this class */
  private static final Logger log = Logger.getLogger(BasicDispatcher.class);


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Static fields (common to all protocols)
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Map containing (con)current invocation identifiers */
  private static final Map<Thread,IID> iids =
    Collections.synchronizedMap(new HashMap<Thread,IID>(7));


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Cache of invocation results */
  private final CacheMap<Timestamp,InvocationResult> cache =
    new CacheMap<Timestamp,InvocationResult>();

  /** Method invocation dispatcher */
  private final GroupInvocationDispatcher dispatcher;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructor
  ////////////////////////////////////////////////////////////////////////////////////////////

  public BasicDispatcher(GroupInvocationDispatcher dispatcher)
  {
    this.dispatcher = dispatcher;
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // ProtocolDispatcher interface
  ////////////////////////////////////////////////////////////////////////////////////////////

  /* (non-Javadoc)
   * @see jgroup.relacs.gmi.protocols.ProtocolDispatcher#dispatch(java.io.InputStream)
   */
  public InvocationResult dispatch(InputStream in)
    throws IOException
  {
    try {
      ObjectInputStream oin = new ObjectInputStream(in);
      IID iid = (IID) oin.readObject();
      iids.put(Thread.currentThread(), iid);

      InvocationResult result = checkCache(oin);
      if (result == null) {
        /* No timestamp attached with the invocation; just dispatch it */
        if (log.isDebugEnabled())
          log.debug(iid);
        result = dispatcher.dispatch(oin);
      }
      return result;

    } catch (IOException ioex) {
      log.warn("Failed reading IID or Timestamp object from stream", ioex);
    } catch (ClassNotFoundException cnfex) {
      log.warn("ClassNotFoundException during unmarshalling of IID or Timestamp", cnfex);
    } finally {
      // Always remove the iid, independent of success or failure of the invocation
      iids.remove(Thread.currentThread());
    }
    return null;
  }

  /**
   * Check if the stream contains a timestamp, and if it does check if the
   * timestamp has already been executed locally; in which case we just return
   * the previous result stored in the cache.  Otherwise we either dispatch the
   * invocation locally and store the result in the cache.  <code>null</code> is
   * returned if the stream has no timestamp.
   */
  private InvocationResult checkCache(ObjectInputStream oin)
    throws IOException, ClassNotFoundException
  {
    // FIXME need to clear the cache every once in a while; for instance
    // in cases were a client group's member failed before performing the
    // invocation while the other members did perform it.

    InvocationResult result = null;
    Timestamp ts = (Timestamp) oin.readObject();
    if (ts != null) {
      if (log.isDebugEnabled())
        log.debug(ts.toString());
      /* Check the cache, since the invocation included a timestamp */
      synchronized (this) {
        result = cache.get(ts);
        if (result == null) {
          /* Not in cache; dispatch it */
          result = dispatcher.dispatch(oin);
          /* Insert the result into the cache */
          cache.put(ts, result, ts.viewSize);
          if (log.isDebugEnabled())
            log.debug("Invocation not in cache; inserted it: " + ts.viewSize);
        } else {
          if (log.isDebugEnabled())

            log.debug("Invocation in cache; returning it.");
        }
      }
      if (log.isDebugEnabled()) {
        log.debug("Size of cache: " + cache.size());
        log.debug(cache.toString());
      }
    }
    return result;
  }

  /**
   * The default implementation does not need access to the server object, and
   * hence the call is ignored. Implementations that need access to the server
   * object needs to override this method.
   */
  public void setServer(Object server) { }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Public methods (static)
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * Returns the invocation identifier (<code>IID</code>) for the currently
   * executing method.
   */
  public static IID getIdentifier()
  {
    return iids.get(Thread.currentThread());
  }

} // END BasicDispatcher
