/*
 * Copyright (c) 1998-2006 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package jgroup.relacs.gmi.protocols;

import static jgroup.relacs.gmi.protocols.Leadercast.MessageType.LEADERCAST_MSG;
import static jgroup.relacs.gmi.protocols.Leadercast.MessageType.STATE_MSG;

import java.io.IOException;
import java.io.InputStream;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.OutputStream;

import jgroup.core.JgroupException;
import jgroup.core.MemberId;
import jgroup.core.MembershipService;
import jgroup.core.StateListener;
import jgroup.core.multicast.ChainIdentifier;
import jgroup.core.multicast.MulticastListener;
import jgroup.core.multicast.MulticastService;
import jgroup.relacs.gmi.GroupAckListener;
import jgroup.relacs.gmi.GroupInvocationDispatcher;
import jgroup.relacs.gmi.InvocationResult;
import jgroup.relacs.gmi.MethodSemantics;

import org.apache.log4j.Logger;
import org.apache.log4j.MDC;

/**
 * Leadercast represents a protocol that will always try to invoke the the
 * group leader first.  The group leader will pass the servers post-invocation
 * state to the other members of the group, if the invocation resulted in a
 * state change.  If a client tries to invoke a non-leader (follower) replica,
 * that replica will forward the invocation to the leader, and will also mediate
 * the result back to the client.  In such failover scenarios, the client will
 * invoke the leader directly on the next interaction.  
 *
 * @author Hein Meling <hein.meling@uis.no>
 */
public class Leadercast
  extends BasicDispatcher
  implements MulticastListener
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Logger
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Obtain logger for this class */
  private static final Logger log = Logger.getLogger(Leadercast.class);


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constants
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Message types supported by this protocol */
  enum MessageType { STATE_MSG, LEADERCAST_MSG }

  /** Protocol name for demux in the multicast layer */
  private static final String PROTOCOL_NAME = MethodSemantics.LEADERCAST.toString();


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** The multicast service interface */
  private final MulticastService mcast;

  /** The membership service interface */
  private final MembershipService pgms;

  /** The state listener object */
  private StateListener stateListener;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructor
  ////////////////////////////////////////////////////////////////////////////////////////////

  public Leadercast(GroupInvocationDispatcher dispatcher,
      MulticastService mcast, MembershipService pgms)
  {
    super(dispatcher);
    this.mcast = mcast;
    this.pgms = pgms;
    /*
     * We need to listen to multicast events, and since this is
     * not a layer it will not be handled automatically by the
     * group manager construction mechanism.
     */
    mcast.addListener(this);
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // ProtocolDispatcher interface (overriding BasicDispatcher)
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * Handle inbound request to the local endpoint with <i>leadercast</i>
   * invocation semantics.
   * 
   * If the local endpoint is the group leader, then just compute the result
   * value and return it to the client (GroupInvocationHandler).
   * 
   * Otherwise, if the local endpoint is a follower, then we multicast the
   * request to all group members, allowing the leader to pick it up
   * (@see deliverStream below) and provide a return value back to this
   * endpoint whom will pass it back to the client (GroupInvocationHandler).
   *
   * @see jgroup.relacs.gmi.protocols.ProtocolDispatcher#dispatch(java.io.InputStream)
   */
  public InvocationResult dispatch(InputStream in)
    throws IOException
  {
    if (pgms.isLeader()) {
      /*
       * Note that in this case, the chainId is set to null, since
       * this request was received from a client directly, and not
       * mediated through a multicast from another group member.
       */
      return dispatchAtLeader(in, null);
    } else {
      /*
       * I'm a follower; multicast the request, await result from leader,
       * and pass it back to the client (GroupInvocationHandler).
       */
      return dispatchLeadercast(in);
    }
  }

  
  /**
   * Set the server object for use by the <code>Leadercast</code> protocol.<p>
   * 
   * This allows the leadercast protocol to obtain access to the <code>StateListener</code>
   * interface of the server.  This overrides the empty implementation in
   * <code>BasicDispatcher</code>.
   * 
   * If the provided server does not implement a <code>StateListener</code> interface,
   * the call is ignored.  This allows the server to use the leadercast protocol without
   * having its state (if any) passed to the followers.
   * 
   * @throws IllegalStateException
   *   If the server has already been set.  The method should not be called multiple times.
   *   
   * @see jgroup.relacs.gmi.protocols.ProtocolDispatcher#setServer(java.lang.Object)
   */
  public void setServer(Object server)
  {
    /*
     * This method protects against updating the 'stateListener', and hence
     * we do not need to synchronize on this method.
     */ 
    if (this.stateListener == null) {
      if (server instanceof StateListener) {
        stateListener = (StateListener) server;
      }
    } else {
      throw new IllegalStateException("Multiple state listeners not supported.");
    }
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // MulticastListener methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /* (non-Javadoc)
   * @see jgroup.core.multicast.MulticastListener#getProtocolName()
   */
  public String getProtocolName()
  {
    return PROTOCOL_NAME;
  }

  /* (non-Javadoc)
   * @see jgroup.core.multicast.MulticastListener#deliverStream(java.io.InputStream, jgroup.core.MemberId, int)
   */
  public Object deliverStream(InputStream msg, MemberId sender, int seqNo)
  {
    if (log.isDebugEnabled()) {
      MDC.put("group", "[Group: " + pgms.getGid() + "]");
      log.debug("deliverStream() start");
    }
    try {
      int ordinal = msg.read();
      MessageType msgType = MessageType.values()[ordinal];
      switch (msgType) {

        case LEADERCAST_MSG:
          // Only the leader will process this leadercast (coming from a follower)
          if (pgms.isLeader()) {
            ChainIdentifier chainId = new ChainIdentifier(sender, seqNo);
            return dispatchAtLeader(msg, chainId);
          }
          break;

        case STATE_MSG:
          // Only followers will process state objects
          if (!pgms.isLeader() && stateListener != null) {
            // Read state object from message stream
            Object state = new ObjectInputStream(msg).readObject();
            if (log.isDebugEnabled())
              log.debug("Received state object from leader: " + state);
            // Apply state object to the state listener (server replica)
            stateListener.putState(state);
          }
          break;

        default:
          log.error("Unknown message type: " + msgType);
          break;
      }
    } catch (ArrayIndexOutOfBoundsException e) {
      log.warn("Unknown message type (ordinal from stream is wrong)", e);
    } catch (Exception e) {
      log.warn("Failed to deliver multicast message", e);
    }
    return null;
  }

  /* (non-Javadoc)
   * @see jgroup.core.multicast.MulticastListener#deliverObject(java.lang.Object, jgroup.core.MemberId, int)
   */
  public Object deliverObject(Object msg, MemberId sender, int seqNo)
  {
    throw new UnsupportedOperationException();
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Private methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * Handles inbound requests with <i>leadercast</i> semantics received at a follower.
   * 
   * Note that this should happen rarely, since the client should select the
   * group leader as its communicating endpoint in most cases.  However, in cases
   * where failover is necessary, the endpoint selected by the client may not be
   * the group leader.
   * 
   * The inbound request will be multicasted to all members,
   * but only the leader will perform the request.  Thus only
   * the leader will provide a result.
   * 
   * @param in <code>InputStream</code> to read invocation from
   *
   * @see jgroup.relacs.gmi.JeriEGMILayer#deliverStream(java.io.InputStream)
   */
  private InvocationResult dispatchLeadercast(InputStream in) 
    throws IOException
  {
    OutputStream mout = mcast.getMessage(PROTOCOL_NAME);
    mout.write(LEADERCAST_MSG.ordinal());
    // Write request into multicast output stream
    byte[] buf = new byte[1500];
    int bytesRead;
    while ((bytesRead = in.read(buf)) != -1) {
      mout.write(buf, 0, bytesRead);
    }
    // Do multicast
    GroupAckListener ackListener = new GroupAckListener(this);
    try {
      if (log.isDebugEnabled())
        log.debug("Leadercast: multicasting invocation to group members");
      mcast.mcast(mout, ackListener.getRemoteAckListener());
    } catch (JgroupException jex) {
      log.warn("Leadercast invocation failed; member is not ready", jex);
      return null;
    }
    // Wait for ack listener
    return (InvocationResult) ackListener.getLeaderResult();
  }


  /**
   * Dispatch the received message to the local endpoint, which is the
   * leader replica.  The message may have been received through another
   * mediating member (follower) or directly from the client (if the provided
   * <code>ChainIdentifier</code> is <code>null</code>).
   * 
   * Checks if there is a <code>StateListener</code>, in which case
   * the state is passed on to the follower replicas.  However, only
   * if the state of the leader replica changed as a consequence of
   * the invocation.
   * 
   * @param in
   *   The input stream encapsulating the method invocation
   * @param chainId 
   *   The chain identifyer used to identify that a potential state message
   *   is associated with another multicast.
   * @return
   *   The <code>InvocationResult</code> of the method invocation
   * @throws IOException
   *   Raised if the state broadcast failed
   */
  private InvocationResult dispatchAtLeader(InputStream in, ChainIdentifier chainId)
    throws IOException
  {
    if (log.isDebugEnabled())
      log.debug("dispatchAtLeader()");
    InvocationResult res = null;
    // Avoid broadcasting state, if there is only a single members
    if (stateListener == null || pgms.members() == 1) {
      // Perform the actual method invocation on the local server (BasicDispatcher)
      res = super.dispatch(in);
    } else {
      // Get the state prior to the invocation on the leader
      Object preInvocState = stateListener.getState();
      int preInvocHash = (preInvocState != null) ? preInvocState.hashCode() : 0;
      if (log.isDebugEnabled())
        log.debug(" preHash=" +  preInvocHash + ",  preState=" +  preInvocState);

      // Perform the actual method invocation on the local server (BasicDispatcher)
      res = super.dispatch(in);

      // Get the state after to the invocation
      Object postInvocState = stateListener.getState();
      int postInvocHash = (postInvocState != null) ? postInvocState.hashCode() : 0;
      if (log.isDebugEnabled())
        log.debug("postHash=" + postInvocHash + ", postState=" + postInvocState);

      // Check if the method invocation changed the replica state 
      if (preInvocHash != postInvocHash) {
        if (log.isDebugEnabled())
          log.debug("State has changed at leader; broadcasting to followers:");
        broadcastState(postInvocState, chainId);
      }
    }
    return res;
  }


  /**
   *  Broadcast the state of the leader replica after an invocation
   *  has been completed.
   * 
   *  @param state
   *    The state of the leader replica
   */
  private void broadcastState(Object state, ChainIdentifier chainId) 
    throws IOException
  {
    OutputStream mout = mcast.getMessage(PROTOCOL_NAME);
    mout.write(STATE_MSG.ordinal());
    // Write the state object to the stream
    new ObjectOutputStream(mout).writeObject(state);
    // Do multicast
    GroupAckListener ackListener = new GroupAckListener(this);
    try {
      if (log.isDebugEnabled())
        log.debug("Multicasting state to group members (including leader)");
      mcast.mcast(mout, ackListener.getRemoteAckListener(), chainId);
    } catch (JgroupException jex) {
      log.warn("Multicasting state failed; member is not ready", jex);
    }
    // Wait for ack listener (ignore the result which is null anyway)
    ackListener.getResult();
  }

} // END Leadercast
