/*
 * Copyright (c) 1998-2002 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package jgroup.relacs.mss;

import java.util.HashMap;
import java.util.Iterator;

import jgroup.core.ConfigurationException;
import jgroup.core.EndPoint;
import jgroup.relacs.config.TransportConfig;

import org.apache.log4j.Logger;

/**
 *  The <code> ClusterTable </code> class
 *
 *  @author Salvatore Cammarata
 *  @author Hein Meling
 *  @since Jgroup 1.2
 */
final class ClusterTable
  implements MssConstants, MssTag
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Logger
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Obtain logger for this class */
  private static Logger log = Logger.getLogger(ClusterTable.class);

  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constants
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Initial capacity of the table */
  private final static int CAPACITY = 32;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Mss level distributed system configuration */
  private MssDS mssds;

  /** Configuration parameters */
  private TransportConfig config;

  /** Hash map containing the clusters */
  private HashMap map;
  
  /** Table of clusters */
  private Cluster[] table;

  /** Actual size of the table */
  private int size;

  /** This cluster */
  private static Cluster me;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructors
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Constructs an empty cluster table with an initial
   *  capacity equal to <CODE>CAPACITY</CODE>.
   */
  ClusterTable(MssDS mssds, TransportConfig config)
  {
    this.mssds = mssds;
    this.config = config;
    map = new HashMap();
    table = new Cluster[CAPACITY];
    size = 0;
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Public methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Inserts cluster <code>cluster</code> in the table, unless a cluster 
   *  with the same key is already in the table.
   *
   *  @param cluster
   *    The cluster to be inserted.
   *  @exception ConfigurationException 
   *    If the <CODE>cluster</CODE> is already in the table.
   *  @exception NullPointerException
   *    If <CODE>cluster</CODE> is null.
   */
  void insert(Cluster cluster)
    throws ConfigurationException
  {
    if (cluster == null)
      throw new NullPointerException();
    if (map.containsKey(cluster.getEndPoint()))
      throw new ConfigurationException("Cluster already present in " + cluster.getEndPoint());
    if (cluster.isLocal()) {
      if (me == null) {
        me = cluster;
      } else {
        throw new ConfigurationException("Local cluster already defined: " + cluster.getEndPoint());
      }
    }
    /* If space in the table has been exhausted, we double it */
    if (size >= table.length) {
      doubleCapacity();
    }
    table[size++] = cluster;
    map.put(cluster.getEndPoint(), cluster);
  }


  /**
   *  Returns the number of cluster contained in the cluster table.
   */
  int size()
  {
    return size;
  }


  /**
   *  Returns the <CODE>index</CODE>-th cluster maintained in the cluster
   *  table.
   *  
   *  @param index the index of the cluster to be returned
   *  @exception IndexOutOfBoundException if <CODE>index</CODE> is not a 
   *  valid index
   */
  Cluster get(int index)
  {
    if (index < 0 || index >= size)
      throw new IndexOutOfBoundsException("Index " + index);
    return table[index];
  }


  /**
   *  Returns the cluster reference to which this endpoint is mapped.
   *
   *  @param   EndPoint endpoint
   *  @return  the cluster reference to which this endpoint is mapped.
   */
  Cluster lookup(EndPoint endpoint)
  {
    return (Cluster) map.get(endpoint);
  }


  Iterator iterator()
  {
    return map.values().iterator();
  }


  /**
   *  Returns the local cluster descriptor.
   *
   *  @throws IllegalStateException
   *    Raised if the local cluster has not yet been initialized.
   */
  public static Cluster getLocalCluster()
  {
    if (me == null)
      throw new IllegalStateException("Local cluster not yet initialized in the ClusterTable.");
    return me;
  }


  /**
   *  Reset the message flow of all clusters in this
   *  <code>ClusterTable</code>.
   */
  void resetMsgFlow()
  {
    for (Iterator iter = map.values().iterator(); iter.hasNext(); ) {
      Cluster cluster = (Cluster) iter.next();
      cluster.resetMsgFlow();
    }
  }


  /**
   *  Update the routing table, based on the information stored in the
   *  routing message.
   */
  void updateRoutingTable(MsgRouting msg)
  {
    TopologyEntry[] topologyTable = msg.getTopologyTable();
    for (int i=0; i < topologyTable.length; i++) {
      /*
       * Note that the sending cluster can be obtained from the message,
       * even though the ROUTING message has been forwarded through an
       * intermediate host.  The forwarding mechanism, preserves the
       * sender field of the message, changing only the message tag to
       * FWDROUTING.
       */
      EndPoint sendingCluster = msg.getCluster().getEndPoint();
      mergeRoute(topologyTable[i], sendingCluster);
    }
  }


  /**
   *  Updates the reachability state of the members of each cluster
   *  according to the route information.  Discards any entry with TTL
   *  equal to zero.  If the TTL is equal to the TTLwarning threshold
   *  the cost is increased in order to make is less appetitible.
   */
  void update()
  {
    if (me == null)
      throw new IllegalStateException("Local cluster not yet initialized in the ClusterTable.");

    /*
     * To ensure that the route for the local cluster is always correct,
     * we reset it for every update.  This should actually not be needed
     * for every update (assuming that the route does not change for the
     * wrong reason), since the route for the local cluster should
     * never change.  FIXME HEIN: Move this method to insert() above, and
     * do a complete test to ensure that the local route never changes.
     */
    me.resetRoute();

    /*
     * Update the route and reachability information for each cluster.
     */
    for (int i = 0; i < size; i++) {
      Cluster cluster = table[i];
      //FIXME HEIN: move the code below to a new method in Cluster.
      RoutingTableEntry rt = cluster.getRoutingEntry();
      if (rt.TTL != 0) {
        int TTLwarning = config.getTTLWarning();

        if (--rt.TTL == 0) {
          /* The cluster became unreachable */
          rt.route = rt.key;
          rt.cost = config.getMaxPathLength();
          
          /*
           * Mark reachable members of the cluster as unreachable.
           */
          cluster.setClusterAsUnreachable();

        } else if (rt.TTL == TTLwarning) {
          /*
           * If the cluster route is quite old set the cost to a less
           * attractive value (delta warning value).
           */
          rt.cost = config.getPathWarning();
          /*
           * Set all reachable members of the cluster to a warning
           * state, and also check for unreachable members.
           */
          cluster.setWarning(TTLwarning);

        } else {

          /*
           * Check for previously reachable members of the cluster, and
           * mark them as unreachable.
           */
          cluster.updateReachability();

        }
      } // end if (rt.TTL...)
    } // end for 
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Routing
  ///////////////////////////////////////////////////////////////////////////////////////////

  /**
   *
   */
  private void mergeRoute(TopologyEntry newEntry, EndPoint sendingCluster)
  {
    log.assertLog(newEntry.routing.cost >= 0, "mergeRoute: negative cost");

    if (log.isDebugEnabled())
      log.debug("mergeRoute: start");

    Cluster cluster = (Cluster) map.get(newEntry.routing.key);
    log.assertLog(cluster != null, "Cluster from routing entry not known locally");
    RoutingTableEntry rt = cluster.getRoutingEntry();
    if (newEntry.routing.cost + 1 < rt.cost) {
      if (log.isDebugEnabled()) {
        log.debug("new.cost=" + (newEntry.routing.cost+1) + " < rt.cost=" + rt.cost);
        log.debug("Found a better route for " + cluster + " --> " + newEntry);
      }
    } else if (sendingCluster.equals(rt.route)) {
      // FIXME: metric for current next-hop may have changed 
      if (log.isDebugEnabled())
        log.debug("Sending cluster same as my route: " + sendingCluster + " --> "+ newEntry);
    } else {
      if (log.isDebugEnabled())
        log.debug("mergeRoute: end");
      return;
    }

    /*
     * Update routing information and cost data for this route
     */
    rt.route = sendingCluster;
    rt.TTL = config.getMaxTTL();
    int maxPathLength = config.getMaxPathLength();
    int pathLengthWarning = config.getPathWarning();
    if (newEntry.routing.cost + 1 <= maxPathLength)
      rt.cost = newEntry.routing.cost + 1;
    else
      rt.cost = maxPathLength;

    /*
     * Update the reachability info; currently, only information leading
     * to new reachable members are considered, without taking into
     * consideration problems in reachability; these should be detected
     * by the periodic refresh.
     */
    if (rt.cost == maxPathLength) {
      /*
       *  Not reachable; do nothing
       */
    } else if (rt.cost >= pathLengthWarning)  {
      /*
       * The cluster is badly reachable; update the reachability info.
       */
      if (log.isDebugEnabled()) {
        log.warn(cluster + " is badly reachable - cost=" + rt.cost);
      }
      for (int j = 0; j < newEntry.reachable.length; j++) {
        // check if there are new reachable members
        MssHost reachable = mssds.hostLookup(newEntry.reachable[j]);
        if (!reachable.pingOK(pathLengthWarning)) {
          // if the member was not reachable before
          mssds.setAsReachable(reachable, newEntry.incarnationId[j]);
        }
      }

      /*
       * Set all reachable members of the cluster to the warning state
       * and update the reachability of other cluster members that may be
       * unreachable.
       */
      cluster.setWarning(config.getTTLWarning());

    } else {
      /*
       * The cluster has good reachability
       */
      if (log.isDebugEnabled()) {
        log.debug(cluster + " is well reachable - cost=" + rt.cost);
      }
      for (int j = 0; j < newEntry.reachable.length; j++) {
        MssHost reachable = mssds.hostLookup(newEntry.reachable[j]);
        // check if there are new reachable members 
        if (!reachable.isLocal() && !reachable.pingOK()) 
          // if the member was not reachable before
          mssds.setAsReachable(reachable, newEntry.incarnationId[j]);
      }
    }
    if (log.isDebugEnabled())
      log.debug("mergeRoute: end");
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Private methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Double the capacity of the table
   */
  private void doubleCapacity()
  {
    Cluster[] newtable = new Cluster[table.length*2];
    System.arraycopy(table, 0, newtable, 0, table.length);
    table = newtable;
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from Object
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Returns a string representation of this object
   */
  public String toString()
  {
    StringBuilder buffer = new StringBuilder();
    buffer.append("[ClusterTable: {");
    for (int i = 0; i < size; i++) {
      if (i != 0)
        buffer.append(",");
      buffer.append(table[i]);
    }
    buffer.append("}]");
    return buffer.toString();
  }

} // END ClusterTable
