/*
 * Copyright (c) 1998-2002 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package jgroup.relacs.mss;

import java.io.Externalizable;
import java.io.IOException;
import java.io.ObjectInput;
import java.io.ObjectOutput;

import jgroup.core.EndPoint;
import jgroup.relacs.types.EndPointImpl;
import jgroup.relacs.types.MessageId;

/**
 *  The <code>FCEntry</code> class.
 * 
 *  @author Salvatore Cammarata
 *  @since Jgroup 1.2
 */
public final class FCEntry 
implements Externalizable
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constants
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Size of this object in bytes */
  public static final int SIZE = EndPointImpl.SIZE + MessageId.SIZE;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  The host endpoint with which we associate the last message
   *  received value below.
   */
  EndPoint key;

  /**
   *  The last message received from the host above.
   */
  int lastMsgRcvd;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructors
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Default constructor for externalization. 
   */
  public FCEntry()
  {
  }

  /**
   *  
   */
  FCEntry(EndPoint key, int lastMsgRcvd)
  {
    this.key = key;
    this.lastMsgRcvd = lastMsgRcvd;
  }


  /**
   *  Create a new <code>FCEntry</code> for the specified host, and
   *  initialize it with the last message delivered within the hosts
   *  cluster window.
   *
   *  @param host
   *    The <code>MssHost</code> to associate with this flow control
   *    entry.
   */
  FCEntry(MssHost host)
  {
    key = host.getEndPoint();
    lastMsgRcvd = host.getMsgFlow().clusterWindow.getLastMsgDelivered();
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods for Externalization
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Restores the content of this object from the marshalled data contained
   *  in the specified input stream.
   * 
   *  @param in the stream to be read
   */
  public void readExternal(ObjectInput in) 
    throws IOException, ClassNotFoundException
  {
    key = new EndPointImpl();
    key.readExternal(in);
    lastMsgRcvd = MessageId.unmarshal(in);
  }   

  /**
   *  Marshals the content of this object to the specified output stream.
   * 
   *  @param out the stream to be written
   */
  public void writeExternal(ObjectOutput out)
    throws IOException
  {
    key.writeExternal(out);
    MessageId.marshal(out, lastMsgRcvd);
  }

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from Object
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Returns a string representation of this object
   */
  public String toString() 
  {
    StringBuilder buf = new StringBuilder(" FC: ");
    buf.append(key);
    buf.append(lastMsgRcvd);
    return buf.toString();
  } 

}
