/*
 * Copyright (c) 1998-2002 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package jgroup.relacs.mss;

import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.SortedMap;
import java.util.TreeMap;

import jgroup.relacs.config.TransportConfig;
import jgroup.util.InMessage;

import org.apache.log4j.Logger;

/**
 *  The <code>MsgFlowRcvrSide</code> class handles the structures needed
 *  to maintain reliability and flow control.
 *
 *  @author Salvatore Cammarata
 *  @author Hein Meling
 *  @since Jgroup 1.2
 */
final class MsgFlowRcvrSide 
  implements MssConstants 
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Logger
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Obtain logger for this class */
  private static final Logger log = Logger.getLogger(MsgFlowRcvrSide.class);

  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Local host */
  private MssHost localHost;

  /** The default payload length of each message */
  private int payload;

  /** Right edge of the sliding window */
  private int lastMsgRcvd;

  /** Sliding window */
  SWindow swindow;

  /** 
   *  clusterWindow is used to obtain the minimum lastDelivered(x)
   *  where x is a member of the local cluster. 
   */
  ClusterWindow clusterWindow;

  /** List of nack sent to host */
  private Map<Integer, ScheduledEvent> nacks =
    Collections.synchronizedMap(new HashMap<Integer, ScheduledEvent>());

  /** Queue of the message received and not yet delivered */
  private SortedMap<Integer, FragmentHeader> rcvd = new TreeMap<Integer, FragmentHeader>();

  /** Queue of the reliable message received and not yet delivered */
  //FIXED HEIN: Was not used; consider if it is needed later?!
  //  TreeMap reliableRcvd = new TreeMap();

  /** Queue of the fragments not yet composed */
  private SortedMap<Integer, byte[]> fragments = new TreeMap<Integer, byte[]>();

  /** Total message byte size */
  private int totlen;
  

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructors
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Creates an empty object not initialized 
   */
  MsgFlowRcvrSide(TransportConfig config) 
  {
    this.payload = config.getPayload();
    swindow = new SWindow(config.getMaxWindow(), UNDEF);
  }

  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  The receive window is closed
   */
  void reset()
  {
    reset(UNDEF);
  }

  /**
   *  We initialize the receive window with to given message fragment
   *  identifier, in order to synchronize.
   */
  void reset(int mid) 
  {
    if (log.isDebugEnabled())
      log.debug("Synchronization - lastMsgSent=" + mid);

    lastMsgRcvd = mid;
    swindow.clear(mid);
    /*
     * Note that we cannot initialize the localHost until here, since
     * each host is associated with a MsgFlowRcvrSide object; and the
     * localhost may not be the first one to be initialized.  We do a
     * test here so that we can avoid creating a new ClusterWindow
     * object for each reset invocation; instead, we just clear it.
     */
    if (localHost == null) {
      localHost = HostTable.getLocalHost();
      clusterWindow = new ClusterWindow(localHost.getCluster().size());
    } else {
      clusterWindow.clear();
    }
    clusterWindow.set(localHost.getClusterIndex(), mid);
    nacks.clear();
//     reliableRcvd.clear();
    rcvd.clear();
    fragments.clear();
    totlen = 0;
  }


  /**
   *  Set the last message received and return the old last received
   *  value.
   *
   *  @return
   *    The identifier for old last received message.
   */
  int setLastMsgRcvd(int mid)
  {
    int oldLastRcvd = lastMsgRcvd;
    lastMsgRcvd = mid;
    return oldLastRcvd;
  }


  int getLastMsgRcvd()
  {
    return lastMsgRcvd;
  }


  int getLastMsgDlvr()
  {
    return swindow.lastMsgDlvr;
  }


  /**
   *  Returns true if the specified message has already been delivered;
   *  otherwise false is returned.
   */
  boolean isDelivered(int fid)
  {
    return swindow.lookup(fid);
  }


  /**
   *  Returns true if the specified message has an associated NACK
   *  timeout that has not yet expired; otherwise false is returned.
   */
  boolean hasNACK(int mid)
  {
    return (mid < lastMsgRcvd && mid > swindow.lastMsgDlvr);
  }


  /**
   *  Returns the number of received messages that has not yet been
   *  delivered due to not respecting FIFO-order.  It also updates the
   *  sliding window datastructures accordingly.
   */
  int getDeliveryCount()
  {
    int pos = swindow.shiftFIFO();
    clusterWindow.set(localHost.getClusterIndex(), swindow.lastMsgDlvr);
    return pos;
  }


  /**
   *  Returns true if the specified message is in FIFO order; otherwise
   *  false is returned.  If the message is in FIFO order, it also
   *  updates the sliding window datastructures accordingly.
   */
  boolean isFIFO(int mid)
  {
    if (mid == swindow.lastMsgDlvr + 1) {
      swindow.shift();
      if (lastMsgRcvd < swindow.lastMsgDlvr) 
        lastMsgRcvd = swindow.lastMsgDlvr;
      clusterWindow.set(localHost.getClusterIndex(), mid);
      return true;
    } else {
      return false;
    }
  }


  /**
   *  Returns true if the specified message is not in FIFO order,
   *  causing a gap in the sequence of messages.  This means that the
   *  received message must be greater than all those already delivered.
   */
  boolean isNotFIFO(int mid)
  {
    return (mid > swindow.lastMsgDlvr + 1);
  }


  /**
   *  Returns true if the specified message causes a gap in the sequence
   *  of received messages; otherwise false is returned.
   */
  boolean causeGap(int mid)
  {
    return (mid > lastMsgRcvd);
  }


  /**
   *  Insert the given message in the ordered queue.  Also update the
   *  sliding window with the received messages.
   *
   *  @param msgFrag
   *    The message fragment to insert in the ordered queue.
   */
  void insertMsgFrag(FragmentHeader msgFrag)
  {
    swindow.set(msgFrag.fragId);
    rcvd.put(msgFrag.fragId, msgFrag);
  }


  /**
   *  Remove and return the first message fragment stored in the ordered
   *  queue.
   *
   *  @return 
   *    The first message stored in the ordered queue.
   */
  FragmentHeader removeMsgFrag()
  {
    return rcvd.remove(rcvd.firstKey());
  }


  /**
   *  Add a fragement to the list of fragments composing the message.
   *  If this is the last fragment, the message is composed and
   *  returned back to the caller.
   *
   *  @param msg
   *    The message to be composed.
   *  @return 
   *    Returns the message stream to be m-received by the upper layer
   *    if this was the last fragment of the message, and if it was for
   *    me; otherwise <code>null</code> is returned.
   */
  InMessage addFragment(FragmentHeader header)
  {

    byte[] msgFragment = header.getFragment();
    int fragLen = header.getFragmentLength() - OVERHEAD_SIZE;
    if (!header.isLastFragment()) {

      /*
       * This is not the last fragment; add it to the fragments TreeMap.
       */
      if (log.isDebugEnabled())
        log.debug("NOT the last fragement of msg (" + header.fragId + ")");
      totlen += payload;
      fragments.put(header.fragId, msgFragment);
      return null;

    } else {

      /*
       * Last fragment; compose and deliver
       */
      totlen += fragLen;
      if (log.isDebugEnabled())
        log.debug("LAST FRAGEMENT: frag.len= "+ fragLen
                  + "  total.len= " + totlen + "  payload=" + payload);
      InMessage msgstream = new InMessage(payload, HEADER_SIZE, TRAILER_SIZE, totlen);
      /*
       * For each fragment of the message, reconstruct the message
       * stream in correct (ascending) order of the message
       * identifiers (keys to the <code>fragments</code> TreeMap).
       * After the fragment is inserted into the message stream, it is
       * removed from the iterator (and the underlying TreeMap.)
       */
      for (Iterator iter = fragments.keySet().iterator(); iter.hasNext(); ) {
        Integer fid = (Integer) iter.next();
        byte[] msgFrag = fragments.get(fid);
        msgstream.insert(msgFrag, payload);
        if (log.isDebugEnabled())
          log.debug("inserting msg (" + fid + ") in stream\n" + msgstream);
        iter.remove();
      }

      /* Insert the last fragment into the message stream. */
      msgstream.insert(msgFragment, fragLen);
      if (log.isDebugEnabled())
        log.debug("final msg (" + header.fragId + ") in stream\n" + msgstream);
      totlen = 0;
      return msgstream;

    }
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Event handling methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  ScheduledEvent getScheduledEvent(int mid)
  {
    return nacks.get(mid);
  }


  ScheduledEvent putScheduledEvent(int mid, ScheduledEvent event)
  {
    return nacks.put(mid, event);
  }


  ScheduledEvent removeScheduledEvent(int mid)
  {
    return nacks.remove(mid);
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Object methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  public String toString()
  {
    StringBuilder buf = new StringBuilder("[MsgFlowRcvrSide: lastMsgRcvd=");
    buf.append(lastMsgRcvd);
    buf.append(", ");
    buf.append(swindow);
    buf.append(", ");
    buf.append(clusterWindow);
    return buf.toString();
  }

} // END MsgFlowRcvrSide 
