/*
 * Copyright (c) 1998-2002 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package jgroup.relacs.mss;

import java.io.IOException;

import jgroup.core.EndPoint;
import jgroup.relacs.types.EndPointImpl;
import jgroup.relacs.types.MessageId;
import jgroup.util.InMessage;
import jgroup.util.MsgFactory;
import jgroup.util.OutMessage;

/**
 *  The <code>MsgNACK</code> class is used to indicate that a message
 *  has been lost and must be resent.
 *
 *  Note that this message is currently queued in the Mss, enabling
 *  rescheduling of the NACK message.  This prevents us from using a
 *  single message instance for all NACKs.
 *
 *  @author Hein Meling
 *  @since Jgroup 1.2
 */
final class MsgNACK
  implements Msg, MssConstants, MssTag
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Size
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Message size (1+4+6) bytes */
  private static final int MSG_SIZE = 11;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Message Fields (static part)
  ////////////////////////////////////////////////////////////////////////////////////////////

  /* Type of NACK message (REMOTENACK, SENTNACK) */
  private byte nackType;

  /** Message identifier for the missing message */
  int mid;

  /** Receiver of this NACK message (source of missing message) */
  private MssHost source;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Transient fields (recomputed during unmarshalling)
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Message to be sent */
  private transient OutMessage outmsg;

  /** The message sender (obtained from the fragment header) */
  private transient MssHost sender;

  /** NACK creation time; used to recompute RTT (not used on the receiver side) */
  private transient long start;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Marshalling and unmarshalling methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Marshal an NACK message.
   */
  static MsgNACK marshal(byte nackType, MssHost source, int missingMid)
    throws IOException
  {
    MsgNACK msg  = new MsgNACK();
    msg.start = System.currentTimeMillis();
    msg.sender = HostTable.getLocalHost();
    msg.outmsg = MsgFactory.get(MSG_SIZE);

    msg.nackType = nackType;
    msg.mid = missingMid;
    msg.source = source;

    msg.outmsg.writeByte(nackType);
    MessageId.marshal(msg.outmsg, missingMid);
    source.getEndPoint().writeExternal(msg.outmsg);
    return msg;
  }


  /**
   *  Unmarshal an NACK message from the given stream.
   */
  static MsgNACK unmarshal(InMessage in, FragmentHeader header, MssDS mssds)
    throws IOException, ClassNotFoundException
  {
    MsgNACK msg  = new MsgNACK();
    msg.sender   = header.getSender();
    msg.nackType = in.readByte();
    msg.mid      = MessageId.unmarshal(in);
    EndPoint srcEndPoint = new EndPointImpl();
    srcEndPoint.readExternal(in);
    msg.source = mssds.hostLookup(srcEndPoint);
    return msg;
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Returns the NACK message type that this message represents.  There
   *  are two possible values: REMOTENACK and SENTNACK.
   */
  byte getNACKType()
  {
    return nackType;
  }


  /**
   *  Returns true if this NACK message is a REMOTENACK; false is
   *  returned if it is a SENTNACK.
   */
  boolean isRemoteNACK()
  {
    return (nackType == REMOTENACK);
  }


  /**
   *  Returns the source of the missing message; the source is equal to
   *  the receiver of this NACK message.
   */
  MssHost getSource()
  {
    return source;
  }


  /**
   * Returns the round trip time for this NACK message.  Used on the
   * sender side to recompute the adaptive timeouts.
   */
  int getRTT()
  {
    return (int) (System.currentTimeMillis() - start);
  }


  /**
   *  Returns the sender cluster of this message.
   */
  Cluster getCluster()
  {
    return sender.getCluster();
  }


  /**
   *  Returns true if the sender of this message is the local host.
   */
  boolean isLocal()
  {
    return sender.isLocal();
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fragmentation handling (from Msg interface)
  ////////////////////////////////////////////////////////////////////////////////////////////

  /*
   * The fragmentation iterator instance for this message; it can be
   * reused for multiple iterations over this message.
   */
  private transient MsgFragmentIterator fragIter = null;


  /**
   *  Returns a <code>FragmentIterator</code> for this
   *  <code>MsgNACK</code> object.  This iterator allows to send the
   *  entire message as multiple fragments of specified size (payload).
   *  At the same time, it marks each fragment with a tag and message
   *  identifier provided through the <code>next()</code> method of the
   *  iterator.
   */
  public FragmentIterator iterator(MsgCntrl msgCntrl)
  {
    /* If there is already an iterator for this message, we reuse it. */
    if (fragIter == null) 
      fragIter = new MsgFragmentIterator(this, msgCntrl);
    else 
      fragIter.reset(outmsg);
    return fragIter;
  }



  ////////////////////////////////////////////////////////////////////////////////////////////
  // Msg interface methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Returns the tag associated with this message.
   */
  public byte getTag() 
  {
    return NACK;
  }


  /**
   *  Returns the message identifier for this message.
   */
  public int getMid()
  {
    return mid;
  }


  /**
   *  Returns the sender of this message.
   */
  public MssHost getSender()
  {
    return sender;
  }


  /**
   *  Returns false always, since NACK messages should never be routed.
   */
  public boolean hasToBeRouted()
  {
    return false;
  }


  /**
   *  Returns the message flow controller for the sender side.
   */
  public MsgFlowSndrSide getMsgFlow()
  {
    return sender.getCluster().getMsgFlow();
  }


  /**
   *  Returns the <code>OutMessage</code> associated with this message.
   */
  public OutMessage getOutMessage()
  {
    return outmsg;
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from Object
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Returns a string representation of this object
   */
  public String toString()
  {
    StringBuilder buf = new StringBuilder();
    switch (nackType) {
    case REMOTENACK:
      buf.append("REMOTENACK");
      break;

    case SENTNACK:
      buf.append("SENTNACK");
      break;

    default:
      buf.append("Illegal NACK type: " + nackType);
    }
    buf.append(", missingMid=");
    buf.append(mid);
    buf.append(", source=");
    buf.append(source);
    if (sender != null) {
      buf.append(", sender=");
      buf.append(sender);
    }
    return buf.toString();
  }

} // END MsgNACK
