/*
 * Copyright (c) 1998-2002 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package jgroup.relacs.mss;

import java.io.IOException;
import java.util.Iterator;

import jgroup.core.ConfigurationException;
import jgroup.core.EndPoint;
import jgroup.relacs.config.DistributedSystemConfig;
import jgroup.relacs.config.Domain;
import jgroup.relacs.config.DomainSet;
import jgroup.relacs.config.HostSet;
import jgroup.relacs.config.TransportConfig;

import org.apache.log4j.Logger;

/**
 *  The <code>MssDS</code> object contains the description of the
 *  distributed system in which the distributed computation is carried
 *  out.  A distributed system is described as a collection of domains
 *  and hosts, and may be constructed in two ways: by adding a
 *  collection of hosts to a initially empty distributed system, or by
 *  reading a configuration file specifying the list of hosts and
 *  domains.
 *
 *  @author Salvatore Cammarata
 *  @since Jgroup 1.2
 */
final class MssDS
  implements MssConstants, MssTag
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Logger
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Obtain logger for this class */
  private static final Logger log = Logger.getLogger(MssDS.class);


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Configuration parameters */
  private TransportConfig tconf;

  /** Topology information: host table */
  private HostTable hosttable;

  /** Topology information: cluster table */
  private ClusterTable clustertable;

  /** Topology information: routing table */
  private RoutingTable routingtable;

  /** reachability view of the Distributed System */
  private DSView view;        

  /** reachability view of the Distributed System */
  private DSView upperview;       

  /** Network interface */
  private NI ni;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructors
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Constructs a distributed system whose description is contained in
   *
   */
  MssDS(EventHandler ehandler, DistributedSystemConfig dsc, TransportConfig tconf)
    throws IOException, ConfigurationException
  {
    /* Copy references to other data structures */
    this.tconf = tconf;
    ni = new NI(ehandler, tconf.getPayload() + OVERHEAD_SIZE, tconf.getMulticastTTL());

    /*
     *  Initialize the tables
     */
    hosttable = new HostTable();
    clustertable = new ClusterTable(this, tconf);

    /*
     *  Extracts the hosts from the distributed system configuration and
     *  creates mss tables for hosts and domains (clusters).
     */
    DomainSet domains = dsc.getDomainSet();
    for (Iterator i = domains.iterator(); i.hasNext(); ) {
      Domain domain = (Domain) i.next();

      /*
       * Construct entries for the cluster table, based on the domains
       * parsed from the configuration file.
       */
      Cluster cluster = new Cluster(this, tconf, domain, ni);
      clustertable.insert(cluster);

      HostSet hosts = domain.getHostSet();
      for (Iterator j = hosts.iterator(); j.hasNext(); ) {
        EndPoint hostEndPoint = (EndPoint) j.next();
        /*
         * Creates a <code>MssHost</code> object and insert it in the
         * hosttable and cluster.
         */
        MssHost host = new MssHost(tconf, hostEndPoint, ni, cluster, ehandler);
        cluster.insertMember(host);
        hosttable.insert(host);
      }
    }

    /* Get the local host */
    MssHost me = HostTable.getLocalHost();

    /* Reset message flows of all clusters and hosts */
    clustertable.resetMsgFlow();
    hosttable.resetMsgFlow();

    /* Initialize the routing table */
    routingtable = new RoutingTable(clustertable, tconf.getMaxTTL() - 1);

    /* Creates a new view */
    view = new DSView(hosttable, me);
    upperview = new DSView(hosttable, me);
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Getter methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  
   */
  DSView getView()
  {
    return view;
  }
 
  /**
   *  
   */
  DSView getUpperView()
  {
    return upperview;
  }
 
  /**
   *  
   */
  void setControl(MsgCntrl control)
  {
    for (Iterator iter = clustertable.iterator(); iter.hasNext(); ) {
      Cluster cluster = (Cluster) iter.next();
      cluster.setControl(control);
    }
  }

  /**
   *  
   */
  ClusterTable getClusterTable() 
  {
    return clustertable;
  }


  /**
   *  
   */
  HostTable getHostTable() 
  {
    return hosttable;
  }


  /**
   *  
   */
  MssHost hostLookup(EndPoint endpoint)
  {
    return hosttable.lookup(endpoint);
  }

  /**
   *  
   */
  Cluster clusterLookup(EndPoint endpoint)
  {
    return clustertable.lookup(endpoint);
  }

  /**
   *  
   */
  void doStart()
  {
    ni.doStart();
  }


  /**
   *  Returns the size of the distributed system; that is the total
   *  number of hosts in the system.  This includes all clusters.
   */
  int size()
  {
    return hosttable.size();
  }


  /**
   *  Returns the number of clusters (domains) in the distributed
   *  system.
   */
  int numOfClusters()
  {
    return clustertable.size();
  }


  /**
   *  Returns a RoutingTable object with references to the clusters
   *  contained in the distributed systems; note that modification
   *  in the RoutingTable object are reflected in this data structure
   */
  RoutingTable getRoutingTable()
  {
    boolean changed = routingtable.updateReachability();
    if (log.isDebugEnabled()) {
      if (changed)
        log.debug("Reachability has changed");
      log.debug(routingtable);
    }
    return routingtable;
  }


  /**
   *  Update the routing table with information obtained from the given
   *  routing message.
   */
  void updateRoutingTable(MsgRouting msg)
  {
    if (log.isDebugEnabled()) {
      log.debug(msg);
      /* Print the old routing table */
      log.debug(routingtable);
    }
    view.clear();
    clustertable.updateRoutingTable(msg);
    /*
     * Note that we don't update the local routing table, since this is
     * done in the <code>getRoutingTable()</code> method above.  This
     * saves the cost of updating it, until we need it for a MsgRouting
     * message.
     */
  }


  /**
   *  Update the reachability information of the sender of given
   *  IAMALIVE message.
   */
  void updateReachability(MsgIamAlive msg)
  {
    if (log.isDebugEnabled()) {
      log.debug(msg);
      /* Print the old routing table */
      log.debug(routingtable);
    }
    view.clear();
    MssHost sender = msg.getSender();

    if (!sender.isLocal() && !sender.pingOK()) {
      // if the sender was not reachable before
      setAsReachable(sender, msg.getIncarnationId());
    }
    /*
     * Note that we don't update the local routing table, since this is
     * done in the <code>getRoutingTable()</code> method above.  This
     * saves the cost of updating it, until we need it for a MsgRouting
     * message.
     */
  }

  void update()
  {
    if (log.isDebugEnabled()) {
      /* Print the old routing table */
      log.debug(routingtable);
    }
    view.clear();
    clustertable.update();
    /*
     * Note that we don't update the local routing table, since this is
     * done in the <code>getRoutingTable()</code> method above.  This
     * saves the cost of updating it, until we need it for a MsgRouting
     * message.
     */
  }


  /**
   *  Check if the sender has the same incarnation as previously known.
   *
   *  @param header
   *    The <code>FragmentHeader</code> object to check for same
   *    incarnation identifier.
   *  @return 
   *    True if the message fragment has a new incarnation identifier
   *    for the sender; false otherwise.
   */
  boolean hasNewIncarnation(FragmentHeader header)
  {
    view.clear();
    MssHost sender = header.getSender();
    if (sender.isReachable() && sender.getIncarnationId() != header.getIncarnationId()) {
      setAsUnreachable(sender);
      return true;
    }
    return false;
  }


  /**
   *  Check the <code>FWDROUTING</code> routing message, to see if the
   *  incarnations are equal to the locally known incarnations.
   *
   *  @param msg
   *    The <code>MsgRouting</code> object to check for incarnation id.
   *  @return 
   *    True if there are some new incarnation identifiers that were
   *    previously unknown; false otherwise.
   */
  boolean checkIncarnation(MsgRouting msg)
  {
    /* Flag to indicate if there are new incarnations */
    boolean newInc = false;

    view.clear();
    TopologyEntry[] table = msg.getTopologyTable();
    for (int i = 0; i < table.length; i++) {
      for (int j = 0; j < table[i].reachable.length; j++) {
        MssHost member = hostLookup(table[i].reachable[j]);
        if (!member.isLocal() && member.isReachable() && 
            member.getIncarnationId() != table[i].incarnationId[j]) {
          setAsUnreachable(member);
          newInc = true;
        }
      }
    }
    return newInc;
  }


  /**
   *  Mark the given host as unreachable.  Also update the message flow
   *  data of both sender and receiver side, and notify the upper layer
   *  of a view change, excluding this partiuclar member.
   *
   *  @param member
   *    The host member to mark as unreachable.
   */
  void setAsUnreachable(MssHost member) 
  {
    if (log.isDebugEnabled())
      log.debug("setAsUnreachable: start " + member);

    member.setAsUnreachable();
    Cluster cluster = member.getCluster();
    cluster.decrementReachableCounter();
    view.setAsUnreachable(member);
    upperview.setAsUnreachable(member);

    /*
     * If none of the hosts included in the cluster containing this
     * member is reachable, resets the flow control information for the
     * sender side (cluster).
     */
    if (!cluster.isReachable()) {
      if (!cluster.isLocal()) {
        cluster.resetMsgFlow(UNDEF);
        if (log.isDebugEnabled()) {
          log.debug("setAsUnreachable: reset Msg Flow of " + cluster);
        }
      } else {
        throw new IllegalStateException("Local cluster unreachable!");
      }
    }

    /*
     * If the new unreachable member is in the local cluster, update the
     * control flow information for all hosts.
     */
    if (cluster.isLocal()) {
      for (Iterator iter = hosttable.iterator(); iter.hasNext(); ) {
        MssHost host = (MssHost) iter.next();
        host.getMsgFlow().clusterWindow.set(member.getClusterIndex(), UNDEF);
      }
    }

    /* Reset the receiver side message flow of the unreachable member */
    member.resetMsgFlow();
    if (log.isDebugEnabled()) {
      log.debug("setAsUnreachable: reset Msg Flow of " + member);
    }

    if (log.isDebugEnabled())
      log.debug("setAsUnreachable: end");
  }


  void setAsReachable(MssHost member, int inc) 
  {
    if (log.isDebugEnabled())
      log.debug("setAsReachable: start " + member + " " + inc);

    Cluster cluster = member.getCluster();
    cluster.incrementReachableCounter();
    view.setAsReachable(member);

    /*
     * Note that we don't need to reinitialize the message flows for the
     * cluster and member at this point since that is done when
     * receiving an ASYN message from the member; see Mss.handleSYNMsg()
     * for details.
     */

    /*
     * If the incarnation number is changed, we must notify the upper
     * layer of this fact.
     */
    if (member.getIncarnationId() != inc) {
      view.setNewIncarnation(member);
      upperview.setNewIncarnation(member);
      member.setIncarnationId(inc);
    }
    if (log.isDebugEnabled())
      log.debug("setAsReachable: end");
  }


  /**
   *  This flow control method updates the last message delivered
   *  by the remote cluster in each of the members of the cluster.
   *  The new value is equal to <CODE>lastMsgRcvd</CODE> of the
   *  cluster window.
   */
  FCEntry[] getClusterFCEntry(EndPoint key)
  {
    Cluster cluster = clustertable.lookup(key);
    if (cluster == null) {
      log.debug("clustertable: "+ clustertable);
      log.warn("No cluster exists in cluster table for endpoint: "+ key);
    }

    FCEntry[] fc = new FCEntry[cluster.size()];
    int i = 0;
    for (Iterator iter = cluster.iterator(); iter.hasNext(); ) {
      MssHost host = (MssHost) iter.next();
      fc[i++] = new FCEntry(host);
    }
    return fc;
  }


  /**
   *  This flow control method returns an array of <CODE>FCEntry</CODE>
   *  objects containing the last delivered message of all members
   *  (local or remote). Clearly, for <CODE>member[i] == me</CODE>
   *  this value is equal to <CODE>lastMsgSent</CODE>.
   */
  FCEntry[] getAllFCEntry() 
  {
    int size = hosttable.size();
    FCEntry[] ret = new FCEntry[size];
    for (int i=0; i < size; i++) {
      MssHost host = hosttable.get(i);
      ret[i] = new FCEntry(host.getEndPoint(), host.getMsgFlow().getLastMsgDlvr());
    }
    return ret; 
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from Object
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Returns a string representation of this object
   */
  public String toString()
  {
    return "[MssDS:" + hosttable + ", "+ clustertable + "]";
  }

} // END MssDS
