/*
 * Copyright (c) 1998-2002 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package jgroup.relacs.mss;

import java.util.Random;

/**
 *  Class for generating random and timeout values.
 *
 *  @author Salvatore Cammarata
 *  @author Alberto Montresor
 *  @author Hein Meling
 *  @since Jgroup 1.2
 */
final class RandomGenerator 
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Static section
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Random generator */
  private static final Random random = new Random();


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Probabilistic Parameters
  ////////////////////////////////////////////////////////////////////////////////////////////

  // FIXME Change this static fields
  private static final long RANDOM_SECTOR = 4;

  /* Random part to be used for computing a random value. */
  private static long RANDOM_TIMEOUT;

  /* Fixed part to be used for computing a random value. */
  private static long FIXED_TIMEOUT;

  /* Last NACK deadline; initialized with the current time. */
  private static long lastNackDeadline = System.currentTimeMillis();


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Static methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Generate a positive integer within the integer range.  This is
   *  used to generate message identifiers.
   */
  public static int newNonce()
  {
    return random.nextInt(Integer.MAX_VALUE);
  }


  /**
   *  Compute a routing timeout value based on a random value and the
   *  values previously set using the <code>setRoutingTimeout</code>
   *  method.
   */
  public static long getRoutingTimeout() {
    return (long) (random.nextDouble() * RANDOM_TIMEOUT + FIXED_TIMEOUT);
  }


  /**
   *  Compute routing timeout values based on the given value.
   */
  public static void setRoutingTimeout(long rt)
  {
    RANDOM_TIMEOUT = rt / RANDOM_SECTOR;
    FIXED_TIMEOUT = rt - RANDOM_TIMEOUT;
  }


  /**
   *  Compute a random timeout value based on the given value.
   */
  public static long getRandomTimeout(long rt)
  {
    long randomTimeout = rt / RANDOM_SECTOR;
    long fixedTimeout = rt - randomTimeout;

    double deltaRandom = random.nextDouble() * randomTimeout;
    long deltaNew = (long) (deltaRandom + fixedTimeout);

    long current = System.currentTimeMillis();
    if (deltaNew + current < lastNackDeadline) {
      lastNackDeadline += deltaRandom; 
      deltaNew = lastNackDeadline - current;
    } else {
      lastNackDeadline = current + deltaNew;
    }
    return deltaNew;
  }

} // END RandomGenerator
