/*
 * Copyright (c) 1998-2002 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package jgroup.relacs.mss;

import java.io.Externalizable;
import java.io.IOException;
import java.io.ObjectInput;
import java.io.ObjectOutput;

import jgroup.core.EndPoint;
import jgroup.relacs.types.EndPointImpl;
import jgroup.relacs.types.GroupIndex;
import jgroup.relacs.types.IncarnationId;

/**
 *  The <code>TopologyEntry</code> class
 *
 *  @author Salvatore Cammarata
 *  @author Hein Meling
 *  @since Jgroup 1.2
 */
final class TopologyEntry 
implements Externalizable
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** */
  RoutingTableEntry routing;

  /** */
  EndPoint[] reachable;

  /** */
  int[] incarnationId;

  /** The number of bytes required to marshal this topology entry instance */
  private transient int marshalSize;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructors
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** 
   *  Default constructor for externalization
   */
  TopologyEntry() 
  {
  }


  /**
   *  
   */
  TopologyEntry(RoutingTableEntry rt, MssHost[] members)
  {
    marshalSize = RoutingTableEntry.SIZE + GroupIndex.SIZE
      + members.length*(EndPointImpl.SIZE + IncarnationId.SIZE);
    routing = rt;
    reachable = new EndPoint[members.length];
    incarnationId = new int[members.length];
    for (int i = 0; i < members.length; i++) {
      reachable[i] = members[i].getEndPoint();
      incarnationId[i] = members[i].getIncarnationId();
    }
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Returns the number of bytes required to marshal this topology
   *  entry instance.
   */
  public int getMarshalSize()
  {
    return marshalSize;
  }


  /**
   *  Returns the maximum number of bytes required to marshal a topology
   *  entry instance, when the reachability set includes all hosts in
   *  the system.
   */
  public static int getMaxSize(int maxNumOfHosts)
  {
    return (RoutingTableEntry.SIZE + GroupIndex.SIZE
      + maxNumOfHosts*(EndPointImpl.SIZE + IncarnationId.SIZE));
  }


  /**
   *  Updates the reachability information for the cluster associated
   *  with this <code>TopologyEntry</code> object.
   */
  public boolean updateReachability(MssHost[] members)
  {
    boolean changed = false;
    /*
     * This at least avoids creating new arrays, if the new set is the
     * same size as before.
     */
    if (reachable.length != members.length) {
      reachable = new EndPoint[members.length];
      incarnationId = new int[members.length];
      marshalSize = RoutingTableEntry.SIZE + GroupIndex.SIZE
        + members.length*(EndPointImpl.SIZE + IncarnationId.SIZE);
      changed = true;
    }
    for (int i = 0; i < members.length; i++) {
      reachable[i] = members[i].getEndPoint();
      incarnationId[i] = members[i].getIncarnationId();
    }
    return changed;
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods for externalization
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Restores the content of this object from the marshalled data contained
   *  in the specified input stream.
   * 
   *  @param in the stream to be read
   */
  public void readExternal(ObjectInput in)
    throws IOException, ClassNotFoundException
  {
    routing = new RoutingTableEntry();
    routing.readExternal(in);
    int len = GroupIndex.unmarshal(in);
    reachable = new EndPoint[len];
    incarnationId = new int[len];
    for (int i=0; i<len; i++){
      reachable[i] = new EndPointImpl();
      reachable[i].readExternal(in);  
      incarnationId[i] = IncarnationId.unmarshal(in);
    }
  }   


  /**
   *  Marshals the content of this object to the specified output stream.
   * 
   *  @param out the stream to be written
   */
  public void writeExternal(ObjectOutput out)
    throws IOException
  {
    routing.writeExternal(out);
    GroupIndex.marshal(out, reachable.length);
    for (int i=0; i<reachable.length; i++) {
      reachable[i].writeExternal(out);
      IncarnationId.marshal(out,incarnationId[i]);
    }
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Object methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Returns a string representation of this object
   */
  public String toString() 
  {
    StringBuilder buf = new StringBuilder("[TopologyEntry: routing=");
    buf.append(routing);
    buf.append(", size=");
    buf.append(reachable.length);
    buf.append(", reachable={");
    for (int i = 0; i < reachable.length; i++) {
      buf.append((i != 0) ? ", " : "");
      buf.append(reachable[i]);
      buf.append(":");
      buf.append(incarnationId[i]);
    }
    buf.append("}]");
    return buf.toString();
  } 

} // END TopologyEntry
