/*
 * Copyright (c) 1998-2002 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package jgroup.relacs.mss;

import java.io.IOException;
import java.net.DatagramPacket;
import java.net.DatagramSocket;
import java.net.InetAddress;
import java.net.SocketException;

import jgroup.relacs.simulator.SocketStatusImpl;

import org.apache.log4j.Logger;

/**
 *  This class represents an unreliable socket for sending and receiving
 *  datagram packets.  It can be used to simulate packet loss.
 *
 *  @author Alberto Montresor
 *  @since Jgroup 0.5
 */
final class UnreliableDatagramSocket
  extends DatagramSocket
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Logger
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Obtain logger for this class */
  private static final Logger log = Logger.getLogger(UnreliableDatagramSocket.class);


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  private SocketStatusImpl status;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructors
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Constructs an unreliable datagram socket and binds it to any
   *  available port on the local host machine.
   *
   *  @exception SocketException
   *    If the socket could not be opened, or the socket could not
   *    be bound to the specified local port.
   */
  UnreliableDatagramSocket()
    throws SocketException
  {
    this(0, null);
  }


  /**
   *  Constructs a unreliable datagram socket and binds it to the
   *  specified port on the local host machine.
   *
   *  @param port
   *    Port to use on the local host.
   *  @exception SocketException
   *    If the socket could not be opened, or the socket could not
   *    be bound to the specified local port.
   */
  UnreliableDatagramSocket(int port)
    throws SocketException
  {
    this(port, null);
  }


  /**
   *  Constructs a unreliable datagram socket, bound to the specified
   *  local address, and binds it to the specified port on the local
   *  host machine.  The local port must be between 0 and 65535
   *  inclusive.
   *
   *  @param port
   *    Port to use on the local host.
   *  @param localAddr
   *    The local address to bind the socket to.
   *  @exception SocketException
   *    If the socket could not be opened, or the socket could not
   *    be bound to the specified local port.
   */
  UnreliableDatagramSocket(int port, InetAddress localAddr)
    throws SocketException
  {
    super(port, localAddr);
    status = SocketStatusImpl.instance();
    if (log.isDebugEnabled())
      log.debug("UnreliableDatagramSocket initialized at port=" + port + ", status="+status);
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Socket methods to override
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Sends a datagram packet using this socket.  The specified
   *  <code>DatagramPacket</code> includes information indicating the
   *  data to be sent, its length, the IP address of the remote host,
   *  and the port number on the remote host.
   *  
   *  This method overrides the DatagramSocket implementation, and
   *  checks if the packet receiver address is reachable (through the
   *  partition simulator) from the local host.  If the packet receiver
   *  is unreachable the packet is not sent, otherwise if the receiver
   *  is reachable the packet is sent.
   *
   *  @param packet
   *    The <code>DatagramPacket</code> to be sent.
   *  @exception IOException
   *    If an I/O error occurs.
   *  @see java.net.DatagramPacket
   */
  public void send(DatagramPacket packet)
    throws IOException
  {
    InetAddress receiver = packet.getAddress();
    if (status == null || status.isReliable(receiver)) {
      if (log.isDebugEnabled())
        log.debug("Sending packet to " + receiver.getHostName());
      super.send(packet);
    } else if (log.isDebugEnabled()) {
      log.debug("Discarding packet meant for " + receiver.getHostName());
    }
  }

} // END UnreliableDatagramSocket
