/*
 * Copyright (c) 1998-2002 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package jgroup.relacs.mss;

import java.io.IOException;
import java.net.DatagramPacket;
import java.net.InetAddress;
import java.net.MulticastSocket;
import java.net.SocketException;

import jgroup.relacs.simulator.SocketStatusImpl;

import org.apache.log4j.Logger;

/**
 *  This class represents an unreliable socket for sending and receiving
 *  multicast datagram packets.  It can be used to simulate packet loss.
 *
 *  @author Alberto Montresor
 *  @author Hein Meling
 *  @since Jgroup 0.5
 */
final class UnreliableMulticastSocket
  extends MulticastSocket
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Logger
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Obtain logger for this class */
  private static final Logger log = Logger.getLogger(UnreliableMulticastSocket.class);


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  private final SocketStatusImpl status;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructors
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Constructs an unreliable multicast socket and binds it to any
   *  available port on the local host machine.
   *
   *  @exception SocketException
   *    If the socket could not be opened, or the socket could not
   *    be bound to the specified local port.
   */
  UnreliableMulticastSocket()
    throws IOException
  {
    this(0);
  }


  /**
   *  Constructs a unreliable multicast socket and binds it to the
   *  specified port on the local host machine.  The local port must be
   *  between 0 and 65535 inclusive.
   *
   *  @param port
   *    Port to use on the local host.
   *  @exception SocketException
   *    If the socket could not be opened, or the socket could not
   *    be bound to the specified local port.
   */
  UnreliableMulticastSocket(int port)
    throws IOException
  {
    super(port);
    status = SocketStatusImpl.instance();
    if (log.isDebugEnabled())
      log.debug("UnreliableMulticastSocket initialized at port=" + port + ", status="+status);
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Socket methods to override
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Sends a datagram packet using this socket.  The specified
   *  <code>DatagramPacket</code> includes information indicating the
   *  data to be sent, its length, the IP address of the remote host,
   *  and the port number on the remote host.
   *
   *  Note that in using the partition simulator, this method will
   *  only enable partitioning between hosts with separate multicast
   *  addresses.  The partition pattern file must therefore define the
   *  link probabilities for multicast addresses rather than host
   *  addresses.
   *
   *  @param packet
   *    The <code>DatagramPacket</code> to be sent.
   *  @exception IOException
   *    If an I/O error occurs.
   *  @see java.net.DatagramPacket
   */
  public void send(DatagramPacket packet)
    throws IOException
  {
    // note that this receiver address is a multicast address
    InetAddress receiver = packet.getAddress();
    if (status == null || status.isReliable(receiver)) {
      if (log.isDebugEnabled())
        log.debug("Sending multicast packet to " + receiver.getHostAddress());
      super.send(packet);
    } else if (log.isDebugEnabled()) {
      log.debug("Discarding multicast packet meant for " + receiver.getHostAddress());
    }
  }

  /**
   *  Receive a datagram packet using this socket.  The specified
   *  <code>DatagramPacket</code> should be filled with the received
   *  data, its length, the IP address and port of the sender host.
   *
   *  Given that the partition simulator is activated for the link
   *  between the receiver and the sender, the packet size will be set
   *  to zero, and discarded in <code>MulticastNetworkInterface</code>.
   *
   *  @param packet
   *    The <code>DatagramPacket</code> object to fill with the
   *    received data.
   *  @exception IOException
   *    If an I/O error occurs.
   *  @see java.net.DatagramPacket
   */
  public void receive(DatagramPacket packet)
    throws IOException
  {
    super.receive(packet);
    InetAddress sender = packet.getAddress();
    if (log.isDebugEnabled()) {
      log.debug("Received packet from: " + sender.getHostName());
      log.debug("status=" + status + ", isReliable=" + status.isReliable(sender));
    }
    if (status != null && !status.isReliable(sender)) {
      packet.setLength(0);
      if (log.isDebugEnabled())
        log.debug("Packet size set to zero; for discarding...");
    }
  }

} // END UnreliableMulticastSocket
