/*
 * Copyright (c) 1998-2002 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package jgroup.relacs.registry;

import java.io.Externalizable;
import java.io.IOException;
import java.io.ObjectInput;
import java.io.ObjectOutput;

import jgroup.core.IID;
import jgroup.core.registry.DependableRegistry.RegistryEntry;
import jgroup.relacs.types.IIDImpl;
import jgroup.util.OrderedList;

/**
 *  Objects of this class are used to maintain a set of bindings
 *  associated to one service name. <p>
 *
 *  FIXME HEIN Oct 5, 2002: Currently it extends the OrderedList
 *  class; we should avoid to use OrderedList and use hash tables
 *  instead; no need for ordering.  In this sense, the correct name
 *  for the class should be just Binding.
 *
 *  @author Alberto Montresor
 *  @since  Jgroup 0.4
 */
public final class BindingList
  extends OrderedList
  implements Externalizable
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  private static final long serialVersionUID = -8269573883688508547L;

  /** The name of the service of this binding */
  String servicename;
  
  /** The remote reference entry of the replica of this binding */
  RegistryEntry entry;
  
  /** The implementation class associated to this binding */
  Class serverClass;
  
  /** Timestamp for this node */
  long timestamp;
  
  /** Leasetime for this node */
  long leasetime;
  
  /** IID for this node */
  IID iid;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructors
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Empty constructor needed for construction of an empty
   *  <code>OrderedList</code> list.  Also needed for externalization.
   */
  public BindingList() { }


  /**
   *  Construct a new binding between a service name and a remote
   *  object. Additional information are a unique numeric identifier
   *  for the binding, created progressively by the server to which
   *  this binding is associated.
   * 
   *  @param servicename 
   *    the string identifier of the service
   *  @param entry 
   *    the remote reference entry of the replica of this binding
   *  @param serverClass
   *    the implementation class associated to this binding
   *  @param iid
   *    the binding identifier for this replica
   */
  BindingList(String servicename, RegistryEntry entry,
              Class serverClass, IID iid)
  {
    this(servicename, entry, serverClass, iid, -1);
  }


  /**
   *  Construct a new binding between a service name and a remote
   *  object. Additional information are a unique numeric identifier
   *  for the binding, created progressively by the server to which
   *  this binding is associated.
   * 
   *  This constructor is used in conjunction with the LeaseLayer.
   * 
   *  @param servicename 
   *    the string identifier of the service
   *  @param entry
   *    the remote reference entry of the replica of this binding
   *  @param serverClass
   *    the implementation class associated to this binding
   *  @param iid
   *    the binding identifier for this replica       
   *  @param leasetime
   *    the leasetime associated with this BindingList
   */ 
  BindingList(String servicename, RegistryEntry entry,
              Class serverClass, IID iid, long leasetime)
  {
    // The key is a numeric progressive identifier of the binding.
    this.key          = iid.getCounter();
    this.servicename  = servicename;
    this.entry        = entry;
    this.serverClass  = serverClass;
    this.timestamp    = System.currentTimeMillis();
    this.leasetime    = leasetime;
    this.iid          = iid;
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Object methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Returns a string representation of this object
   */
  public String toString()
  {
    StringBuilder buf = new StringBuilder();
    buf.append("key: ");
    buf.append(key);
    buf.append(", service: ");
    buf.append(servicename);
//    buf.append(", entry: ");
//    buf.append(entry);
//    buf.append(", class: ");
//    buf.append(serverClass.getName());
    return buf.toString();
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Externalization
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Marshals the content of this object to the specified output stream.
   * 
   *  @param out the stream to be written
   */
  public void writeExternal(ObjectOutput out)
    throws IOException
  {
    out.writeInt(key);
    out.writeUTF(servicename);
    out.writeObject(entry);
    out.writeObject(serverClass);
    out.writeLong(leasetime);
    /*
     * Compute the remaining time of the lease for this server
     * (i.e. the delta time) and send this to the other registry
     * replicas.
     */
    out.writeLong(System.currentTimeMillis() - timestamp);
    iid.writeExternal(out);
  }

  /**
   *  Restores the content of this object from the marshalled data contained
   *  in the specified input stream.
   * 
   *  @param in the stream to be read
   */
  public void readExternal(ObjectInput in)
    throws IOException, ClassNotFoundException
  {
    key         = in.readInt();
    servicename = in.readUTF();
    entry       = (RegistryEntry) in.readObject();
    serverClass = (Class) in.readObject();
    leasetime   = in.readLong();
    /*
     * Compute a new timestamp for this server as the delta time plus
     * the current time.  That is, we will give the server a little
     * extra time (the communicatio delay) to perform the refresh.
     */
    timestamp   = in.readLong() + System.currentTimeMillis();
    iid = new IIDImpl();
    iid.readExternal(in);
  }

} // END BindingList
