/*
 * Copyright (c) 1998-2002 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package jgroup.relacs.registry;

import java.rmi.NotBoundException;
import java.rmi.RemoteException;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import jgroup.core.MemberId;
import jgroup.core.VMID;
import jgroup.core.View;
import jgroup.relacs.gm.LeaseInfo;
import jgroup.util.IntList;

import org.apache.log4j.Logger;

/**
 *  This class implements the main data structure used to maintain
 *  the bindings stored in the registry.
 *
 *  @author Alberto Montresor
 *  @author Hein Meling
 *  @since Jgroup 0.5
 */
final class RegistryTable
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Logger
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Obtain logger for this class */
  private static final Logger log = Logger.getLogger(RegistryTable.class);


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** 
   *  This map associates VMIDs (virtual machine identifiers) to
   *  the <CODE>ServerData</CODE> data structure associated to them
   */
  private final Map<VMID, ServerData> servertable = new HashMap<VMID, ServerData>(51);
  
  /** 
   *  This map associates service names (constituted by string identifiers) 
   *  to the <CODE>ServiceData</CODE> data structure associated to them
   */
  private final Map<String, ServiceData> servicetable = new HashMap<String, ServiceData>(51);
  
  /** Last view installed */
  private View view;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructor
  ////////////////////////////////////////////////////////////////////////////////////////////

  RegistryTable() { }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Add a binding for the specified <CODE>VMID</CODE>.
   *  
   *  @param vmid 
   *    the virtual machine identifier to which associate this binding
   *  @param binding
   *    the binding to be associated to the specified <CODE>VMID</CODE>
   *    
   */
  synchronized void add(VMID vmid, BindingList bl)
    throws RemoteException
  {
    // Inserts in service table
    ServiceData service = servicetable.get(bl.servicename);
    if (service == null) {
      service = new ServiceData(bl.serverClass, bl.servicename, bl.entry);
      servicetable.put(bl.servicename, service);
    }
    service.add(bl);

    // Insert in server table
    ServerData server = get(vmid);
    server.add(bl.key, bl);
  }


  /**
   *  Remove a binding associated to the specified <CODE>VMID</CODE>
   *  and identifier by the specified numeric progressive identifier.
   *  
   *  @param vmid 
   *    the vmid associated to the binding to be removed
   *  @param key 
   *    the identifier of the binding
   */
  synchronized void remove(VMID vmid, int key)
  {
    ServerData server = get(vmid);
    BindingList binding = server.remove(key);
    if (binding != null) {
      ServiceData service = servicetable.get(binding.servicename);
      if (service != null) {
        service.remove(binding);
        if (service.isEmpty())
          servicetable.remove(binding.servicename);
      }
    }
  }
  
  /**
   *  Get a binding associated to the specified <CODE>VMID</CODE>
   *  and identifier by the specified numeric progressive identifier.
   * 
   *  @param vmid
   *    the vmid associated to the binding to be fetched
   *  @param key
   *    the identifier of the binding
   *  @return
   *    the BindingList node of the specified server
   */
  public BindingList lookup(VMID vmid, int key)
  {
    ServerData server = get(vmid);
    BindingList binding = server.lookup(key);
    return binding;
  }
  
  /**
   *  Returns the IIDs and timestamps for all servers
   *  registered under the given servicename, as well 
   *  as returning the leasetime for the given service.
   * 
   *  @param servicename
   *    the name of the service
   *  @return
   *    datastructure containing IIDs, timestamps and
   *    leasetime for servers under the given servicename
   */
  public LeaseInfo getLookupInf(String servicename)
    throws NotBoundException
  {
    ServiceData service = (ServiceData) servicetable.get(servicename);
    if (service == null)
      throw new NotBoundException();
    List list = service.getLinkedList();
    LeaseInfo leaseInfo = new LeaseInfo(list.size());

    BindingList first = (BindingList) list.get(0);
    leaseInfo.setLeasetime(first.leasetime);
    leaseInfo.setTimeStamp(first.timestamp, 0);
    leaseInfo.setIID(first.iid, 0);

    for (int i = 1; i < list.size(); i++) {
      BindingList bl = (BindingList) list.get(i);
      leaseInfo.setTimeStamp(bl.timestamp, i);
      leaseInfo.setIID(bl.iid, i);
    }
	
    return leaseInfo;
  }


  /**
   *
   */
  synchronized void viewChange(View view)
  {
    this.view = view;
    for (ServerData server : servertable.values()) {
      server.viewChange(view);
    }
  }


  /**
   *  Returns the list of all of the service names currently maintained
   *  by the registry.
   */
  synchronized String[] list()
  {
    String[] list = new String[servicetable.size()];
    int j = 0;
    for (String serviceName : servicetable.keySet())
      list[j++] = serviceName;
    return list;
  }
  
  /**
   * Returns the service data for the given service name.
   */
  synchronized ServiceData getService(String serviceName)
    throws NotBoundException
  {
    ServiceData service = servicetable.get(serviceName);
    if (service==null)
      throw new NotBoundException(serviceName);
    return service;
  }


  /**
   *  Returns a <CODE>MergingData</CODE> object containing the 
   *  information needed by the registry replicas contained in
   *  <CODE>dests</CODE> to merge their databases. 
   *  
   *  @param dests the set of replicas to which the merging information
   *    is addressed.
   *  @param me identifier of the local replica, needed to identify
   *    the origin of the data.
   */
  synchronized MergingData getState(MemberId[] dests, MemberId me)
  {
    MergingData merging = new MergingData();
    int size = servertable.size();

    merging.vmids = new VMID[size];
    merging.removed = new IntList[size];
    merging.bindings = new BindingList[size][];

    int j = 0;
    for (ServerData server : servertable.values()) {
      merging.vmids[j] = server.vmid;
      merging.removed[j] = server.removed;
      merging.bindings[j++] = server.getState(dests, me);
    }
    return merging;
  }


  /**
   *  Takes a <CODE>MergingData</CODE> object and inserts the data
   *  contained in it.
   *  
   *  @param merging the merging data sent by another replica
   *  @param dests the destination set of this merging data
   */
  synchronized void putState(MergingData merging, MemberId[] dests)
    throws RemoteException, ClassNotFoundException
  {
    for (int i=0; i < merging.vmids.length; i++) {
      if (log.isDebugEnabled())
        log.debug("putState: merging.vmids["+i+"]: " + merging.vmids[i]);
      ServerData server = get(merging.vmids[i]);
      server.putState(servicetable, merging.removed[i], merging.bindings[i]);
    }
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Private methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Returns the <CODE>ServerData</CODE> object associated with the 
   *  specified <CODE>VMID</CODE>, or creates a new one if not present.
   */
  private ServerData get(VMID vmid)
  {
    ServerData server = servertable.get(vmid);
    if (server == null) {
      server = new ServerData(vmid);
      server.viewChange(view);
      servertable.put(vmid, server);
    }
    return server;
  }

} // END RegistryTable
