/*
 * Copyright (c) 1998-2002 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package jgroup.relacs.registry;

import java.rmi.RemoteException;
import java.util.Map;

import jgroup.core.MemberId;
import jgroup.core.MemberTable;
import jgroup.core.VMID;
import jgroup.core.View;
import jgroup.util.IntList;

import org.apache.log4j.Logger;

/**
 *
 *
 *  @author Alberto Montresor
 *  @since Jgroup 0.4
 */
final class ServerData
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Logger
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Obtain logger for this class */
  private static final Logger log = Logger.getLogger(ServerData.class);


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /* Virtual machine identifier */
  VMID vmid;

  /* List of active bindings */
  BindingList bindings;

  /* List of removed bindings */
  IntList removed;

  /* New binding ids inserted during last view */
  private IntList nbids;

  /*
   * Member table containing the status of other dependable registry
   * instances.
   */
  MemberTable table;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructor
  ////////////////////////////////////////////////////////////////////////////////////////////

  ServerData(VMID vmid)
  {
    this.vmid = vmid;
    bindings = new BindingList();
    removed = new IntList();
    nbids = new IntList();
    table = new MemberTable();
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  void add(int key, BindingList bl)
  {
    if (!removed.contains(key)) {
      bindings.insert(bl);
      nbids.insert(key);
    }
  }

  BindingList remove(int key)
  {
    removed.insert(key);
    return (BindingList) bindings.remove(key);
  }
  
  BindingList lookup(int key)
  {
    return (BindingList) bindings.lookup(key);
  }

  void viewChange(View view)
  {
    table.viewChange(view);
    MemberId[] members = table.members();
    for (int i=0; i < members.length; i++) {
      switch (table.getStatus(members[i])) {
      case MemberTable.NEWMEMBER:
        /* This is a new member */
        if (log.isDebugEnabled())
          log.debug("New member " + members[i]);
        table.put(members[i], new IntList());
        break;
      case MemberTable.RECOVERING:
        /* This member has crashed and then recovered */
        if (log.isDebugEnabled())
          log.debug("New incarnation " + members[i]);
        table.put(members[i], new IntList());
        break;
      case MemberTable.SURVIVED:
        /* This member is survived from the previous view */
        if (log.isDebugEnabled())
          log.debug("Inserting nbids: " + nbids);
        IntList list = (IntList) table.get(members[i]);
        if (log.isDebugEnabled())
          log.debug("Before: list[" + members[i] + "] = " + list);
        list.merge(nbids);
        if (log.isDebugEnabled())
          log.debug("After : list[" + members[i] + "] = " + list);
        break;
      }
    }
    nbids.clean();
  }


  BindingList[] getState(MemberId[] dests, MemberId me)
  {
    IntList slist = ((IntList) table.get(me));
    slist = slist.copy();
    if (log.isDebugEnabled())
      log.debug("Starting slist: " + slist);
    for (int i = 0; i < dests.length; i++) {
      if (dests[i].equals(me))
        slist.remove((IntList) table.get(dests[i]));
    }
    if (log.isDebugEnabled())
      log.debug("Local removed: " + removed);
    slist.remove(removed);
    if (log.isDebugEnabled())
      log.debug("Final: " + slist);
    int[] bids = slist.toIntArray();
    BindingList[] ret = new BindingList[bids.length];
    BindingList scan = (BindingList) bindings.getFirst();
    int i = 0;
    while (scan != null && i < bids.length) {
      if (scan.key == bids[i]) {
        ret[i] = scan;
        scan = (BindingList) scan.getNext();
        i++;
      } else if (scan.key < bids[i]) {
        scan = (BindingList) scan.getNext();
      } else {
        i++;
      }
    }
    return ret;
  }

  void putState(Map<String, ServiceData> servicetable, IntList toremove, BindingList[] toadd)
    throws ClassNotFoundException, RemoteException
  {
    toremove.remove(removed);
    removed.merge(toremove);
    int[] array = toremove.toIntArray();
    for (int j = 0; j < array.length; j++) {
      BindingList bl = (BindingList) bindings.remove(array[j]);
      if (bl != null) {
        ServiceData service = servicetable.get(bl.servicename);
        if (service != null) {
          service.remove(bl);
          if (service.isEmpty())
            servicetable.remove(bl.servicename);
        }
      }
    }

    IntList nlist = new IntList();
    for (int i = 0; i < toadd.length; i++) {
      BindingList bl = toadd[i];
      if (!removed.contains(bl.key)) {
        ServiceData service = servicetable.get(bl.servicename);
        if (service == null) {
          service = new ServiceData(bl.serverClass, bl.servicename, bl.entry);
          servicetable.put(bl.servicename, service);
        }
        service.add(bl);
        bindings.insert(bl);
        nlist.insert(bl.key);
      }
    }

    // Add the new bindings to list of new bindings
    nbids.merge(nlist);
  }

} // END ServerData
