/*
 * Copyright (c) 1998-2002 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package jgroup.relacs.rmi;

import java.rmi.RemoteException;
import java.rmi.server.ExportException;
import java.util.HashSet;
import java.util.Iterator;

import jgroup.core.Callback;
import jgroup.core.MemberId;
import jgroup.core.View;
import jgroup.core.multicast.AckListener;
import net.jini.export.Exporter;
import net.jini.jeri.BasicILFactory;
import net.jini.jeri.BasicJeriExporter;
import net.jini.jeri.tcp.TcpServerEndpoint;


/**
 *  Acknowledgement listener for asynchronous invocations, that is
 *  invocations for which the client does not wait for the reply.
 *
 *  @author Alberto Montresor
 *  @author Hein Meling
 *  @since Jgroup 0.9
 */
public final class AsynchAckListener
  implements AckListener
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Callback to be notified at the end. */
  private Callback cb;

  /** Members from which we are waiting a result or an exception. */
  private HashSet waiting;

  /** The BasicJeriExporter used by this AckListener */
  private Exporter exporter;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructor
  ////////////////////////////////////////////////////////////////////////////////////////////

  public AsynchAckListener(Callback cb)
    throws RemoteException
  {
    this.cb = cb;
  }

  /**
   *  Get a reference for the remote AckListener.
   *
   *  @return
   *    A remote proxy for the <code>AsynchAckListener</code>.
   *  @throws ExportException
   *    Raised if the <code>AsyncAckListener</code> could not be exported.
   */
  public AckListener getRemoteAckListener()
    throws ExportException
  {
    exporter =
      new BasicJeriExporter(TcpServerEndpoint.getInstance(0), new BasicILFactory());
    return (AckListener) exporter.export(this);
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from the AckListener interface
  ////////////////////////////////////////////////////////////////////////////////////////////

  public void notifyView(View view)
    throws RemoteException
  {
    MemberId[] members = view.getMembers();
    for (int i = 0; i < members.length; i++)
      waiting.add(members[i]);
    /*
     * Do nothing; we don't need to collect results since we are
     * asynchronous.
     */
  }


  public void ack(MemberId id, int vpos, Object obj)
    throws RemoteException
  {
    waiting.remove(id);
    if (obj instanceof Exception)
      cb.exception(id, (Exception) obj);
    else
      cb.result(id, obj);
  }


  public void viewChange()
    throws RemoteException
  {
    if (!waiting.isEmpty()) {
      RemoteException remoteException = new RemoteException("Member unreachable.");
      for (Iterator iter = waiting.iterator(); iter.hasNext(); )
        cb.exception((MemberId) iter.next(), remoteException);
    }
    // Unexport this remote AckListener; no more remote calls should occur
    exporter.unexport(false);
  }

} // END AsynchAckListener
