/*
 * Copyright (c) 1998-2002 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package jgroup.relacs.rmi;

import java.io.InputStream;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.OutputStream;
import java.lang.reflect.InvocationHandler;
import java.lang.reflect.Method;

import jgroup.core.Callback;
import jgroup.core.InternalGMIListener;
import jgroup.core.InternalGMIService;
import jgroup.core.JgroupException;
import jgroup.core.MemberId;
import jgroup.core.multicast.MulticastListener;
import jgroup.core.multicast.MulticastService;
import jgroup.relacs.gmi.GroupAckListener;
import jgroup.util.Abort;

import org.apache.log4j.Logger;

/**
 *  Handler for group internal invocations.  It is used to dynamically
 *  generate a proxy for internal GMI.
 *
 *  Note that the <code>InternalGMIService</code> interface is not
 *  implemented by the <code>IntGroupHandler</code> layer, but rather by
 *  the dynamically generated proxy object.  The interface is included
 *  here, only to follow the rules of the layer configuration structure
 *  which does not deal with proxy based layers as is.  That is a layer
 *  must implement its service interface
 *  (<code>InternalGMIService</code> in this case), and since the
 *  dynamically generated proxy is the object actually implementing the
 *  service interface, we must do this trick.  The method from the
 *  <code>InternalGMIService</code> interface simply throws an
 *  <code>UnsupportedOperationException</code> and thus should never be
 *  called.
 *
 *  @author Alberto Montresor
 *  @author Hein Meling
 *  @since Jgroup 0.9
 */
public class IntGroupHandler
  implements InternalGMIService, InvocationHandler, MulticastListener
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Logger
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Obtain logger for this class */
  private static final Logger log = Logger.getLogger(IntGroupHandler.class);


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Static section
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Protocol name used to distinguish messages from other protocols. */
  private static final String PROTOCOL_NAME = "IGMI";

  /** Method invoke of interface InternalGMIService */
  private static Method invokeMethod;

  static {
    try {
      Class[] types = new Class[] { Method.class, Object[].class, Callback.class };
      Class cl = InternalGMIService.class;
      invokeMethod = cl.getDeclaredMethod("invoke", types);
    } catch (NoSuchMethodException e) {
      // Can't happen, unless InternalGMIService has been modified
      Abort.exit("Method invoke() in class InternalGMIService not found", e, 1);
    }
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  The method table holds entries for all IGMI methods implemented
   *  by the server and/or layer that is registered as a listener for
   *  IGMI events.
   */
  private ServerMethodTable methodTable;

  /** The multicast service */
  private MulticastService multicastService;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructors
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Constructs a new <code>IntGroupHandler</code> object.
   */
  private IntGroupHandler(MulticastService multicastService)
  {
    this.multicastService = multicastService;
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Static factory
  ////////////////////////////////////////////////////////////////////////////////////////////

  public static IntGroupHandler getLayer(MulticastService mcastService)
  {
    return new IntGroupHandler(mcastService);
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Layer interface methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Add a server/layer that is listening for internal group method
   *  invocations.
   *
   *  @exception IllegalStateException
   *    Thrown if this method has already been invoked with another
   *    listener for group method invocations.  It is not allowed to
   *    have multiple local listeners for group method invocations.
   *  @exception IllegalArgumentException
   *    Raised if the specified listener is not an InternalGMIListener.
   */
  public void addListener(Object listener)
  {
    if (listener instanceof InternalGMIListener) {
      /*
       * Compute the method table for the internal interfaces
       * implemented by the provided listener.
       */
      //FIXME use the new method table once it supports IGMI
      if (methodTable == null)
        methodTable = new ServerMethodTable();
      methodTable.addMethods(listener, InternalGMIListener.class);
    } else {
      /*
       * Since the InternalGMIListener is a marker interface, we cannot
       * throw an exception here since the application using the IGMI
       * service may actually use IGMI only in a layer or in the server
       * object, and thus the addListener() method may get invoked also
       * for objects (server or layer) that does not actually implement
       * a InternalGMIListener interface.  Thus, we just print a log
       * message indicating this.
       */
      if (log.isDebugEnabled()) {
        log.debug(listener.getClass().getName() + " does not implement an InternalGMIListener");
      }
    }
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from InternalGMIService interface
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Dummy method.  The method of the InternalGMIService interface is
   *  not implemented IntGroupHandler layer, but the interface is
   *  implemented by the dynamically generated proxy object.  This
   *  method is included here, only to follow the rules of the layer
   *  configuration structure which does not deal with proxy based
   *  layers as is.  That is a layer must implement its service
   *  interface (InternalGMIService in this case), and since the
   *  dynamically generated proxy is the object actually implementing
   *  the service interface, we must do this trick.
   */
  public Object invoke(Method m, Object[] args, Callback callback)
    throws Exception
  {
    throw new UnsupportedOperationException();
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from InvocationHandler interface
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * Processes a method invocation on a proxy instance and returns
   * the result.  This method will be invoked on an invocation handler
   * when a method is invoked on a proxy instance that it is
   * associated with.
   */
  public Object invoke(Object proxy, Method m, Object[] args)
    throws Throwable
  {
    if (log.isDebugEnabled())
      log.debug("igmi: " + m);
    boolean asynchronous = false;
    Callback callback = null;

    /*
     * Check whether the method has asynchronous semantics; if so, read
     * the arguments.  Otherwise, just use the method directly with
     * synchronous semantics.
     */
    if (m.equals(invokeMethod)) {
      asynchronous = true;
      m = (Method) args[0];
      callback = (Callback) args[2];
      args = (Object[]) args[1];
    }

    try {
      /*
       * Prepare message
       */
      long hash = methodTable.get(m);
      OutputStream message = multicastService.getMessage(PROTOCOL_NAME);
      ObjectOutputStream out = new MarshalOutputStream(message);
      out.writeLong(hash);
      int len = (args == null ? 0 : args.length); 
      out.writeInt(len);
      for (int i=0; i < len; i++)
        out.writeObject(args[i]);
      out.flush();

      if (asynchronous) {
        // Asynchronous semantics
        if (callback != null)
          multicastService.mcast(message,
              new AsynchAckListener(callback).getRemoteAckListener());
        else
          multicastService.mcast(message, null);
        return null;
      } else {
        // Synchronous semantics
        GroupAckListener listener = new GroupAckListener(this);
        multicastService.mcast(message, listener.getRemoteAckListener());
        return listener.getResults();
      }

    } catch (JgroupException e) {
      throw e;
    } catch (Exception e) {
      log.warn("Unable to perform internal group method invocation", e);
      throw e;
    }
  }

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from MulticastListener
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Returns a string naming the protocol implemented by this multicast
   *  listener.
   */
  public String getProtocolName() 
  {
    return PROTOCOL_NAME;
  }


  /**
   *  Upcall that is invoked by Jgroup to deliver a message
   *  <code>msg</code> related to the <i>IGMI</i> protocol.
   *
   *  @param inmsg      the stream from which the message may be read.
   */
  public Object deliverStream(InputStream inmsg, MemberId sender, int seqNo)
  {
    if (log.isDebugEnabled())
      log.debug("IntGroupHandler: received stream message");
    try {
      long hash;
      Object[] args;
      synchronized (inmsg) {
        /*
         * This must be synchronized on the inmsg object so as to avoid
         * concurrent access to the inmsg object, since most methods in 
         * the InMessage class modify its state (the position field in
         * particular), including methods such as seek(), read(), insert()
         * and so on.  Such concurrent access may occur when two competing
         * threads (Ehandler and Dispatcher) want to access the inmsg object.
         * For example, one thread may be reading the inmsg object for
         * delivery to an upper-layer (see IntGroupHandler.deliverStream()),
         * while another thread wants to send the inmsg to another member of
         * the group, and thus needs to seek to the START_DATA position;
         * see MsgMcast.getInMessage().
         */
        ObjectInputStream in = new MarshalInputStream(inmsg);
        hash = in.readLong();
        int size = in.readInt();
        args = new Object[size];
        for (int i=0; i < args.length; i++)
          args[i] = in.readObject();
      }

      /* Obtain the method from the method hash */
      Method m = methodTable.get(hash);
      Object server = methodTable.getServer(hash);
      try {
        return m.invoke(server, args);
      } catch (Exception e) {
        return e;
      }
    } catch (Exception e) {
      /*
       * If we for some reason received an incorrectly formated message;
       * we simply log the raised exception and ignore the message since
       * we cannot know its origin.  That is it may be spam.
       */
      log.warn("IntGroupHandler: Unable to unmarshal stream message", e);
      return null;
    }
  }


  /**
   *  Upcall that is invoked by Jgroup to deliver a message <code> obj</code>
   *  related to the group identified by <code> group</code>. In this case,
   *  the message is an object marshalled and unmarshalled through java
   *  serialization.
   *
   *  @param obj        the object containing the message.
   */
  public Object deliverObject(Object obj, MemberId sender, int seqNo)
  {
    throw new UnsupportedOperationException();
  }

} // END IntGroupHandler
