/*
 * Copyright (c) 1998-2002 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package jgroup.relacs.rmi;

import java.io.IOException;
import java.io.ObjectOutputStream;
import java.io.OutputStream;
import java.rmi.MarshalledObject;
import java.rmi.Remote;

/**
 *  This class overrides the replaceObject method to replace each
 *  exported remote object with its corresponding stub instance.
 *  This is obtained by using the <CODE>java.rmi.MarshalledObject</CODE>
 *  class, that contains a byte stream with the serialized 
 *  representation of an object given to its constructor.  The 
 *  contained object is serialized and deserialized with the same 
 *  serialization semantics used for marshaling and unmarshaling 
 *  parameters and return values of RMI calls. 
 *
 *  @author     Alberto Montresor
 *  @since      Jgroup 0.8
 */
public class      MarshalOutputStream
extends           ObjectOutputStream
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructors
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * Provide a way for subclasses that are completely reimplementing
   * MarshalOutputStream to not have to allocate private data just used by
   * this implementation of ObjectOutputStream.
   *
   * <p>If there is a security manager installed, this method first calls the
   * security manager's <code>checkPermission</code> method with a
   * <code>SerializablePermission("enableSubclassImplementation")</code>
   * permission to ensure it's ok to enable subclassing.
   *
   * @exception IOException   Thrown if not called by a subclass.
   *
   * @throws SecurityException
   *    if a security manager exists and its
   *    <code>checkPermission</code> method denies
   *    enabling subclassing.
   *
   */
  protected MarshalOutputStream()
    throws IOException
  {
    super();
    enableReplaceObject(true);
  }

  /**
   *
   */
  public MarshalOutputStream(OutputStream out)
    throws IOException
  {
    super(out);
    enableReplaceObject(true);
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  If the specified object is a Remote object, substitutes it with 
   *  a marshalled form of it.
   *  
   */
  protected Object replaceObject(Object obj)
    throws IOException
  {
    if (obj instanceof Remote)
      obj = new MarshalledObject(obj);
    return obj;
  }

} // END MarshalOutputStream



