/*
 * Copyright (c) 1998-2002 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package jgroup.relacs.rmi;

import java.lang.reflect.InvocationTargetException;
import java.rmi.RemoteException;
import java.rmi.server.ExportException;

import jgroup.core.MemberId;
import jgroup.core.View;
import jgroup.core.multicast.AckListener;
import jgroup.relacs.gm.DispatcherService;
import net.jini.export.Exporter;
import net.jini.jeri.BasicILFactory;
import net.jini.jeri.BasicJeriExporter;
import net.jini.jeri.tcp.TcpServerEndpoint;

import org.apache.log4j.Logger;

/**
 *  This class is used to collect return values of synchronous internal
 *  group method invocations. For each invocation, an object of this
 *  class is instantiated. When an invocation is performed, method
 *  notifyView() is invoked to communicate to this object the number and
 *  the identities of the members included in the view in which the
 *  method is executed.  Method ack() is used to add the return value
 *  returned by the specified member.  Method viewChange() is used to
 *  notify objects of this class that a new view has been installed and
 *  members not responding are to be considered unreachable.
 *
 *  @author Alberto Montresor
 *  @since Jgroup 0.9
 */
public final class SynchAckListener
  implements AckListener
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Logger
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Obtain logger for this class */
  private static final Logger log = Logger.getLogger(SynchAckListener.class);


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Array containing results */
  private Object[] results;
  
  /** Boolean array for results */
  private boolean[] completed;
  
  /** Counter */
  private int missing;

  /** Dispatcher used to move the dispatcher */
  private DispatcherService dispatcher;

  /** The BasicJeriExporter used by this AckListener */
  private Exporter exporter;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructor
  ////////////////////////////////////////////////////////////////////////////////////////////

  public SynchAckListener(DispatcherService dispatcher)
  {
    this.dispatcher = dispatcher;
  }

  /**
   *  Get a reference for the remote AckListener.
   *
   *  @return
   *    A remote proxy for the <code>SynchAckListener</code>.
   *  @throws ExportException
   *    Raised if the <code>SyncAckListener</code> could not be exported.
   */
  public AckListener getRemoteAckListener()
    throws ExportException
  {
    /*
     * Create a new exporter on any available port for this AckListener,
     * and enable the distributed garbage collection (DGC), and enable
     * the keep alive attribute.
     */
    exporter = new BasicJeriExporter(TcpServerEndpoint.getInstance(0),
        new BasicILFactory(), true, true);
    return (AckListener) exporter.export(this);
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from AckListener
  ////////////////////////////////////////////////////////////////////////////////////////////

  public synchronized void notifyView(View view)
    throws RemoteException
  {
    missing = view.size();
    completed = new boolean[missing];
    results = new Object[missing];
    if (log.isDebugEnabled())
      log.debug("SynchAckListener: notifyView: " + missing);
  }


  public synchronized void ack(MemberId id, int vpos, Object obj)
    throws RemoteException
  {
    if (log.isDebugEnabled())
      log.debug("SynchAckListener: ack: " + id + ", " + obj + ", " + vpos);
    if (obj instanceof InvocationTargetException) {
      Throwable targetException = ((InvocationTargetException) obj).getTargetException();
      log.warn("Caught: ", targetException);
    }

    if (!completed[vpos]) {
      completed[vpos] = true;
      results[vpos] = obj;
      missing--;
      if (missing == 0) {
        if (log.isDebugEnabled())
          log.debug("SynchAckListener: ack: completed");
        notifyAll();
      }
    }
  }


  public synchronized void viewChange()
    throws RemoteException
  {
    if (log.isDebugEnabled())
      log.debug("SynchAckListener: viewChange");
    if (missing > 0) {
      for (int i=0; i < completed.length; i++) {
        if (!completed[i]) {
          completed[i] = true;
          results[i] = new RemoteException("Member not reachable");
          if (log.isDebugEnabled())
            log.debug("SynchAckListener: viewChange: unreachable pos " + i);
        }
      }
    }
    missing = 0;
    // Unexport this remote AckListener; no more remote calls should occur
    exporter.unexport(false);
    // Ensure that any waiting ends.
    notifyAll();
  }


  /**
   * Wait for and obtain the results.
   */
  public synchronized Object getResults()
  {
    if (log.isDebugEnabled())
      log.debug("SynchAckListener: getResults");
    while (missing > 0 || results == null) {
      dispatcher.dispatch(this);
    }
    /*
     * Unexport this remote AckListener; note that calls to the
     * AckListener.viewChange() method may occur after the remote
     * AckListener has been unexported.  This should be dealt with
     * at the client-side, as this will cause a NoSuchObjectException
     */
    exporter.unexport(false);
    if (log.isDebugEnabled())
      log.debug("SynchAckListener: getResults completed");
    return results;
  }

} // END SynchAckListener
