/*
 * Copyright (c) 1998-2002 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 8 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package jgroup.relacs.types;

import java.io.IOException;
import java.io.ObjectInput;
import java.io.ObjectOutput;

import jgroup.core.EndPoint;

/**
 *  The <code>ViewId</code> class wraps a value of the relacs daemon
 *  type "group id" in an object.  A group id uniquely identifies
 *  groups.  An object of type <code>ViewId</code> contains a single
 *  field whose type is long.
 *
 *  @author Alberto Montresor
 *  @since  Jgroup 1.2
 */

public final class ViewId
{
  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constants
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Size of this object in bytes (when marshalled) */
  public static final int SIZE = 8;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////
  
  /** Wrapped value */
  private long value;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructor
  ////////////////////////////////////////////////////////////////////////////////////////////
  
  /**
   *  Creates a <CODE>ViewId</CODE> object containing the specified
   *  value.
   */
  public ViewId(long value)
  {
    this.value = value;
  }
  
  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Accessor method
  ////////////////////////////////////////////////////////////////////////////////////////////
  
  /**
   *  Returned the wrapped value.
   */
  public long getValue()
  {
    return value;
  }
  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Marshaling / unmarshaling methods
  ////////////////////////////////////////////////////////////////////////////////////////////
  
  /**
   *  Write the specified message id to an outgoing message.
   */
  public static void marshal(ObjectOutput msg, long value)
    throws IOException
  {
    msg.writeLong(value);
  }
  
  /**
   *  Unmarshals a view id from an incoming message.
   */
  public static long unmarshal(ObjectInput msg)
    throws IOException
  {
    return msg.readLong();
  }

  /**
   * Returns a new complete view identifier obtained from the host identifier and
   * a sequence number.
   *
   * @param     host    the host idientifer.
   * @param     counter sequence number.
   */
  public static long create(EndPoint host, int counter)
  {
    int hid = host.getIntAddress();
    // FIXME Should we use also port to construct the view id?
    return ((long)(hid) << 32) | (counter & 0xFFFFFFFFL);
  }
  
  /**
   * Returns a new partial view identifier obtained by a complete view identifier
   * and an additional sequence number.
   *
   * @param     cvid    the complete view identifier.
   * @param     counter the additional counter.
   */
  public static long createPartial(long cvid, int counter)
  {
    return (cvid + counter);
  }

} // END ViewId
