/*
 * Copyright (c) 1998-2002 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package jgroup.util;

import java.io.Externalizable;
import java.io.IOException;
import java.io.ObjectInput;
import java.io.ObjectOutput;

/**
 *  The <CODE>IntList</CODE> class represents ordered lists of integers.
 *  This class has been optimized to maintain list of integers 
 *  containing long sequence of consecutive numbers. It is particularly
 *  useful for maintaining list of received ack numbers, that are 
 *  normally consecutive (except in the case of connectivity problems).
 *  <p>
 *  The list is maintained by a chain of integer pairs; in each pair,
 *  the first integer represents the beginning of a consecutive sequence,
 *  while the second integer represents the end of a consecutive sequence.
 *  <p>
 *  For example, a list of this type:
 *  <pre> 1 2 3 4 5  7 8 9 10 11  13 14 15 16 17 18 19 20 21 22 23 </pre>
 *  is mantained in this way:
 *  <pre> (1-5),(7-11),(13-23) </pre>
 *  thus saving both memory and time to scan a list searching for a number.
 *  <p>
 *  <CODE>IntList</CODE> contains methods to add, lookup and remove 
 *  integers, and to merge the contents of other <CODE>IntList</CODE>s.
 *  <p>
 *  Implementation note: As mentioned, <CODE>IntList</CODE> is based on a 
 *  linked list of pairs. Each of these pairs are <CODE>IntList</CODE> 
 *  objects, to avoid the need to introduce a inner class to hold pairs. 
 *  For this reason, each integer list starts with an "empty" 
 *  <CODE>IntList</CODE> containing no pair. The first pair of an 
 *  <CODE>IntList</CODE> l is contained in <CODE>l.next</CODE>.
 *
 *  @author Alberto Montresor
 *  @author Hein Meling
 *  @since Jgroup 0.3
 */
public final class IntList
  implements Externalizable
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  private static final long serialVersionUID = -6522311553289690117L;

  /** First element in a consecutive sequence */
  private int start;

  /** Last element in a consecutive sequence */
  private int stop;
  
  /** Next sequence in the list */
  private IntList next;

  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructors
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Build an empty <CODE>IntList</CODE>. 
   */
  public IntList()
  {
  }

  /**
   *  Instantiates a pair representing a sequence
   *  starting and ending with <CODE>key</CODE>,
   *  and whose next element is <CODE>next</CODE>.
   *  For private use in the class.
   */
  private IntList(int key, IntList next)
  {
    this.start = key;
    this.stop  = key;
    this.next  = next;
  }

  /**
   *  Instantiates a pair representing a sequence
   *  starting with <CODE>start</CODE> and ending
   *  with <CODE>stop</CODE>. The next pair is set
   *  to null. For private use in the class.
   */
  private IntList(int start, int stop)
  {
    this.start = start;
    this.stop  = stop;
    this.next  = null;
  }

  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Returns the first element of the list, or -1 if the list is empty.
   */
  public int getFirst()
  {
    if (next != null)
      return next.start;
    else
      return -1;
  }

  /**
   *  Returns the last element of the list, or -1 if the list is empty.
   */
  public int getLast()
  {
    IntList scan = this.next;
    if (scan == null)
      return -1;
    while (scan.next != null)
      scan = scan.next;
    return scan.stop;
  }

  /**
   *  Returns true if the list is empty; false otherwise.
   */
  public boolean isEmpty()
  {
    if (next == null)
      return true;
    else 
      return false;
  }

  /**
   *  Returns the end value of the first consecutive sequence of integers 
   *  contained in the list. For example, if the list contains 
   *  <pre> 1 2 3 4 5  7 8 9 10 11 </pre>
   *  this methods return 5. If the list is empty, it returns -1.
   */
  public int getMaxConsecutive()
  {
    if (next != null)
      return next.stop;
    else
      return -1;
  }
  
  /**
   *  Returns true if value <CODE>key</CODE> is contained in this 
   *  <CODE>IntList</CODE>.
   */
  public boolean contains(int key)
  {
    // Used to scan the list
    IntList scan = this.next;
    while (scan != null && key >= scan.start) {
      if (key <= scan.stop)
        return true;
      scan = scan.next;
    }
    return false;
  }

  /**
   *  Insert value <CODE>key</CODE> in the <CODE>IntList</CODE>.
   */
  public void insert(int key)
  {
    // Used to scan the list
    IntList scan = this;

    // Scan the list
    while (scan.next != null && key > scan.next.stop + 1)
      scan = scan.next;
    if (scan.next == null || key < scan.next.start - 1) {
      // Append at the beginning
      scan.next = new IntList(key, scan.next);
    } else if (key == scan.next.stop + 1) {
      // Increment the stop position
      IntList item = scan.next.next;  // Temporary variable
      if (item != null && item.start == key + 1) {
        // Merge with the next item
        scan.next.stop = item.stop;
        scan.next.next = item.next;
      } else {
        // Just increment
        scan.next.stop = key;
      }
    } else if (key > scan.next.stop + 1) {
      // Append at the end
      scan.next.next = new IntList(key, scan.next.next);
    } else if (key == scan.next.start - 1) {
      // Decrement start
      scan.next.start = key;
    }
  }

  /**
   *  Merge <CODE>IntList addend</CODE> to the <CODE>IntList</CODE>.
   */
  public void merge(IntList addend)
  {
    if (addend == null)
      return;

    IntList scan = this;
    addend = addend.next;
    while (addend != null) {
      IntList snext = scan.next;
      if (snext == null) {
        // Append the rest of <addend> at the end of the list
        scan.next = new IntList(addend.start, addend.stop);
        addend = addend.next;
        scan = snext;
      }
      else if (addend.stop < snext.start - 1) {
        // Insert the current entry of <addend> before the
        // current entry of the list
        IntList temp = snext;
        scan.next = new IntList(addend.start, addend.stop);
        snext.next = temp;
        scan = snext;
        addend = addend.next;
      }
      else if (snext.stop+1 < addend.start) {
        // Skip to the next entry of the list
        scan = snext;
      }
      else {
        // The current entry of <addend> and the current entry of
        // the list intersect
        if (snext.start > addend.start)
          snext.start = addend.start;
        if (snext.stop < addend.stop) {
          snext.stop = addend.stop;
          IntList temp = snext.next;
          if (temp != null && temp.start <= addend.stop + 1) {
            // Merge two elements
            snext.next = temp.next;
            snext.stop = temp.stop;
          }
        }
        addend = addend.next;
      }
    }
  }


  /**
   *  Split the <code>IntList</code> at position <code>key</code>,
   *  leaving the upper part in this <code>IntList</code>, while
   *  returning the lower part of the list, <code>key</code> inclusive.
   *
   *  @param key  the position in the list to make the split.
   *  @return the lower part of this list is returned.
   */
  public IntList split(int key)
  {
    IntList ret = copy();
    IntList toRemove = new IntList();
    toRemove.next = new IntList(key + 1, getLast());
    ret.remove(toRemove);
    toRemove.next = new IntList(getFirst(), key);
    remove(toRemove);
    return ret;
  }


  /**
   *  Remove range from <code>start</code> to <code>stop</code>.  If
   *  <code>stop</code> is zero, the remainder of the list is deleted.
   */
  public void remove(int start, int stop)
  {
    IntList toRemove = new IntList();
    if (stop == 0)
      stop = getLast();
    if (getLast() != stop)
      stop--;
    toRemove.next = new IntList(start, stop);
    remove(toRemove);
  }


  /**
   *  Returns and removes the first element of the list; -1 is returned
   *  if the list is empty.
   */
  public int removeFirst()
  {
    if (next != null) {
      int ret = next.start;
      remove(ret);
      return ret;
    } else
      return -1;
  }


  /**
   *  Remove value <CODE>key</CODE> from the <CODE>IntList</CODE>.
   */
  public void remove(int key)
  {
    IntList toRemove = new IntList();
    toRemove.insert(key);
    remove(toRemove);
  }


  /**
   *  Remove <CODE>IntList removed</CODE> from the <CODE>IntList</CODE>.
   */
  public void remove(IntList removed)
  {
    if (removed == null)
      return;

    IntList scan = this;
    removed = removed.next;
    while (removed != null && scan.next != null) {
      if (removed.stop < scan.next.start) {
        /*
         * The current <removed> entry is less than the current entry of
         * the list; skip to the next <removed> entry.
         */
        removed = removed.next;
      }
      else if (scan.next.stop < removed.start) {
        /*
         * The current entry of the list is less than the current
         * <removed> entry; skip to the next entry of the list.
         */
        scan = scan.next;
      }
      else if (removed.start > scan.next.start && removed.stop < scan.next.stop) {
        /*
         * The current <removed> entry is included in a list entry;
         * split the list entry in two entries.
         */
        IntList temp = scan.next.next;
        scan.next.next = new IntList(removed.stop+1, scan.next.stop);
        scan.next.stop = removed.start-1;
        scan.next.next.next = temp;
      }
      else {
        /*
         * The current <removed> entry is included in a list entry,
         * either entirely or starting from the beginning or ending at
         * the same number; reorganize the list or remove it entirely.
         */
        if (removed.stop >= scan.next.stop)
          scan.next.stop = removed.start-1;
        if (removed.start <= scan.next.start)
          scan.next.start = removed.stop+1;
        if (scan.next.stop < scan.next.start)
          scan.next = scan.next.next;
      }
    }
  }

  /**
   *  Return an <CODE>int</CODE> array containing the values contained
   *  in this <CODE>IntList</CODE>.
   */
  public int[] toIntArray()
  {
    IntList scan;

    int size = 0;
    for (scan = this.next; scan != null; scan = scan.next)
      size += (scan.stop - scan.start + 1);
    int[] ret = new int[size];

    int i = 0;
    for (scan = this.next; scan != null; scan = scan.next)
      for (int j = scan.start; j <= scan.stop; i++, j++)
        ret[i] = j;
    return ret;
  }

  /**
   *  Clear the content of this <CODE>IntList</CODE>.
   */
  public void clean()
  {
    this.next = null;
  }

  /**
   *  Deep clone this <CODE>IntList</CODE>.
   */
  public IntList copy()
  {
    IntList ret = new IntList();
    IntList scan1 = ret;
    IntList scan2 = this.next;
    while (scan2 != null) {
      scan1.next = new IntList(scan2.start, scan2.stop);
      scan1 = scan1.next;
      scan2 = scan2.next;
    }
    return ret;
  }

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from External
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Write the contents of this <CODE>IntList</CODE> to
   *  the object output stream <CODE>out</CODE>. The external
   *  format of the string is the following: each pair is
   *  written to the stream, terminated by -1.
   * 
   *  @param out the stream to be written
   */
  public void writeExternal(ObjectOutput out)
    throws IOException
  {
    IntList scan = this.next;
    while (scan != null) {
      out.writeInt(scan.start);
      out.writeInt(scan.stop);
      scan = scan.next;
    }
    out.writeInt(-1);
  }

  /**
   *  Restores the content of this object from the marshalled data contained
   *  in the specified input stream.
   * 
   *  @param in the stream to be read
   */
  public void readExternal(ObjectInput in)
    throws IOException, ClassNotFoundException
  {
    // Used to construct the IntList
    IntList scan = this;

    // Start value
    int v1;
    while ((v1 = in.readInt()) != -1) {
      scan.next = new IntList(v1, in.readInt());
      scan = scan.next;
    }
  }

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Object methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Returns a string representation of this object
   */
  public String toString()
  {
    StringBuilder buf = new StringBuilder();
    buf.append("[IntList: {");
    IntList scan = next;
    while (scan != null) {
      buf.append(scan.start);
      buf.append("-");
      buf.append(scan.stop);
      scan = scan.next;
      if (scan != null)
        buf.append(", ");
    }
    buf.append("}]");
    return buf.toString();
  }

} // END IntList
