/*
 * Copyright (c) 1998-2002 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package jgroup.util;

import java.util.Calendar;


/**
 *  The <code>Util</code> class implements various useful static methods
 *  that can be invoked from anywhere.
 *
 *  @author Alberto Montresor
 *  @author Hein Meling
 *  @since Jgroup 0.1
 */
public final class Util
{

  /**
   *  Private constructor (Util cannot have instance objects)
   */
  private Util() { }


  /**
   * Searches the specified array of ints for the specified value using
   * the sequential search algorithm. The array <strong>must</strong>
   * be sorted prior to making this call.
   *
   * @param array       the array to be searched.
   * @param key         the value to be searched for.
   * @return            index of the search key, if it is contained in the array;
   *                    otherwise, array.length.
   */
  public static int valueAt(int[] array, int key)
  {
    int         i;                      // Counter

    for (i=0; i < array.length && array[i] < key; i++);
    return ((i < array.length && array[i] == key) ? i : -1);
  }

  public static int valueAt(Object[] array, Object key)
  {
    int         i;                      // Counter

    try {
      for(i=0; !array[i].equals(key); i++);
      return i;
    } catch (Exception e) {}
    return -1;
  }

  public static int referenceAt(Object[] array, Object key)
  {
    int         i;                      // Counter

    try {
      for(i=0; array[i] != key; i++);
      return i;
    } catch (Exception e) {}
    return -1;
  }

  
  /**
   *  Return true if an object equal to <CODE>obj</CODE> is contained in
   *  <CODE>array</CODE>.
   */
  public static boolean in(Object[] array, Object obj)
  {
    for (int i=array.length-1; i >=0; i--)
      if (array[i].equals(obj))
        return true;
    return false;
  }


  /**
   *  Return a stringified date value.  The format is as follows:<p>
   *
   *  <pre>yyyymmdd</pre>
   *
   *  @return A string representation of the current date.
   */
  public static String getDateString()
  {
    Calendar gc = Calendar.getInstance();
    StringBuilder buffer = new StringBuilder();
    buffer.append(Integer.toString(gc.get(Calendar.YEAR)));
    int month = gc.get(Calendar.MONTH) + 1;
    if (month < 10)
      buffer.append("0");
    buffer.append(Integer.toString(month));
    int day = gc.get(Calendar.DAY_OF_MONTH);
    if (day < 10)
      buffer.append("0");
    buffer.append(Integer.toString(day));
    return buffer.toString();
  }


  /**
   *  Return a stringified clock time value.  The format is as follows:<p>
   *
   *  <pre>hhmmss</pre>
   *
   *  @return A string representation of the current time.
   */
  public static String getTimeString()
  {
    Calendar gc = Calendar.getInstance();
    StringBuilder buffer = new StringBuilder();
    int hour = gc.get(Calendar.HOUR_OF_DAY);
    if (hour < 10)
      buffer.append("0");
    buffer.append(hour);
    int minute = gc.get(Calendar.MINUTE);
    if (minute < 10)
      buffer.append("0");
    buffer.append(minute);
    int second = gc.get(Calendar.SECOND);
    if (second < 10)
      buffer.append("0");
    buffer.append(second);
    return buffer.toString();
  }


  /**
   *  Return a stringified date and clock time value.
   *  The format is as follows:<p>
   *
   *  <pre>yyyymmdd-hhmm</pre>
   *
   *  @return A string representation of the current date and time.
   */
  public static String getDateTimeString()
  {
    Calendar gc = Calendar.getInstance();
    StringBuilder buffer = new StringBuilder();
    buffer.append(Integer.toString(gc.get(Calendar.YEAR)));
    int month = gc.get(Calendar.MONTH) + 1;
    if (month < 10)
      buffer.append("0");
    buffer.append(Integer.toString(month));
    int day = gc.get(Calendar.DAY_OF_MONTH);
    if (day < 10)
      buffer.append("0");
    buffer.append(Integer.toString(day));
    buffer.append("-");
    int hour = gc.get(Calendar.HOUR_OF_DAY);
    if (hour < 10)
      buffer.append("0");
    buffer.append(Integer.toString(hour));
    int minute = gc.get(Calendar.MINUTE);
    if (minute < 10)
      buffer.append("0");
    buffer.append(Integer.toString(minute));
    int second = gc.get(Calendar.SECOND);
    if (second < 10)
      buffer.append("0");
    buffer.append(Integer.toString(second));
    return buffer.toString();
  }


  /**
   *  Converts the given <code>byte</code> value to a string value
   *  aligned to three characters; it can be used to make message
   *  outputs print nicely.
   */
  public static String byte2str(byte value)
  {
    int val = (value & 0xFF);
    StringBuilder buf = new StringBuilder();
    if (val < 100) {
      buf.append("0");
      if (val < 10)
        buf.append("0");
    }
    buf.append(val);
    return buf.toString();
  }


  /**
   * Wrapper for the <code>Thread.sleep()</code> method.
   * 
   * @param time the time in milliseconds to sleep.
   */
  public static void sleep(long time)
  {
    try {
      Thread.sleep(time);
    } catch (InterruptedException e) {
      // Ignore interruptions
    }
  }

} // END Util
