/*
 * Copyright (c) 1998-2005 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */
package jgroup.util.log;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.ObjectOutputStream;

import jgroup.util.Network;
import jgroup.util.Util;

import org.apache.log4j.Logger;

/**
 * The event logger class provides methods used for logging specific events
 * to a disk file.  It uses a simple memory based buffer, which is flushed
 * to disk after a given number of log events, or on demand.  Also, it should
 * flush and close the file on "clean" system exits, using the shutdown hook
 * mechanism.
 * 
 * @author Rune Vestvik
 * @author Hein Meling
 * @since Jgroup 3.0
 */
public class Eventlogger
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Logger
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Obtain logger for this class */
  private static final Logger log = Logger.getLogger(Eventlogger.class);


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constants
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** The filename prefix for the event logs */
  public static final String PREFIX = "eventlog-";

  /** The number of event objects to store in memory before flushing */
  private static final int MEM_SIZE = 10000;

  /** True if the measurement function should be activated. */
  public static final boolean ENABLED;

  /** Initialize static variables */
  static {
    ENABLED = Boolean.getBoolean("jgroup.log.measurements");
    if (log.isDebugEnabled()) {
      log.debug("Measurement logging " + (ENABLED ? "enabled" : "disabled"));
    }
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * Holds true only if no events has yet been logged;
   * used to determine if a file needs to be created.
   * If the file is closed, this is reset to true to allow
   * creation of a new file.
   */
  private static boolean firstEvent = true;

  /** The object output stream to which events (objects) from this JVM should be logged */
  private static ObjectOutputStream out;

  /** Index into the log memory array */
  private static int index = 0;

  /** The log memory array */
  private static final Event[] memLog = new Event[MEM_SIZE];

  /** The shutdown thread */
  private static Thread shutdownThread;

  /** The log file currently active */
  private static String logFile;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Private methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  private static void onShutdown()
  {
    if (shutdownThread == null) {
      shutdownThread = new Thread("EventLogger-ShutdownHook") {
        public void run()
        {
          if (log.isDebugEnabled())
            log.debug("Running EventLogger-ShutdownHook");
          close();
        }
      };
      Runtime.getRuntime().addShutdownHook(shutdownThread);
    }
  }

  private static void initLog()
  {
    // Set to false to avoid creating multiple log files unecessarily
    firstEvent = false;
    onShutdown();

    String logDir = System.getProperty("jgroup.log.dir");
    int ran = (int) (Math.random() * 99);
    StringBuilder buf = new StringBuilder(logDir);
    buf.append(File.separator);
    buf.append(PREFIX);
    buf.append(Network.getLocalMachineName());
    buf.append("-");
    buf.append(Util.getTimeString());
    buf.append("-");
    buf.append(ran);
    logFile = buf.toString();
    if (log.isDebugEnabled())
      log.debug("Creating log file: " + logFile);
    try {
      out = new ObjectOutputStream(new FileOutputStream(logFile));
    } catch (Exception e) {
      log.warn("Failed to create: " + logFile, e);
    }
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Public logger methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * Log simple event.
   */
  public static void logEvent(String description)
  {
    logEvent(null, description);
  }

  /**
   * Log event with both type and description.
   */
  public static void logEvent(EventType type, String description)
  {
    logEvent(new Event(type, description));
  }

  /**
   * Log any event type.
   */
  public static void logEvent(Event event)
  {
    if (firstEvent)
      initLog();
    memLog[index++] = event;
    if (index == MEM_SIZE) 
      flushMem();
  }
  
  /**
   * Log simple event and flush to disk.
   */
  public static void logEventFlush(String description)
  {
    logEventFlush(null, description);
  }

  /**
   * Log event with both type and description and flush to disk.
   */
  public static void logEventFlush(EventType type, String description)
  {
    logEventFlush(new Event(type, description));    
  }

  /**
   * Log any event type and flush to disk.
   */
  public static void logEventFlush(Event event)
  {
    if (firstEvent)
      initLog();
    memLog[index++] = event;
    flushMem();
  }
  
  /**
   * Log the given event, ignoring system time and group identifier.
   */
  public static void logEventNoSysTime(String description)
  {
    logEvent(new Event(0, null, description));
  }

  public static void flushMem()
  {
    if (out == null) {
      log.warn("Failed to flush memory; file already closed: " + logFile);
      return;
    }
    try {
      synchronized (Eventlogger.class) {
        for (int i = 0; i < index; i++) {
          out.writeObject(memLog[i]);
        }
        out.flush();
      }
    } catch (IOException e) {
      log.warn("Failed to flush memory: " + logFile, e);
    }
    index = 0;
    if (log.isDebugEnabled())
      log.debug("Flushed the log file: " + logFile);
  }

  public static void close()
  {
    // Ignore duplicate calls
    if (out == null)
      return;
    flushMem();
    if (log.isDebugEnabled())
      log.debug("Closing log file: " + logFile);
    synchronized (Eventlogger.class) {
      try {
        out.close();
      } catch (IOException e) {
        // Ignore, may fail if the file was already closed from another thread.
      }
      out = null;
      // Ensure that a new file can be created after a file has been closed
      firstEvent = true;
    }
  }

} // END EventLogger
