/*
 * Copyright (c) 1998-2005 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */
package jgroup.util.log;

import java.net.InetAddress;

import jgroup.relacs.config.AppConfig;
import jgroup.util.Network;

/**
 * A replica event to be logged.
 * 
 * @author Hein Meling
 */
public class ReplicaEvent
  extends Event
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constants
  ////////////////////////////////////////////////////////////////////////////////////////////

  private static final long serialVersionUID = 442667830497627858L;

  /**
   * Supported replica event types.
   */
  public enum Type implements EventType {
    Created, Initialized, Leaving, Removed, ForcedRemove, Failed, Shutdown
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** The group identifier for this ReplicaEvent */
  private final int groupId;

  /** The host on which the event takes place or a remote host if it is a created event */
  private final InetAddress host;

  /** The AppConfig object associated with this event */
  private AppConfig thisApp;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructor
  ////////////////////////////////////////////////////////////////////////////////////////////

  public ReplicaEvent(EventType type, int groupId)
  {
    this(type, groupId, null);
  }

  public ReplicaEvent(EventType type, int groupId, InetAddress host)
  {
    super(type, "");
    this.groupId = groupId;
    if (host == null)
      this.host = super.eventHost;
    else
      this.host = host;
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Access Methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * @return Returns the groupId.
   */
  public int getGroupId()
  {
    return groupId;
  }

  /**
   * @return Returns the <code>AppConfig</code> object associated with
   *   this event.
   */
  public AppConfig getAppConfig()
  {
    if (thisApp == null)
      thisApp = AppConfig.getApplication(this.groupId);
    return thisApp;
  }

  /**
   * @return Returns the location of this event.
   */
  public InetAddress getLocation()
  {
    return host;
  }

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from Object
  ////////////////////////////////////////////////////////////////////////////////////////////

  public String toString()
  {
    StringBuilder buf = commonToString(true);
    buf.append(" @ ");
    if (host != null)
      buf.append(Network.getMachineName(host.getHostName()));
    buf.append(" ");
    buf.append(getAppConfig().getClassData().getShortName());
    buf.append(" gid=");
    buf.append(groupId);
    return buf.toString();
  }

} // END ReplicaEvent
